////////////////////////////////////////////////////////////////////////////////////
/*
*	DrawingEditor.java -
*			Sample editor program
*	* Author			Elpida Tzafestas
*	* Original release date		December 1997
*	* Latest release date		December 2005
*
*	Features:
*	1	Input of graphical elements with the buttons and mouse
*		* Input two end points for line
*		* Input two end points of a diameter for circle
*		* Input vertices for polygon, until 30 points are given
*			or control-c/C is pressed or double click is detected.
*	2	Code works both as an applet and as an application
*	3	Selection of a shape with click, multiple shapes with
*		shift-click.
*		Select all with control-a/A (alternatively use alt-a/A)
*	4	Mouse dragging of a shape
*	5	Move selected shapes with arrows
*		(move by 1 pixel with plain arrow, by 10 pixels with alt+arrow).
*/
////////////////////////////////////////////////////////////////////////////////////

import java.awt.*;
import java.awt.event.*;
import java.util.Vector;
import javax.swing.*;

public class DrawingEditor extends JPanel
	implements Constants, AboutDialogUse,
	ActionListener, KeyListener, MouseListener, MouseMotionListener
{
// Parent frame for standalone mode or applet
public DrawingFrame parent;
DrawingApplet applet;

// Shapes input by the user
Drawing target;

// Components
JPanel shapePanel;
JButton circle, line, polygon;

JPanel toolbar;
JButton
	deleteButton, clearButton, exampleButton,
	editButton, resizeButton, rotateButton,
	alignButton, distributeButton, fitButton, equalizeButton;
DrawingView drawingView;

// Selections
Vector selections = new Vector(10,5);

// Positions defined with the mouse
Vector positions = new Vector(10,5);
int curr_x = -1, curr_y = -1;

// Variable used for input
int input_flag = 0;

// Position set with the mouse
int posX = -1, posY = -1;

// Variables used for shapes dragging : flag and previous drag position
boolean shapeMouseDrag = false;
int prev_x = -1, prev_y = -1;

// Colors
static Color BackgroundColor = Color.white;
static Color InputColor = Color.red;

// Sizes
public static int SizeX=500, SizeY=400;
final static Dimension IMAGE_VIEW_DIMENSION = new Dimension(400,400);

// Help message
protected static String helpText =
	"1. Input of shapes\r\n"+
	"  1.1 Use the upper panel buttons (circle, line or polygon).\r\n"+
	"  1.2 Press a button first, then click on two points to input a line\r\n"+
	"    or a circle, or click on any number of points up to 30 to input\r\n"+
	"    a polygon. Polygon input terminates with double-click or <Control>-c/C.\r\n"+
	"  1.3 The newly inserted shape will have the default color.\r\n"+
	"  1.4 By definition, the position of a shape is its upper left\r\n"+
	"    corner (the leftmost or uppermost end for a segment).\r\n"+
	"\r\n"+
	"2. Selection of shapes\r\n"+
	"  2.1 Click on an shape to select it. For lines click over or close to the line.\r\n"+
	"  2.2 Use <Shift>-click to select multiple shapes\r\n"+
	"  2.3 Use <Control>-a/A to select all shapes. Alternatively use <Alt>-a/A\r\n"+
	"  2.4 Click on the free space to deselect all selected shapes.\r\n"+
	"\r\n"+
	"3. Translation of selected shapes\r\n"+
	"  3.1 Use plain arrow keys <Left>, <Down>, <Right> and <Up> to move the selected\r\n"+
	"    shapes 1 pixel to the left, bottom, right or top, respectively.\r\n"+
	"  3.2 Use <Alt>-ed arrow keys <Left>, <Down>, <Right> and <Up> to move them\r\n"+
	"    10 pixels to the left, bottom, right or top, respectively.\r\n"+
	"  3.3 Use the mouse to drag them to a new position. The dragging operation\r\n"+
	"    is cancelled if the target position is outside the drawing boundaries.\r\n"+
	"\r\n"+
	"4. Miscellaneous\r\n"+
	"  4.1 Press <F5> to refresh the display at any time.\r\n"+
	"  4.2 Press <F1> to open a help dialog.\r\n"+
	"  4.3 Press <F7> to open an 'About' Dialog.\r\n"+
	"  4.4 When in standalone (non-applet) mode, use also the <Refresh> option\r\n"+
	"    from the <Window> menu, the <Help> option and the <About> option\r\n"+
	"    from the <Help> menu.\r\n"+
	"  4.5 Use the <Clear> button in the right button panel, to open a new blank\r\n"+
	"    design. When in standalone (non-applet) mode, use also the <New>\r\n"+
	"    option from the <File> menu or press <Control>-n or <Control>-N.\r\n"+
	"  4.6 Use the <Example> button in the right button panel to open a new\r\n"+
	"    exemplar design.\r\n"+
	"\r\n";

public static void main(String args[])
{
	DrawingEditor ed = new DrawingEditor();
	DrawingFrame centralFrame = new DrawingFrame(ed);

	ed.parent = centralFrame;
	ed.init();

	centralFrame.setSize(SizeX,SizeY+100);
	centralFrame.setResizable(false);
	centralFrame.setTitle("Mini Drawing Editor");

	centralFrame.setVisible(true);
}

public Container getContentPane()
{
	if (parent != null)
		// Standalone mode
		return parent.getContentPane();
	else
		// Applet mode (applet != null)
		return applet.getContentPane();
}

public void init()
{
	initializeGUI();

	// Initialize variables
	target = new Drawing(this);
}

void initializeGUI()
{
//	Shape panel
	shapePanel = new JPanel();
	shapePanel.setLayout(new FlowLayout());
	shapePanel.setBackground(Color.gray);
	circle = new JButton("Circle");
	circle.addActionListener(this);
	line = new JButton("Line");
	line.addActionListener(this);
	polygon = new JButton("Polygon");
	polygon.addActionListener(this);
	shapePanel.add(circle);
	shapePanel.add(line);
	shapePanel.add(polygon);

//	Toolbar
	toolbar = new JPanel();
	toolbar.setLayout(new GridLayout(0,1));
	toolbar.setBackground(Color.gray);
	deleteButton = new JButton("Delete");
	deleteButton.addActionListener(this);
	clearButton = new JButton("Clear");
	clearButton.addActionListener(this);
	exampleButton = new JButton("Example");
	exampleButton.addActionListener(this);
	editButton = new JButton("Edit");
	editButton.addActionListener(this);
	resizeButton = new JButton("Resize");
	resizeButton.addActionListener(this);
	rotateButton = new JButton("Rotate");
	rotateButton.addActionListener(this);
	alignButton = new JButton("Align");
	alignButton.addActionListener(this);
	distributeButton = new JButton("Distribute");
	distributeButton.addActionListener(this);
	fitButton = new JButton("Fit");
	fitButton.addActionListener(this);
	equalizeButton = new JButton("Equalize");
	equalizeButton.addActionListener(this);
	toolbar.add(deleteButton);
	toolbar.add(clearButton);
	toolbar.add(exampleButton);
	toolbar.add(editButton);
	toolbar.add(rotateButton);
	toolbar.add(resizeButton);
	toolbar.add(alignButton);
	toolbar.add(distributeButton);
	toolbar.add(equalizeButton);
	toolbar.add(fitButton);

//	Initialize drawing view
	drawingView = new DrawingView(this,IMAGE_VIEW_DIMENSION);

//	Initialize layout
	getContentPane().setLayout(new BorderLayout());
	getContentPane().add("North",shapePanel);
	getContentPane().add("East",toolbar);
	getContentPane().add("West",drawingView);
}

public void repaint()
{
	super.repaint();
	if (drawingView != null)
		drawingView.repaint();
}

public void actionPerformed(ActionEvent evt)
{
  if (evt.getSource() instanceof JButton)
  {
	JButton tgt = (JButton)evt.getSource();

	if (tgt == line) input_flag = LINE;
	else if (tgt == circle) input_flag = CIRCLE;
	else if (tgt == polygon) input_flag = POLYGON;
	else if (tgt == clearButton)
	{
		clear();
		repaint();
	}
	else if (tgt == exampleButton)
	{
		loadExample();
		repaint();
	}
	drawingView.requestFocus();
  }
}

public void keyTyped(KeyEvent evt) {}
public void keyReleased(KeyEvent evt) {}
public void keyPressed(KeyEvent evt)
{
	char ch = evt.getKeyChar();
	int key = evt.getKeyCode();

	boolean
		shift=false, control=false, alt=false,
		single_selection, has_selections;
	Shape comp;

	shift = evt.isShiftDown();
	control = evt.isControlDown();
	alt = (evt.isMetaDown() || evt.isAltDown());

	boolean noModifiers = (!shift) && (!control) && (!alt);
	boolean onlyShift = (shift) && (!control) && (!alt);
	boolean onlyControl = (!shift) && (control) && (!alt);
	boolean onlyAlt = (!shift) && (!control) && (alt);
	boolean altControl = (!shift) && (control) && (alt);
	boolean shiftControl = (shift) && (control) && (!alt);
	boolean shiftAlt = (shift) && (!control) && (alt);
	boolean shiftAltControl = (shift) && (control) && (alt);

	single_selection = (selections.size() == 1);
	has_selections = (selections.size() > 0);

	if ((ch=='a') || (ch=='A')) ch='a';
	else if ((ch=='c') || (ch=='C')) ch='c';
	else ch=' ';

	// Help and about box
	if ((key == KeyEvent.VK_F1) && noModifiers)
	{
		help();
		return;
	}
	else if ((key == KeyEvent.VK_F7) && noModifiers)
	{
		about();
		return;
	}

	// Refresh
	if ((key == KeyEvent.VK_F5) && noModifiers)
	{
		repaint();
		return;
	}

	// Move with the arrows
	else if (key == KeyEvent.VK_LEFT)
	{
		if (!has_selections) return;
		if (noModifiers)
		{	// Move by 1
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				comp.moveBy(-1,0);
			}
			repaint();
		}
		else if (onlyAlt)
		{	// Move by 10
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				comp.moveBy(-10,0);
			}
			repaint();
		}
	}

	else if (key == KeyEvent.VK_RIGHT)
	{
		if (!has_selections) return;
		if (noModifiers)
		{	// Move by 1
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				comp.moveBy(1,0);
			}
			repaint();
		}
		else if (onlyAlt)
		{	// Move by 10
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				comp.moveBy(10,0);
			}
			repaint();
		}
	}

	else if (key == KeyEvent.VK_UP)
	{
		if (!has_selections) return;
		if (noModifiers)
		{	// Move by 1
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				comp.moveBy(0,-1);
			}
			repaint();
		}
		else if (onlyAlt)
		{	// Move by 10
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				comp.moveBy(0,-10);
			}
			repaint();
		}
	}

	else if (key == KeyEvent.VK_DOWN)
	{
		if (!has_selections) return;
		if (noModifiers)
		{	// Move by 1
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				comp.moveBy(0,1);
			}
			repaint();
		}
		else if (onlyAlt)
		{	// Move by 10
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				comp.moveBy(0,10);
			}
			repaint();
		}
	}

	else if ((ch == 'a') && (onlyControl || onlyAlt)) selectAll();
	else if ((ch == 'c') && (onlyControl || onlyAlt))
	{
		System.out.println("Control-C pressed.");
		if (input_flag == POLYGON)
		{
			if (positions.size() > 0)
			{
				comp = new Shape(POLYGON,positions);
				target.add(comp);
				reset_input();
				clearSelections();
				select(comp);
			}
			repaint();
			return;
		}
	}
}

public void mouseEntered(MouseEvent evt) {}
public void mouseExited(MouseEvent evt) {}
public void mousePressed(MouseEvent evt)
{
	int x = evt.getX(), y = evt.getY();
	Shape comp;

	boolean
		shift = evt.isShiftDown(),
		control = evt.isControlDown(),
		alt = evt.isMetaDown() || evt.isAltDown();

	drawingView.requestFocus();

	if (control || alt) return;
	
	// Input of a shape
	if (!target.includes(x,y)) {reset_input(); repaint(); return;}
	if (input_flag == LINE)
	{
			positions.addElement((Object)(new Point(x,y)));
			curr_x = x; curr_y = y;
			if (positions.size() == 2)
			{
				comp = new Shape(LINE,(Point)positions.elementAt(0),(Point)positions.elementAt(1));
				target.add(comp);
				reset_input();
				clearSelections();
				select(comp);
			}
			repaint();
			return;
	}
	else if (input_flag == CIRCLE)
	{
			positions.addElement((Object)(new Point(x,y)));
			curr_x = x; curr_y = y;
			if (positions.size() == 2)
			{
				comp = new Shape(CIRCLE,(Point)positions.elementAt(0),
							(Point)positions.elementAt(1));
				target.add(comp);
				reset_input();
				clearSelections();
				select(comp);
			}
			repaint();
			return;
	}
	else if (input_flag == POLYGON)
	{
			positions.addElement((Object)(new Point(x,y)));
			curr_x = x; curr_y = y;
			if ((positions.size() == Shape.SizeLimit) || (evt.getClickCount() == 2))
			{
				comp = new Shape(POLYGON,positions);
				target.add(comp);
				reset_input();
				clearSelections();
				select(comp);
			}
			repaint();
			return;
	}
	else
		// Selection with dragging and edition of shapes
	{
		posX = x;
		posY = y;
		int i = target.shapes.size()-1;
		while (i >= 0)
		{
			comp = (Shape)target.shapes.elementAt(i);
			if (comp.includes(x,y))
			{
				if (!shift) clearSelections();
				select(comp);
	
				// Start dragging selections
				setShapeMouseDrag();
				prev_x = x;
				prev_y = y;
				for (int ind=0;ind<selections.size();ind++)
					((Shape)selections.elementAt(ind)).startDragging();
				posX = x;
				posY = y;
				repaint();
				return;
			}
			i--;
		}
		clearSelections();
		repaint();
	}
}

public void mouseReleased(MouseEvent evt)
{
	int x = evt.getX(), y = evt.getY();
	Shape comp;

	drawingView.requestFocus();

	if (!shapeMouseDrag) return;
	if (shapeMouseDrag)
	{
		resetShapeMouseDrag();
		if (!target.includes(x,y)) return;
		Shape sel = (Shape)selections.elementAt(0);
		if (sel.hasBeenDragged())
		{
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				comp.dragBy(x-prev_x,y-prev_y);
				comp.confirmDragged(target);
			}
		}
		prev_x = -1; prev_y = -1;
	}
	repaint();
}

public void mouseClicked(MouseEvent evt) {}

public void mouseMoved(MouseEvent evt)
{
	int x = evt.getX(), y = evt.getY();
	drawingView.requestFocus();
	if (input_flag != 0)
		if (positions.size() > 0) {curr_x = x; curr_y = y; repaint();}
}

public void mouseDragged(MouseEvent evt)
{
	int x = evt.getX(), y = evt.getY();
	Shape comp;

	drawingView.requestFocus();

	if (!shapeMouseDrag) return;
	if (shapeMouseDrag)
	{
		for (int i=0;i<selections.size();i++)
		{
			comp = (Shape)selections.elementAt(i);
			comp.dragBy(x-prev_x,y-prev_y);
		}
		prev_x = x; prev_y = y;
	}
	repaint();
}

public void paint(Graphics g)
{
	Color pen = g.getColor();
	super.paint(g);
	g.setColor(BackgroundColor);
	g.fillRect(0,0,SizeX,SizeY);
	g.setColor(pen);
	target.draw(g);
	drawPositions(g);

	if (shapeMouseDrag)
		for (int i=0;i<selections.size();i++)
			((Shape)selections.elementAt(i)).
				drawDragged(g);
}

public void clearSelections()
{
	for (int i=0;i<selections.size();i++)
	{
		Shape comp = (Shape)selections.elementAt(i);
		comp.deselect();
	}
	selections.removeAllElements();
}

void reset_input()
{
	positions.removeAllElements();
	input_flag = 0;
	curr_x = -1;
	curr_y = -1;
}

void message(String mesg)
{
	System.out.println(mesg);
}

private void setShapeMouseDrag() {shapeMouseDrag = true;}
private void resetShapeMouseDrag() {shapeMouseDrag = false;}

public void clear()
{
	clearSelections();
	target.clear();
	reset_input();
}

public void loadExample()
{
	target.initExample();
}

void selectAll()
{
	clearSelections();
	for (int i=0;i<target.shapes.size();i++)
		select((Shape)target.shapes.elementAt(i));
	repaint();
}

void select(Shape s)
{
	if (s.isSelected()) return;
	s.select();
	selections.addElement(s);
}

void deselect(Shape s)
{
	s.deselect();
	selections.removeElement(s);
}

public void help()
{
	TextAreaDialog.open("Help",helpText);
}

public void about()
{
	AboutDialog.open(this,
		"About Drawing Editor",
		"Drawing Editor", "Editor for Geometric Shapes (Lines, Circles and Polygons)",
		"Elpida Tzafestas, 1997", "brensham@softlab.ece.ntua.gr",
		"http://www.softlab.ece.ntua.gr/~brensham");
}

void drawPositions(Graphics g)
{
//	This method tracks user input in red : all given points are drawn
//	and each one of them is connected to the next one with a straight
//	line. This arrangement is particularly useful for polygon input,
//	where one can see the intermediate state of the polygon as it
//	evolves. Circle and line input is traced in a similar manner.
//	The final mouse pointer is also tracked as moved around.

	Point prev, curr;
	Color pen = g.getColor();
	g.setColor(InputColor);

	if ((input_flag == POLYGON) || (input_flag == LINE))
	{
		if (positions.size() > 0)
		{
			prev = (Point)positions.elementAt(0);
			drawPoint(g,prev);
			for (int i=1;i<positions.size();i++)
			{
				curr = (Point)positions.elementAt(i);
				g.drawLine(prev.x,prev.y,curr.x,curr.y);
				prev=curr;
			}
			g.drawLine(prev.x,prev.y,curr_x,curr_y);
		}
	}
	else if (input_flag == CIRCLE)
	{
		if (positions.size() > 0) // It will be necessarily 1.
		{
			prev = (Point)positions.elementAt(0);
			drawPoint(g,prev);

			// Draw the circle from (prev.x,prev.y) to (curr_x,curr_y)
			int diameter = (int)Math.sqrt(((prev.x-curr_x)*(prev.x-curr_x))
					+ ((prev.y-curr_y)*(prev.y-curr_y)));
			Point up = new Point((prev.x+curr_x-diameter)/2,
						(prev.y+curr_y-diameter)/2);
			g.drawOval(up.x,up.y,diameter,diameter);
		}
	}

	g.setColor(pen);
}

void drawPoint(Graphics g, int x, int y)
	{g.drawLine(x,y,x,y);}

void drawPoint(Graphics g, Point po)
	{g.drawLine(po.x,po.y,po.x,po.y);}

}	// End DrawingEditor

class Drawing implements Constants
{
Vector shapes;
Dimension size;

DrawingEditor parent;

static Dimension DefaultSize=new Dimension(450,450);
static Dimension MaximumSize=new Dimension(1000,1000);

//	Light yellow background
Color backgroundColor=Colors.LightYellow;

private void init(Dimension dim, DrawingEditor editor)
{
	size = new Dimension(dim);
	parent = editor;

	shapes = new Vector(10,5);
}

Drawing(DrawingEditor editor)
{
	init(DefaultSize, editor);
}

Drawing(Dimension dim, DrawingEditor editor)
{
	init(dim, editor);
}

public Color getBackgroundColor() {return backgroundColor;}

public boolean includes(int x, int y)
{
	if ((x>0) && (x<size.width) && (y>0) && (y<size.height)) return true;
	return false;
}

private void setSize(Dimension dim)
{
	size = new Dimension(dim);
}

public void add(Shape comp)
{
	shapes.addElement(comp);
}

public void clear()
{
	shapes.removeAllElements();
}

public void draw(Graphics g)
{
	Color c = g.getColor();

	g.setColor(backgroundColor);
	g.fillRect(0,0,size.width,size.height);
	g.setColor(c);

	for (int i=0;i<shapes.size();i++)
		((Shape)(shapes.elementAt(i))).draw(g);
}

//////////////////////////////////////////////////////////////////
//			Example : Happy cartoon
//////////////////////////////////////////////////////////////////

public void initExample()
{
	Shape temp;

	int xpos1[] = {87,337,274,150,147};
	int ypos1[] = {225,228,344,344,343};
	shapes.addElement(new Shape(POLYGON,xpos1,ypos1,5,Color.black));
	int xpos2[] = {128,302,216,214};
	int ypos2[] = {225,227,337,337};
	shapes.addElement(new Shape(POLYGON,xpos2,ypos2,4,Color.white));
	int xpos3[] = {104,119,173,247,295,328,328,215,105,104};
	int ypos3[] = {109,176,231,232,188,109,66,35,66,66};
	shapes.addElement(new Shape(POLYGON,xpos3,ypos3,10,Colors.LightPink));

	temp = new Shape(CIRCLE,new Point(123,76),new Point(141,81));
	temp.setFillColor(Color.yellow);
	shapes.addElement(temp);
	
	temp = new Shape(CIRCLE,new Point(178,80),new Point(150,110));
	temp.setFillColor(Color.yellow);
	shapes.addElement(temp);

	temp = new Shape(LINE,new Point(183,93),new Point(243,93));
	temp.setFillColor(Color.blue);
	shapes.addElement(temp);
	
	temp = new Shape(CIRCLE,new Point(279,80),new Point(251,110));
	temp.setFillColor(Color.yellow);
	shapes.addElement(temp);
	
	temp = new Shape(CIRCLE,new Point(287,76),new Point(305,81));
	temp.setFillColor(Color.yellow);
	shapes.addElement(temp);
	
	temp = new Shape(CIRCLE,new Point(224,122),new Point(202,145));
	temp.setFillColor(Color.yellow);
	shapes.addElement(temp);

	int xpos4[] = {158,214,266,263,215,162,163};
	int ypos4[] = {158,186,157,173,212,173,173};
	shapes.addElement(new Shape(POLYGON,xpos4,ypos4,7,Color.yellow));

	temp = new Shape(CIRCLE,new Point(251,100),new Point(267,103));
	temp.setFillColor(Color.blue);
	shapes.addElement(temp);

	temp = new Shape(CIRCLE,new Point(150,99),new Point(166,102));
	temp.setFillColor(Color.blue);
	shapes.addElement(temp);
	
	int xpos5[] = {168,169,260,261,259};
	int ypos5[] = {233,287,233,286,286};
	shapes.addElement(new Shape(POLYGON,xpos5,ypos5,5,Color.red));

	temp = new Shape(CIRCLE,new Point(212,249),new Point(225,267));
	temp.setFillColor(Color.red);
	shapes.addElement(temp);
}

}	// End Drawing

class Shape implements Constants
{
protected boolean selected = false;
Point dragPosition = new Point(0,0);

Color fillColor = Color.yellow;
Color borderColor = Color.black;
static int SizeLimit = 100;
int type;
int size = 0;
private int xpoints[] = new int[SizeLimit];
private int ypoints[] = new int[SizeLimit];

Shape()
{
	type = LINE;
}

Shape(int typ, Point start, Point end)
{
	// Line or circle
	switch (typ)
	{
		case LINE:
		default:
			type = LINE;
			xpoints[0]=start.x; xpoints[1]=end.x;
			ypoints[0]=start.y; ypoints[1]=end.y;
			size=2;
			break;
		case CIRCLE:
			type = CIRCLE;
			int cx = (start.x+end.x)/2;
			int cy = (start.y+end.y)/2;
			int r = (int)(Math.sqrt(((start.x-end.x)*(start.x-end.x))
					+ ((start.y-end.y)*(start.y-end.y))) / 2);
			xpoints[0]= cx - r; xpoints[1]= cx + r;
			ypoints[0]= cy - r; ypoints[1]= cy + r;
			size=2;
			break;
	}
}

Shape(int typ, Vector points)
{
	// Polygon
	Point temp;
	size = Math.min(points.size(),SizeLimit);

	type = POLYGON;
	for (int i=0;i<size;i++)
	{
		temp = (Point)points.elementAt(i);
		xpoints[i]=temp.x;
		ypoints[i]=temp.y;
	}
}

Shape(int typ, int xpos[], int ypos[], int s)
{
	// Practically used only for polygons
	type = typ;
	size = Math.min(Math.min(Math.min(xpos.length,ypos.length), SizeLimit), s);
	for (int i=0;i<size;i++)
		xpoints[i]=xpos[i];
	for (int j=0;j<size;j++)
		ypoints[j]=ypos[j];
}

Shape(int typ, int xpos[], int ypos[], int s, Color c)
{
	// Practically used only for polygons or during copying
	type = typ;
	size = Math.min(Math.min(Math.min(xpos.length,ypos.length), SizeLimit), s);
	for (int i=0;i<size;i++)
		xpoints[i]=xpos[i];
	for (int j=0;j<size;j++)
		ypoints[j]=ypos[j];
	fillColor = c;
}

Shape(int typ, int xpos[], int ypos[], int s, Color f, Color b)
{
	// Practically used only for polygons or during copying
	type = typ;
	size = Math.min(Math.min(Math.min(xpos.length,ypos.length), SizeLimit), s);
	for (int i=0;i<size;i++)
		xpoints[i]=xpos[i];
	for (int j=0;j<size;j++)
		ypoints[j]=ypos[j];
	fillColor = f;
	borderColor = b;
}

public void setType(int typ)
{
	type = typ;
}

public void setFillColor(Color c) {fillColor = c;}

public void select() {selected = true;}
public void deselect() {selected = false;}
public boolean isSelected() {return selected;}

public boolean includes1(int x, int y)
{
	return rectangleIncludes(getPosition(), width(), height(), x, y);
}

public boolean rectangleIncludes(Point upperleft, int width, int height, int x, int y)
{
	if ((x >= upperleft.x) && (x <= (upperleft.x+width))
			&& (y >= upperleft.y) && (y <= (upperleft.y+height)))
		return true;
	return false;
}

public boolean includes(int x, int y)
{
	switch (type)
	{
		case LINE: return lineIncludes(x,y);
		case CIRCLE: return circleIncludes(x,y);
		case POLYGON: return polygonIncludes(x,y);
	}
	return false;
}

public void moveBy(Point po) { moveBy(po.x,po.y); }

public Point topleft() { return getPosition(); }

public String name()
{
	return nameFor(type);
}

static String nameFor(int typ)
{
	switch (typ)
	{
		case LINE: return "Line";
		case CIRCLE: return "Circle";
		case POLYGON: return "Polygon";
		default: return "Simple Shape";
	}
}

public void draw(Graphics g)
{
	drawUnselected(g);
	if (selected) drawSelected(g);
}

/*
protected void drawSelected1(Graphics g)
{
	Color pen = g.getColor();
	g.setColor(getBorderColor());
	drawSelectedRectangle(g,getPosition(),width(),height());
	g.setColor(pen);
}
*/

protected void drawSelectedRectangle(Graphics g, Point upperleft, int width, int height)
{
	if (selected)
	{
		drawSelectedPoint(g, upperleft.x, upperleft.y);
		drawSelectedPoint(g, upperleft.x, upperleft.y + (height/2));
		drawSelectedPoint(g, upperleft.x, upperleft.y + height);

		drawSelectedPoint(g, upperleft.x + (width / 2), upperleft.y);
		drawSelectedPoint(g, upperleft.x + (width / 2),
			upperleft.y + height);
		drawSelectedPoint(g, upperleft.x + width, upperleft.y);
		drawSelectedPoint(g, upperleft.x + width,
			upperleft.y + (height/2));
		drawSelectedPoint(g, upperleft.x + width,
			upperleft.y + height);
	}
}

protected void drawSelectedPoint(Graphics g, int x, int y)
{
	g.fillRect(x-2, y-2, 5, 5);
}

void drawDragged(Graphics g)
{ drawDraggedAt(g,dragPosition.x,dragPosition.y); }

public int width()
{
	int leftmost=10000, rightmost=-1, i;

	switch (type)
	{
		case CIRCLE: 
				leftmost = xpoints[0];
				rightmost = xpoints[1];
				break;
		case LINE:
		case POLYGON: 
		default:
				// Find rightmost and leftmost
				for (i=0;i<size;i++)
				{
					if (xpoints[i]>rightmost) rightmost=xpoints[i];
					if (xpoints[i]<leftmost) leftmost=xpoints[i];
				}
				break;
	}

	return rightmost-leftmost;
}

public int height()
{
	int bottommost=-1, topmost=10000, i;

	switch (type)
	{
		case CIRCLE: 
				topmost = ypoints[0];
				bottommost = ypoints[1];
				break;
		case LINE:
		case POLYGON: 
		default:
				// Find bottommost and topmost
				for (i=0;i<size;i++)
				{
					if (ypoints[i]<topmost) topmost=ypoints[i];
					if (ypoints[i]>bottommost) bottommost=ypoints[i];
				}
				break;
	}

	return bottommost-topmost;
}

public Point getPosition()
{
	int leftmost=10000, topmost=10000, i, j;
	switch (type)
	{
		case CIRCLE: 
				leftmost = xpoints[0];
				topmost = ypoints[0];
				break;
		case LINE:
		case POLYGON: 
		default:
				//	Find leftmost and topmost
				for (i=0;i<size;i++)
					if (xpoints[i]<leftmost) leftmost=xpoints[i];
				for (j=0;j<size;j++)
					if (ypoints[j]<topmost) topmost=ypoints[j];
				break;
	}
	return new Point(leftmost,topmost);
}

public Point bottomright()
{
	int rightmost=-1, bottommost=-1, i, j;
	switch (type)
	{
		case LINE:
		case POLYGON: 
		default:
			//	Find rightmost and bottommost
				for (i=0;i<size;i++)
					if (xpoints[i]>rightmost) rightmost=xpoints[i];
				for (j=0;j<size;j++)
					if (ypoints[j]>bottommost) bottommost=ypoints[j];
				break;
		case CIRCLE: 
				rightmost = xpoints[1];
				bottommost = ypoints[1];
				break;
	}
	return new Point(rightmost,bottommost);
}

public void setPosition(int x, int y)
{
	int leftmost=10000, topmost=10000, i, j;

//	Find leftmost and topmost
	Point po = topleft();

	// Reset topmost and leftmost (reset all points relative to them)
	for (i=0;i<size;i++)
		xpoints[i] += (x-po.x);
	for (j=0;j<size;j++)
		ypoints[j] += (y-po.y);
}

public void setPosition(Point po)
	{setPosition(po.x,po.y);}

void drawUnselected(Graphics g)
{
//	Point up;
	int diameter;
	Color pen = g.getColor();

	switch (type)
	{
		case LINE:
			g.setColor(borderColor);
			g.drawLine(xpoints[0],ypoints[0],
					xpoints[1],ypoints[1]);
			g.setColor(pen);
			break;
		case CIRCLE:
			diameter = xpoints[1] - xpoints[0];
			g.setColor(fillColor);
			g.fillOval(xpoints[0],ypoints[0],diameter,diameter);
			g.setColor(borderColor);
			g.drawOval(xpoints[0],ypoints[0],diameter,diameter);
			g.setColor(pen);
			break;
		case POLYGON:
			g.setColor(fillColor);
			g.fillPolygon(xpoints,ypoints,size);
			g.setColor(borderColor);
			g.drawPolygon(xpoints,ypoints,size);
			g.setColor(pen);
			break;
	}
}

void drawDraggedAt(Graphics g, int x, int y)
{
	Point up, offset, pos = getPosition();
	int diameter;
	Color pen = g.getColor();

	offset = new Point(x - getPosition().x, y - getPosition().y);

	switch (type)
	{
		case LINE:
			g.setColor(borderColor);
			g.drawLine(xpoints[0]+offset.x,ypoints[0]+offset.y,
					xpoints[1]+offset.x,ypoints[1]+offset.y);
			g.setColor(pen);
			break;
		case CIRCLE:
			g.setColor(borderColor);
			g.drawOval(xpoints[0]+offset.x,ypoints[0]+offset.y,
					xpoints[1]-xpoints[0],ypoints[1]-ypoints[0]);
			g.setColor(pen);
			break;
		case POLYGON:
			g.setColor(borderColor);
			int newxpoints[] = new int[SizeLimit];
			int newypoints[] = new int[SizeLimit];
			for (int i=0;i<size;i++)
			{
				newxpoints[i]=xpoints[i]+offset.x;
				newypoints[i]=ypoints[i]+offset.y;
			}
			g.drawPolygon(newxpoints,newypoints,size);
			g.setColor(pen);
			break;
	}
}

protected void drawSelected(Graphics g)
{
	Color pen = g.getColor();
	g.setColor(borderColor);

	switch (type)
	{
		case LINE: lineDrawSelected(g); break;
		case CIRCLE: circleDrawSelected(g); break;
		case POLYGON: polygonDrawSelected(g); break;
	}

	g.setColor(pen);
}

void lineDrawSelected(Graphics g)
{
	drawSelectedPoint(g, xpoints[0], ypoints[0]);
	drawSelectedPoint(g, (xpoints[0] + xpoints[1]) / 2, (ypoints[0] + ypoints[1])/ 2);
	drawSelectedPoint(g, xpoints[1], ypoints[1]);
}

void circleDrawSelected(Graphics g)
{
	//drawSelectedRectangle(g,new Point(xpoints[0],ypoints[0]),xpoints[1]-xpoints[0],ypoints[1]-ypoints[0]);
	Point tl = topleft();
	Point br = bottomright();
	drawSelectedRectangle(g,tl,br.x-tl.x,br.y-tl.y);
}

void polygonDrawSelected(Graphics g)
{
	Point upperleft = new Point(10000,10000);
	Point bottomright = new Point(0,0);
	int width, height;
	for (int i=0;i<size;i++)
	{
		if (xpoints[i] < upperleft.x) upperleft.x = xpoints[i];
		if (xpoints[i] > bottomright.x) bottomright.x = xpoints[i];
		if (ypoints[i] < upperleft.y) upperleft.y = ypoints[i];
		if (ypoints[i] > bottomright.y) bottomright.y = ypoints[i];
	}
	width = bottomright.x - upperleft.x;
	height = bottomright.y - upperleft.y;

	drawSelectedRectangle(g,upperleft,width,height);
}

public boolean lineIncludes(int x,int y)
{
	int lowX;
	int lowY;
	int upX;
	int upY;

	if (xpoints[0] <= xpoints[1])
	{
		lowX = xpoints[0]-1;
		upX = xpoints[1]+1;
	}
	else
	{
		lowX = xpoints[1]-1;
		upX = xpoints[0]+1;
	}
	if (ypoints[0] <= ypoints[1])
	{
		lowY = ypoints[0]-1;
		upY = ypoints[1]+1;
	}
	else
	{
		lowY = ypoints[1]-1;
		upY = ypoints[0]+1;
	}
	if ((x >= lowX) && (x <= upX) && (y >= lowY) & (y <= upY))
		return true;
	return false;
}

public boolean circleIncludes(int x, int y)
{
	return rectangleIncludes(new Point(xpoints[0],ypoints[0]),
					xpoints[1]-xpoints[0],ypoints[1]-ypoints[0],x,y);
}

public boolean polygonIncludes(int x, int y)
{
	Point upperleft = new Point(10000,10000);
	Point bottomright = new Point(0,0);
	int width, height;
	for (int i=0;i<size;i++)
	{
		if (xpoints[i] < upperleft.x) upperleft.x = xpoints[i];
		if (xpoints[i] > bottomright.x) bottomright.x = xpoints[i];
		if (ypoints[i] < upperleft.y) upperleft.y = ypoints[i];
		if (ypoints[i] > bottomright.y) bottomright.y = ypoints[i];
	}
	width = bottomright.x - upperleft.x;
	height = bottomright.y - upperleft.y;

	return rectangleIncludes(upperleft, width, height, x, y);
}

public void moveBy(int x, int y)
{
	int i;
	for (i=0;i<xpoints.length;i++)
		{xpoints[i]+=x;}
	for (i=0;i<ypoints.length;i++)
		{ypoints[i]+=y;}
}

public void dragBy(int x, int y) {dragPosition.x +=x; dragPosition.y +=y;}
public void startDragging()
		{dragPosition.x = getPosition().x; dragPosition.y = getPosition().y;}
public boolean hasBeenDragged()
{
	Point pos = getPosition();
	return ((dragPosition.x != pos.x) || (dragPosition.y != pos.y));
}

public void confirmDragged(Drawing target)
{
	if (target.includes(dragPosition.x,dragPosition.y) &&
		target.includes(dragPosition.x+width(),
						dragPosition.y+height()))
	setPosition(dragPosition.x,dragPosition.y);
}

}	// End Shape

class DrawingView extends JPanel
{
DrawingEditor gui;
private Dimension size;

DrawingView(DrawingEditor ed, Dimension dim)
{
	gui = ed;
	size = new Dimension(dim);

	addKeyListener(gui);
	addMouseListener(gui);
	addMouseMotionListener(gui);
}

public Dimension getPreferredSize()
	{return new Dimension(size.width,size.height);}

public void paint(Graphics g)
{
	if (g == null) return;
	super.paint(g);

	// Draw the editor view
	gui.paint(g);
}

}	// End DrawingView

class DrawingFrame extends JFrame implements ActionListener,
				KeyListener, MouseListener, MouseMotionListener
{
DrawingEditor target;

JMenuItem
	fileNew, fileExit, editSelectAll,
	windowRefresh, helpHelp, helpAbout;

DrawingFrame(DrawingEditor app)
{
	super();
	target = app;

	initBar();
}

private void initBar()
{
	JMenuBar bar = new JMenuBar();

	JMenu fileMenu = new JMenu("File");
	fileNew = new JMenuItem("New",KeyEvent.VK_N);
	fileNew.addActionListener(this);
	fileExit = new JMenuItem("Exit");
	fileExit.addActionListener(this);
	fileMenu.add(fileNew);
	fileMenu.addSeparator();
	fileMenu.add(fileExit);

	JMenu editMenu = new JMenu("Edit");
	editSelectAll = new JMenuItem("Select All",KeyEvent.VK_A);
	editSelectAll.addActionListener(this);
	editMenu.add(editSelectAll);

	JMenu windowMenu = new JMenu("Window");
	windowRefresh = new JMenuItem("Refresh -- F5");
	windowRefresh.addActionListener(this);
	windowMenu.add(windowRefresh);

	JMenu helpMenu = new JMenu("Help");
	helpHelp = new JMenuItem("Help -- F1");
	helpHelp.addActionListener(this);
	helpAbout = new JMenuItem("About -- F7");
	helpAbout.addActionListener(this);
	helpMenu.add(helpHelp);
	helpMenu.addSeparator();
	helpMenu.add(helpAbout);

	bar.add(fileMenu);
	bar.add(editMenu);
	bar.add(windowMenu);
	bar.add(helpMenu);
	setJMenuBar(bar);

	addKeyListener(this);
	addMouseListener(this);
	addMouseMotionListener(this);
	addWindowListener(new WindowAdapter()
	{
		public void windowClosing(WindowEvent evt)
			{dispose();System.exit(1);}
	});
	requestFocus();
}

public void keyPressed(KeyEvent evt) {target.keyPressed(evt); requestFocus();}
public void keyTyped(KeyEvent evt) {target.keyTyped(evt); requestFocus();}
public void keyReleased(KeyEvent evt) {target.keyReleased(evt); requestFocus();}

public void mousePressed(MouseEvent evt) {target.mousePressed(evt); requestFocus();}
public void mouseClicked(MouseEvent evt) {target.mouseClicked(evt); requestFocus();}
public void mouseReleased(MouseEvent evt) {target.mouseReleased(evt); requestFocus();}
public void mouseEntered(MouseEvent evt) {target.mouseEntered(evt); requestFocus();}
public void mouseExited(MouseEvent evt) {target.mouseExited(evt); requestFocus();}
public void mouseMoved(MouseEvent evt) {target.mouseMoved(evt); requestFocus();}
public void mouseDragged(MouseEvent evt) {target.mouseDragged(evt); requestFocus();}

public void actionPerformed(ActionEvent evt)
{
  if (evt.getSource() instanceof JMenuItem)
  {
	JMenuItem tgt = (JMenuItem)evt.getSource();

	if (tgt == fileNew)
		//	New file
		doNewFile();
	else if (tgt == fileExit) {dispose();System.exit(1);}
	else if (tgt == editSelectAll) target.selectAll();
	else if (tgt == windowRefresh) target.repaint();
	else if (tgt == helpHelp) target.help();
	else if (tgt == helpAbout) target.about();
  }
  requestFocus();
}

public void doNewFile()
{
	target.clear();
	target.message("New design");
	target.repaint();
}

}	// End DrawingFrame


////////////////////////////////////////////////////
//
//			Various dialogs
//
////////////////////////////////////////////////////

class TextAreaDialog extends JDialog implements ActionListener
{
protected static JFrame localFrame;
protected JButton okButton;
protected JTextArea message;

TextAreaDialog(JFrame parent, String title, String text)
{
	super(parent,title,true);

	okButton = new JButton("OK");
	okButton.addActionListener(this);
	JPanel okPanel = new JPanel();
	okPanel.setLayout(new FlowLayout(FlowLayout.CENTER));
	okPanel.add(okButton);

	message = new JTextArea(text,20,70);
	message.setEditable(false);
	JScrollPane messagePanel = new JScrollPane();
	messagePanel.setPreferredSize(new Dimension(550,550));
	messagePanel.setViewportView(message);

	getContentPane().setLayout(new BorderLayout());
	getContentPane().add("North",messagePanel);
	getContentPane().add("South",okPanel);

	addWindowListener(new WindowAdapter()
	{
		public void windowClosing(WindowEvent e)
		{
			setVisible(false);
			if (localFrame != null)
				localFrame.setVisible(false);
		}
	});
}

public void actionPerformed(ActionEvent evt)
{
	if ((JButton)evt.getSource() == okButton)
	{
		setVisible(false);
		if (localFrame != null)
			localFrame.setVisible(false);
	}
}

// Shortcut to create a frame automatically, the frame is a static
// variable, so that all message dialogs can use the same frame

public static void open(String title, String text)
{
	// If the frame has not been created yet, create it.
	if (localFrame == null)
		localFrame = new JFrame("");
	// Create the dialog now
	TextAreaDialog dlg = new TextAreaDialog(localFrame,title,text);

	// Shrink the frame to just fit the dialog
	localFrame.setSize(190,130);

	// Show the dialog
	dlg.pack();
	dlg.setVisible(true);
}

}	// End TextAreaDialog

class AboutDialog extends JDialog implements ActionListener
{
protected static JFrame localFrame;
protected JLabel title1, title2, author, mailto, link;
protected JButton okButton = new JButton("OK");

static Color backColor = Color.cyan;

private AboutDialogUse target;

AboutDialog(JFrame parent, AboutDialogUse app, String title,
	String tit1, String tit2, String auth, String mto, String http)
{
	super(parent,title,true);

	target = app;

	title1 = new JLabel(tit1,JLabel.CENTER);
	title1.setFont(new Font("Dialog",Font.BOLD,18));
	title2 = new JLabel(tit2,JLabel.CENTER);
	title2.setFont(new Font("Dialog",Font.BOLD,16));
	author = new JLabel(auth,JLabel.CENTER);
	author.setFont(new Font("Courier",Font.PLAIN,14));
	mailto = new JLabel(mto,JLabel.CENTER);
	mailto.setFont(new Font("Courier",Font.ITALIC,12));
	link = new JLabel(http,JLabel.CENTER);
	link.setFont(new Font("Courier",Font.ITALIC,12));

//	Component foreground colors
	title1.setForeground(Color.blue);
	title2.setForeground(Color.black);
	author.setForeground(Color.magenta);
	mailto.setForeground(Color.gray);
	link.setForeground(Color.gray);
	okButton.setForeground(Color.black);
	okButton.addActionListener(this);

	initLayout();

	addWindowListener(new WindowAdapter()
	{
		public void windowClosing(WindowEvent e)
		{
			setVisible(false);
			if (localFrame != null)
				localFrame.setVisible(false);
		}
	});
}

private void initLayout()
{
	JPanel componentPanel = new JPanel();
	componentPanel.setLayout(new GridLayout(0,1));
	componentPanel.setBackground(backColor);

	componentPanel.add(title1);
	componentPanel.add(title2);
	componentPanel.add(author);
	componentPanel.add(mailto);
	componentPanel.add(link);

	JPanel buttonPanel = new JPanel();
	buttonPanel.setLayout(new FlowLayout());
	buttonPanel.setBackground(backColor);
	buttonPanel.add(okButton);
	getContentPane().setLayout(new BorderLayout());
	setBackground(backColor);
	setForeground(backColor);
	getContentPane().add("North",componentPanel);
	getContentPane().add("Center",buttonPanel);

	setSize(500,190);
	setResizable(false);
}

public void actionPerformed(ActionEvent evt)
{
	if ((JButton)evt.getSource() == okButton)
	{
		setVisible(false);
		if (localFrame != null)
			localFrame.setVisible(false);
	}
}

// Shortcut to create a frame automatically, the frame is a static
// variable, so that all message dialogs can use the same frame

public static void open(AboutDialogUse app, String title,
		String tit1, String tit2, String auth, String mto, String address)
{
	// If the frame has not been created yet, create it.
	if (localFrame == null)
		localFrame = new JFrame("");
	// Create the dialog now
	AboutDialog dlg = new AboutDialog(localFrame,app,title,
						tit1,tit2,auth,mto,address);

	dlg.setVisible(true);
}

}	// End AboutDialog

interface AboutDialogUse {}

interface Colors
{
final Color LightYellow = new Color(255,255,200);
final Color LightPink = new Color(255,220,220);
final Color LightGreen = new Color(200,255,200);
final Color LightMagenta = new Color(255,200,255);
final Color Brown = new Color(185,120,100);
}

interface Constants
{
// Constants used for input
static int LINE = 1;
static int CIRCLE = 2;
static int POLYGON = 3;

}	// End Constants


