////////////////////////////////////////////////////////////////////////////////////
/*
*	DrawingEditor.java -
*			Sample editor program
*	* Author			Elpida Tzafestas
*	* Original release date		December 1997
*	* Latest release date		December 2005
*
*	Features:
*	1	Input of graphical elements with the buttons and mouse
*		* Input two end points for line
*		* Input two end points of a diameter for circle
*		* Input vertices for polygon, until 30 points are given
*			or control-c/C is pressed or double click is detected.
*	2	Code works both as an applet and as an application
*	3	Selection of a shape with click, multiple shapes with
*		shift-click.
*		Select all with control-a/A (alternatively use alt-a/A)
*		Select with the aid of a rubberband drawn around shapes by dragging
*	4	Delete selected shapes with the <DELETE> or <BACKSPACE> key
*	5	Clipboard : control-x/X, control-c/C and control-v/V, for
*		cut, copy and paste of shapes, respectively.
*		Alternatively use alt-x/X, alt-c/C and alt-v/V
*	6	Mouse dragging of a shape
*	7	Move selected shapes with arrows
*		(move by 1 pixel with plain arrow, by 10 pixels with alt+arrow).
*	8	Edit shapes fill and border colors by double-clicking.
*	9	I/O of designs when in standalone mode
*	10	Undo operations with control-z/Z. Alternatively use alt-z/Z
*	11	Redo the previous editing operation with control-y/Y. Alternatively
*		use alt-y/Y. Redo-able editing operations for selected shapes are move,
*		set the fill and border colors and delete
*/
////////////////////////////////////////////////////////////////////////////////////

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.StringTokenizer;
import java.util.Vector;
import javax.swing.*;

public class DrawingEditor extends JPanel
	implements Constants, AboutDialogUse,
	ActionListener, KeyListener, MouseListener, MouseMotionListener
{
// Parent frame for standalone mode or applet
public DrawingFrame parent;
DrawingApplet applet;

// Shapes input by the user
Drawing target;

// Components
JPanel shapePanel;
JButton circle, line, polygon;

JPanel toolbar;
JButton
	deleteButton, clearButton, exampleButton,
	editButton, resizeButton, rotateButton,
	alignButton, distributeButton, fitButton, equalizeButton;
DrawingView drawingView;

// Selections, clipboard and operations
Vector selections = new Vector(10,5);
Vector clipboard = new Vector(10,5);
Vector operations = new Vector(10,5);

// Redo data
int redo_operator = 0;
int redo_operands_size = 6;
int redo_operands[] = new int[redo_operands_size];

// Positions defined with the mouse
Vector positions = new Vector(10,5);
int curr_x = -1, curr_y = -1;

// Variable used for input
int input_flag = 0;

// Position set with the mouse
int posX = -1, posY = -1;

// Variables used for shapes dragging : flag and previous drag position
boolean shapeMouseDrag = false;
int prev_x = -1, prev_y = -1;

//	Variables used for rubberbanding : flag and start position
boolean rubberbandMouseDrag = false;
Point rubberbandStart = new Point(-1,-1);
Point rubberbandEnd = new Point(-1,-1);

// Colors
static Color BackgroundColor = Color.white;
static Color InputColor = Color.red;

// Sizes
public static int SizeX=500, SizeY=400;
final static Dimension IMAGE_VIEW_DIMENSION = new Dimension(400,400);

// Help message
protected static String helpText =
	"1. Input of shapes\r\n"+
	"  1.1 Use the upper panel buttons (circle, line or polygon).\r\n"+
	"  1.2 Press a button first, then click on two points to input a line\r\n"+
	"    or a circle, or click on any number of points up to 30 to input\r\n"+
	"    a polygon. Polygon input terminates with double-click or <Control>-c/C.\r\n"+
	"  1.3 The newly inserted shape will have the default color.\r\n"+
	"  1.4 By definition, the position of a shape is its upper left\r\n"+
	"    corner (the leftmost or uppermost end for a segment).\r\n"+
	"  1.5 A shape's foreground color may be edited, as well as the view's background\r\n"+
	"    color. (Cf. Clipboard Operations below)\r\n"+
	"\r\n"+
	"2. Selection of shapes\r\n"+
	"  2.1 Click on an shape to select it. For lines click over or close to the line.\r\n"+
	"  2.2 Use <Shift>-click to select multiple shapes\r\n"+
	"  2.3 Use <Control>-a/A to select all shapes. Alternatively use <Alt>-a/A\r\n"+
	"  2.4 Click on the free space to deselect all selected shapes.\r\n"+
	"  2.5 Select all shapes completely included in a rubberband drawn by pressing\r\n"+
	"    the mouse on a start point, then dragging it around and releasing it on an end point.\r\n"+
	"\r\n"+
	"3. Deletion of selected shapes\r\n"+
	"  3.1 Use the <Delete> button on the right button panel or press the <Delete>\r\n"+
	"    or <Backspace> key.\r\n"+
	"  3.2 Use <Control>-x/X to delete the selected shapes and put them on the\r\n"+
	"    clipboard. Alternatively use <Alt>x/X.\r\n"+
	"    (Cf. Clipboard Operations below)\r\n"+
	"\r\n"+
	"4. Translation of selected shapes\r\n"+
	"  4.1 Use plain arrow keys <Left>, <Down>, <Right> and <Up> to move the selected\r\n"+
	"    shapes 1 pixel to the left, bottom, right or top, respectively.\r\n"+
	"  4.2 Use <Alt>-ed arrow keys <Left>, <Down>, <Right> and <Up> to move them\r\n"+
	"    10 pixels to the left, bottom, right or top, respectively.\r\n"+
	"  4.3 Use the mouse to drag them to a new position. The dragging operation\r\n"+
	"    is cancelled if the target position is outside the drawing boundaries.\r\n"+
	"\r\n"+
	"5. Edition\r\n"+
	"  Double click on an shape to edit it or use the <Edit> button on the\r\n"+
	"  right button panel. This button has effect only in the case of a\r\n"+
	"  single object selected. Edition concerns changing its fill and border\r\n"+
	"  colors.\r\n"+
	"\r\n"+
	"6. Clipboard operations\r\n"+
	"  Use <Control>-c/C to copy the selected shapes to the clipboard,\r\n"+
	"  <Control>-x/X to cut the selected shapes, <Control>-v/V to paste them\r\n"+
	"  on a free position defined by clicking. Alternatively use <Alt>-c/C, <Alt>-x/X and <Alt>-v/V\r\n"+
	"\r\n"+
	"7. Cancellation of operations\r\n"+
	"  Use <Control>-z/Z to cancel previous operations. Alternatively use <Alt>-z/Z\r\n"+
	"\r\n"+
	"8. Redo editing operations\r\n"+
	"  Redo the previous editing operation with control-y/Y. Alternatively use\r\n"+
	"  use alt-y/Y. Redo-able editing operations for selected shapes are move,\r\n"+
	"  set the fill and border colors and delete.\r\n"+
	"\r\n"+
	"9. I/O of designs in standalone (non-applet) mode\r\n"+
	"  9.1 Use <Control>-o/O or option <Open> from the <File> menu to open a design file.\r\n"+
	"  9.2 Use <Control>-s/S or option <Save> from the <File> menu to save\r\n"+
	"    a previously opened drawing file.\r\n"+
	"    Use option <Save As> from the <File> menu to save a drawing under a new\r\n"+
	"    name. All drawing filenames are text files and use the extension '.dre'.\r\n"+
	"\r\n"+
	"10. Miscellaneous\r\n"+
	"  10.1 Press <F5> to refresh the display at any time.\r\n"+
	"  10.2 Press <F1> to open a help dialog.\r\n"+
	"  10.3 Press <F7> to open an 'About' Dialog.\r\n"+
	"  10.4 When in standalone (non-applet) mode, use also the <Refresh> option\r\n"+
	"    from the <Window> menu, the <Help> option and the <About> option\r\n"+
	"    from the <Help> menu.\r\n"+
	"  10.5 Use the <Clear> button in the right button panel, to open a new blank\r\n"+
	"    design. When in standalone (non-applet) mode, use also the <New>\r\n"+
	"    option from the <File> menu or press <Control>-n or <Control>-N.\r\n"+
	"  10.6 Use the <Example> button in the right button panel to open a new\r\n"+
	"    exemplar design.\r\n"+
	"\r\n";

public static void main(String args[])
{
	DrawingEditor ed = new DrawingEditor();
	DrawingFrame centralFrame = new DrawingFrame(ed);

	ed.parent = centralFrame;
	ed.init();

	centralFrame.setSize(SizeX,SizeY+100);
	centralFrame.setResizable(false);
	centralFrame.setTitle("Mini Drawing Editor");

	centralFrame.setVisible(true);

//	Parsing argument string for a design file name
	if (args.length > 0)
		centralFrame.tryToOpen(args[0]);
}

public Container getContentPane()
{
	if (parent != null)
		// Standalone mode
		return parent.getContentPane();
	else
		// Applet mode (applet != null)
		return applet.getContentPane();
}

public void init()
{
	initializeGUI();

	// Initialize variables
	target = new Drawing(this);
}

void initializeGUI()
{
//	Shape panel
	shapePanel = new JPanel();
	shapePanel.setLayout(new FlowLayout());
	shapePanel.setBackground(Color.gray);
	circle = new JButton("Circle");
	circle.addActionListener(this);
	line = new JButton("Line");
	line.addActionListener(this);
	polygon = new JButton("Polygon");
	polygon.addActionListener(this);
	shapePanel.add(circle);
	shapePanel.add(line);
	shapePanel.add(polygon);

//	Toolbar
	toolbar = new JPanel();
	toolbar.setLayout(new GridLayout(0,1));
	toolbar.setBackground(Color.gray);
	deleteButton = new JButton("Delete");
	deleteButton.addActionListener(this);
	clearButton = new JButton("Clear");
	clearButton.addActionListener(this);
	exampleButton = new JButton("Example");
	exampleButton.addActionListener(this);
	editButton = new JButton("Edit");
	editButton.addActionListener(this);
	resizeButton = new JButton("Resize");
	resizeButton.addActionListener(this);
	rotateButton = new JButton("Rotate");
	rotateButton.addActionListener(this);
	alignButton = new JButton("Align");
	alignButton.addActionListener(this);
	distributeButton = new JButton("Distribute");
	distributeButton.addActionListener(this);
	fitButton = new JButton("Fit");
	fitButton.addActionListener(this);
	equalizeButton = new JButton("Equalize");
	equalizeButton.addActionListener(this);
	toolbar.add(deleteButton);
	toolbar.add(clearButton);
	toolbar.add(exampleButton);
	toolbar.add(editButton);
	toolbar.add(rotateButton);
	toolbar.add(resizeButton);
	toolbar.add(alignButton);
	toolbar.add(distributeButton);
	toolbar.add(equalizeButton);
	toolbar.add(fitButton);

//	Initialize drawing view
	drawingView = new DrawingView(this,IMAGE_VIEW_DIMENSION);

//	Initialize layout
	getContentPane().setLayout(new BorderLayout());
	getContentPane().add("North",shapePanel);
	getContentPane().add("East",toolbar);
	getContentPane().add("West",drawingView);
}

public void repaint()
{
	super.repaint();
	if (drawingView != null)
		drawingView.repaint();
}

public void actionPerformed(ActionEvent evt)
{
  if (evt.getSource() instanceof JButton)
  {
	JButton tgt = (JButton)evt.getSource();

	if (tgt == line) input_flag = LINE;
	else if (tgt == circle) input_flag = CIRCLE;
	else if (tgt == polygon) input_flag = POLYGON;
	else if (tgt == clearButton)
	{
		clear();
		repaint();
	}
	else if (tgt == editButton)
	{
		if (selections.size() == 1)
			((Shape)selections.elementAt(0)).edit(this);
		else if (selections.size() > 1)
	MessageDialog.open("Warning : Cannot edit multiple selections at once!");
	}
	else if (tgt == deleteButton)
	{
		deleteSelections();
		repaint();
	}
	else if (tgt == exampleButton)
	{
		loadExample();
		repaint();
	}
	drawingView.requestFocus();
  }
}

public void keyTyped(KeyEvent evt) {}
public void keyReleased(KeyEvent evt) {}
public void keyPressed(KeyEvent evt)
{
	char ch = evt.getKeyChar();
	int key = evt.getKeyCode();

	boolean
		del=false, edit=false,
		shift=false, control=false, alt=false,
		single_selection, has_selections;
	Shape comp;

	shift = evt.isShiftDown();
	control = evt.isControlDown();
	alt = (evt.isMetaDown() || evt.isAltDown());

	boolean noModifiers = (!shift) && (!control) && (!alt);
	boolean onlyShift = (shift) && (!control) && (!alt);
	boolean onlyControl = (!shift) && (control) && (!alt);
	boolean onlyAlt = (!shift) && (!control) && (alt);
	boolean altControl = (!shift) && (control) && (alt);
	boolean shiftControl = (shift) && (control) && (!alt);
	boolean shiftAlt = (shift) && (!control) && (alt);
	boolean shiftAltControl = (shift) && (control) && (alt);

	single_selection = (selections.size() == 1);
	has_selections = (selections.size() > 0);

	if ( (key == KeyEvent.VK_BACK_SPACE) || (key == KeyEvent.VK_DELETE) ) del=true;
	else if (key == KeyEvent.VK_ENTER) edit=true;

	if ((ch=='a') || (ch=='A')) ch='a';
	else if ((ch=='c') || (ch=='C')) ch='c';
	else if ((ch=='v') || (ch=='V')) ch='v';
	else if ((ch=='x') || (ch=='X')) ch='x';
	else if ((ch=='y') || (ch=='Y')) ch='y';
	else if ((ch=='z') || (ch=='Z')) ch='z';
	else ch=' ';

	if (del && noModifiers)
	{
		deleteSelections();
		repaint();
		return;
	}

	if (edit && noModifiers)
	{
		if (single_selection)
		{
			((Shape)selections.elementAt(0)).edit(this);
			return;
		}
		else if (selections.size() > 1)
	MessageDialog.open("Warning : Cannot edit multiple selections at once!");
		return;
	}

	// Help and about box
	if ((key == KeyEvent.VK_F1) && noModifiers)
	{
		help();
		return;
	}
	else if ((key == KeyEvent.VK_F7) && noModifiers)
	{
		about();
		return;
	}

	// Refresh
	if ((key == KeyEvent.VK_F5) && noModifiers)
	{
		repaint();
		return;
	}

	// Move with the arrows
	else if (key == KeyEvent.VK_LEFT)
	{
		if (!has_selections) return;
		if (noModifiers)
		{	// Move by 1
			if (!single_selection) record(null,END_BLOCK);
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				record(comp,MOVE,comp.getPosition().x,comp.getPosition().y);
				comp.moveBy(-1,0);
			}
			if (!single_selection) record(null,START_BLOCK);
			if (selections.size() > 0) set_redo(MOVE_BY,-1,0);
			repaint();
		}
		else if (onlyAlt)
		{	// Move by 10
			if (!single_selection) record(null,END_BLOCK);
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				record(comp,MOVE,comp.getPosition().x,comp.getPosition().y);
				comp.moveBy(-10,0);
			}
			if (!single_selection) record(null,START_BLOCK);
			if (selections.size() > 0) set_redo(MOVE_BY,-10,0);
			repaint();
		}
	}

	else if (key == KeyEvent.VK_RIGHT)
	{
		if (!has_selections) return;
		if (noModifiers)
		{	// Move by 1
			if (!single_selection) record(null,END_BLOCK);
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				record(comp,MOVE,comp.getPosition().x,comp.getPosition().y);
				comp.moveBy(1,0);
			}
			if (!single_selection) record(null,START_BLOCK);
			if (selections.size() > 0) set_redo(MOVE_BY,1,0);
			repaint();
		}
		else if (onlyAlt)
		{	// Move by 10
			if (!single_selection) record(null,END_BLOCK);
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				record(comp,MOVE,comp.getPosition().x,comp.getPosition().y);
				comp.moveBy(10,0);
			}
			if (!single_selection) record(null,START_BLOCK);
			if (selections.size() > 0) set_redo(MOVE_BY,10,0);
			repaint();
		}
	}

	else if (key == KeyEvent.VK_UP)
	{
		if (!has_selections) return;
		if (noModifiers)
		{	// Move by 1
			if (!single_selection) record(null,END_BLOCK);
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				record(comp,MOVE,comp.getPosition().x,comp.getPosition().y);
				comp.moveBy(0,-1);
			}
			if (!single_selection) record(null,START_BLOCK);
			if (selections.size() > 0) set_redo(MOVE_BY,0,-1);
			repaint();
		}
		else if (onlyAlt)
		{	// Move by 10
			if (!single_selection) record(null,END_BLOCK);
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				record(comp,MOVE,comp.getPosition().x,comp.getPosition().y);
				comp.moveBy(0,-10);
			}
			if (!single_selection) record(null,START_BLOCK);
			if (selections.size() > 0) set_redo(MOVE_BY,0,-10);
			repaint();
		}
	}

	else if (key == KeyEvent.VK_DOWN)
	{
		if (!has_selections) return;
		if (noModifiers)
		{	// Move by 1
			if (!single_selection) record(null,END_BLOCK);
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				record(comp,MOVE,comp.getPosition().x,comp.getPosition().y);
				comp.moveBy(0,1);
			}
			if (!single_selection) record(null,START_BLOCK);
			if (selections.size() > 0) set_redo(MOVE_BY,0,1);
			repaint();
		}
		else if (onlyAlt)
		{	// Move by 10
			if (!single_selection) record(null,END_BLOCK);
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				record(comp,MOVE,comp.getPosition().x,comp.getPosition().y);
				comp.moveBy(0,10);
			}
			if (!single_selection) record(null,START_BLOCK);
			if (selections.size() > 0) set_redo(MOVE_BY,0,10);
			repaint();
		}
	}

	else if ((ch == 'a') && (onlyControl || onlyAlt)) selectAll();
	else if ((ch == 'c') && (onlyControl || onlyAlt))
	{
		System.out.println("Control-C pressed.");
		if (input_flag == POLYGON)
		{
			if (positions.size() > 0)
			{
				comp = new Shape(POLYGON,positions);
				target.add(comp);
				record(comp,DELETE);
				reset_input();
				clearSelections();
				select(comp);
			}
			repaint();
			return;
		}
		// Copy selections
		copy();
	}
	else if ((ch == 'v') && (onlyControl || onlyAlt))
	{	// Paste selections
		paste();
	}
	else if ((ch == 'x') && (onlyControl || onlyAlt))
	{	// Cut selections
		cut();
		if (selections.size() > 0) set_redo(DELETE);
	}
	else if ((ch == 'y') && (onlyControl || onlyAlt))
	{
		redo();
		repaint();
	}
	else if (ch == 'z')
	{
		if (onlyControl || onlyAlt)
		{
			undo();
			repaint();
		}
	}
}

public void mouseEntered(MouseEvent evt) {}
public void mouseExited(MouseEvent evt) {}
public void mousePressed(MouseEvent evt)
{
	int x = evt.getX(), y = evt.getY();
	Shape comp;

	boolean
		shift = evt.isShiftDown(),
		control = evt.isControlDown(),
		alt = evt.isMetaDown() || evt.isAltDown();

	drawingView.requestFocus();

	if (control || alt) return;
	
	// Input of a shape
	if (!target.includes(x,y)) {reset_input(); repaint(); return;}
	if (input_flag == LINE)
	{
			positions.addElement((Object)(new Point(x,y)));
			curr_x = x; curr_y = y;
			if (positions.size() == 2)
			{
				comp = new Shape(LINE,(Point)positions.elementAt(0),(Point)positions.elementAt(1));
				target.add(comp);
				record(comp,DELETE);
				reset_input();
				clearSelections();
				select(comp);
			}
			repaint();
			return;
	}
	else if (input_flag == CIRCLE)
	{
			positions.addElement((Object)(new Point(x,y)));
			curr_x = x; curr_y = y;
			if (positions.size() == 2)
			{
				comp = new Shape(CIRCLE,(Point)positions.elementAt(0),
							(Point)positions.elementAt(1));
				target.add(comp);
				record(comp,DELETE);
				reset_input();
				clearSelections();
				select(comp);
			}
			repaint();
			return;
	}
	else if (input_flag == POLYGON)
	{
			positions.addElement((Object)(new Point(x,y)));
			curr_x = x; curr_y = y;
			if ((positions.size() == Shape.SizeLimit) || (evt.getClickCount() == 2))
			{
				comp = new Shape(POLYGON,positions);
				target.add(comp);
				record(comp,DELETE);
				reset_input();
				clearSelections();
				select(comp);
			}
			repaint();
			return;
	}
	else
		// Selection with dragging and edition of shapes
	{
		posX = x;
		posY = y;
		int i = target.shapes.size()-1;
		while (i >= 0)
		{
			comp = (Shape)target.shapes.elementAt(i);
			if (comp.includes(x,y))
			{
				if (!shift) clearSelections();
				select(comp);
	
				// Start dragging selections
				setShapeMouseDrag();
				prev_x = x;
				prev_y = y;
				for (int ind=0;ind<selections.size();ind++)
					((Shape)selections.elementAt(ind)).startDragging();
				posX = x;
				posY = y;
				// If double-clicked, edit the selection
				if (evt.getClickCount() == 2)
				{
					resetShapeMouseDrag();
					if (selections.size() == 1)
					{
						((Shape)selections.elementAt(0)).edit(this);
						repaint();
						return;
					}
					else if (selections.size() > 1)
					{
	MessageDialog.open("Warning : Cannot edit multiple selections at once!");
							return;
					}
				}
				repaint();
				return;
			}
			i--;
		}
		clearSelections();
		setRubberbandMouseDrag();
		rubberbandStart.x = rubberbandEnd.x = x;
		rubberbandStart.y = rubberbandEnd.y = y;
		repaint();
	}
}

public void mouseReleased(MouseEvent evt)
{
	int x = evt.getX(), y = evt.getY();
	Shape comp;

	drawingView.requestFocus();

	if ((!shapeMouseDrag) && (!rubberbandMouseDrag)) return;
	if (shapeMouseDrag)
	{
		resetShapeMouseDrag();
		if (!target.includes(x,y)) return;
		Shape sel = (Shape)selections.elementAt(0);
		if (sel.hasBeenDragged())
		{
			if (selections.size()>1) record(null,END_BLOCK);
			set_redo(-1);
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				record(comp,MOVE,comp.getPosition().x,comp.getPosition().y);
				comp.dragBy(x-prev_x,y-prev_y);
				comp.confirmDragged(target);
			}
			if (selections.size()>1) record(null,START_BLOCK);
		}
		prev_x = -1; prev_y = -1;
	}
	else 		// rubberbandMouseDrag
	{
		clearSelections();
		for (int i=0;i<target.shapes.size();i++)
		{
			comp = (Shape)target.shapes.elementAt(i);
			if (inRubberband(comp.topleft())
					&& inRubberband(comp.bottomright()))
				select(comp);
		}
		resetRubberbandMouseDrag();
		rubberbandStart.x = rubberbandStart.y
			= rubberbandEnd.x = rubberbandEnd.y = -1;
	}
	repaint();
}

public void mouseClicked(MouseEvent evt) {}

public void mouseMoved(MouseEvent evt)
{
	int x = evt.getX(), y = evt.getY();
	drawingView.requestFocus();
	if (input_flag != 0)
		if (positions.size() > 0) {curr_x = x; curr_y = y; repaint();}
}

public void mouseDragged(MouseEvent evt)
{
	int x = evt.getX(), y = evt.getY();
	Shape comp;

	drawingView.requestFocus();

	if ((!shapeMouseDrag) && (!rubberbandMouseDrag)) return;
	if (shapeMouseDrag)
	{
		for (int i=0;i<selections.size();i++)
		{
			comp = (Shape)selections.elementAt(i);
			comp.dragBy(x-prev_x,y-prev_y);
		}
		prev_x = x; prev_y = y;
	}
	else		// rubberbanding
	{
		rubberbandEnd.x = x;
		rubberbandEnd.y = y;
	}
	repaint();
}

public void paint(Graphics g)
{
	Color pen = g.getColor();
	super.paint(g);
	g.setColor(BackgroundColor);
	g.fillRect(0,0,SizeX,SizeY);
	g.setColor(pen);
	target.draw(g);
	drawPositions(g);

	if (shapeMouseDrag)
		for (int i=0;i<selections.size();i++)
			((Shape)selections.elementAt(i)).
				drawDragged(g);

	if (rubberbandMouseDrag)
		drawRubberband(g,rubberbandStart,rubberbandEnd);
}

public void drawRubberband(Graphics g, Point start, Point end)
{
	Color pen = g.getColor();
	int	originx = Math.min(start.x, end.x),
		originy = Math.min(start.y, end.y),
		width = Math.abs(start.x - end.x),
		height = Math.abs(start.y - end.y),
		endx = Math.max(start.x, end.x),
		endy = Math.max(start.y, end.y);

	g.setColor(Color.blue);

//	Draw a dashed rectangle
	int i, j, step = 5;
	boolean draw = true;

	for (i=originx;i<=(endx-step);i+=step)
	{
		if (draw) g.drawLine(i, originy, i+step, originy);
		draw = (!draw);
	}
	draw = true;
	for (j=originy;j<=(endy-step);j+=step)
	{
		if (draw) g.drawLine(endx, j, endx, j+step);
		draw = (!draw);
	}
	draw = true;
	for (i=endx;i>=(originx+step);i-=step)
	{
		if (draw) g.drawLine(i, endy, i-step, endy);
		draw = (!draw);
	}
	draw = true;
	for (j=endy;j>=(originy+step);j-=step)
	{
		if (draw) g.drawLine(originx, j, originx, j-step);
		draw = (!draw);
	}

	g.setColor(pen);
}

public boolean inRubberband(Point po)
{
	int	leftmost = Math.min(rubberbandStart.x, rubberbandEnd.x),
		topmost = Math.min(rubberbandStart.y, rubberbandEnd.y),
		rightmost = Math.max(rubberbandStart.x, rubberbandEnd.x),
		bottommost = Math.max(rubberbandStart.y, rubberbandEnd.y);

	return ((po.x >= leftmost) && (po.x <= rightmost)
			&& (po.y >= topmost) && (po.y <= bottommost));
}

public void clearSelections()
{
	for (int i=0;i<selections.size();i++)
	{
		Shape comp = (Shape)selections.elementAt(i);
		comp.deselect();
	}
	selections.removeAllElements();
}

void reset_input()
{
	positions.removeAllElements();
	input_flag = 0;
	curr_x = -1;
	curr_y = -1;
}

void message(String mesg)
{
	System.out.println(mesg);
}

private void setShapeMouseDrag() {shapeMouseDrag = true;}
private void resetShapeMouseDrag() {shapeMouseDrag = false;}
private void setRubberbandMouseDrag() {rubberbandMouseDrag = true;}
private void resetRubberbandMouseDrag() {rubberbandMouseDrag = false;}

public void clear()
{
	clearSelections();
	clearOperations();
	target.clear();
	reset_input();
}

public void loadExample()
{
	target.initExample();
	record(null,CLEAR_ALL);
}

void selectAll()
{
	clearSelections();
	for (int i=0;i<target.shapes.size();i++)
		select((Shape)target.shapes.elementAt(i));
	repaint();
}

void select(Shape s)
{
	if (s.isSelected()) return;
	s.select();
	selections.addElement(s);
}

void deselect(Shape s)
{
	s.deselect();
	selections.removeElement(s);
}

void copy()
{
	clipboard.removeAllElements();
	for (int i=0;i<selections.size();i++)
	{
		Shape comp = ((Shape)selections.elementAt(i)).copy();
		comp.deselect();
		clipboard.addElement(comp);
	}
}

void cut()
{
	clipboard.removeAllElements();
	if (selections.size() != 1) record(null,END_BLOCK);
	for (int i=0;i<selections.size();i++)
	{
		Shape comp = (Shape)selections.elementAt(i);
		clipboard.addElement(comp);
		target.remove(comp);
		record(comp,ADD,comp.getPosition().x,comp.getPosition().y);
	}
	if (selections.size() != 1) record(null,START_BLOCK);
	clearSelections();
	repaint();
}

void paste()
{
	if ((posX < 0) || (posY < 0)) return;
	if (selections.size() != 1) record(null,END_BLOCK);
	for (int i=0;i<clipboard.size();i++)
	{
		Shape comp = (Shape)clipboard.elementAt(i);
		comp.setPosition(posX,posY);
		comp.deselect();
		comp = comp.copy();
		target.add(comp);
		record(comp,DELETE);
	}
	if (selections.size() != 1) record(null,START_BLOCK);
	repaint();
}

public void clearOperations()
{
	for (int i=0;i<operations.size();i++)
	{
		UndoOperation oper = (UndoOperation)operations.elementAt(i);
		oper.target = null;
	}
	operations.removeAllElements();
	set_redo(-1);
}

public void deleteSelections()
{
	if (selections.size() > 1) record(null,END_BLOCK);
	for (int i=0;i<selections.size();i++)
	{
		Shape comp = (Shape)selections.elementAt(i);
		target.remove(comp);
		record(comp,ADD,comp.getPosition().x,comp.getPosition().y);
	}
	if (selections.size() > 1) record(null,START_BLOCK);
	if (selections.size() > 0) set_redo(DELETE);
	clearSelections();
}

public void help()
{
	TextAreaDialog.open("Help",helpText);
}

public void about()
{
	AboutDialog.open(this,
		"About Drawing Editor",
		"Drawing Editor", "Editor for Geometric Shapes (Lines, Circles and Polygons)",
		"Elpida Tzafestas, 1997", "brensham@softlab.ece.ntua.gr",
		"http://www.softlab.ece.ntua.gr/~brensham");
}

public void undo()
{
	int index;
	UndoOperation current;

	if (operations.size() == 0) return;
	clearSelections();
	index = operations.size()-1;
	current = (UndoOperation)operations.elementAt(index);
	if (current.operator == CLEAR_ALL) clear();
	else if (current.operator == START_BLOCK)
	{
		operations.removeElement(current);
		index--;
		current = (UndoOperation)operations.elementAt(index);
		while (current.operator != END_BLOCK)
		{
			current.target.undo(current.operator,
				current.operands[0],current.operands[1],
				current.operands[2],this);
			operations.removeElement(current);
			index--;
			current = (UndoOperation)operations.elementAt(index);
		}
		// Now, current.operator == END_BLOCK, so remove the UndoOperation.
		operations.removeElement(current);
	}
	else
	{
		current.target.undo(current.operator,
			current.operands[0],current.operands[1],current.operands[2],this);
		operations.removeElement(current);
	}
}

public void record(UndoEntity target, int operator)
{
	operations.addElement(new UndoOperation(target,operator));
}

public void record(UndoEntity target, int operator, int op1)
{
	operations.addElement(new UndoOperation(target,operator,op1));
}

public void record(UndoEntity target, int operator, int op1, int op2)
{
	operations.addElement(new UndoOperation(target,operator,op1,op2));
}

public void record(UndoEntity target, int operator, int op1, int op2, int op3)
{
	operations.addElement(new UndoOperation(target,operator,op1,op2,op3));
}

public void redo()
{
	if (selections.size() == 0) return;
	switch (redo_operator)
	{
		case SET_FILL_AND_BORDER_COLOR:
			record(null,END_BLOCK);
			for (int i=0;i<selections.size();i++)
			{
				Shape comp = (Shape)selections.elementAt(i);
				Color f = comp.getFillColor(), b = comp.getBorderColor();
				comp.recordFillColor(this);
				comp.recordBorderColor(this);
				comp.setFillColor(new Color(redo_operands[0],
								redo_operands[1],redo_operands[2]));
				comp.setBorderColor(new Color(redo_operands[3],
								redo_operands[4],redo_operands[5]));
			}
			record(null,START_BLOCK);
			break;
		case -1 :
			break;
		default:
			if (selections.size() > 1) record(null,END_BLOCK);
			for (int i=0;i<selections.size();i++)
				((Shape)selections.elementAt(i)).redo(redo_operator,
											redo_operands,this);
			if (selections.size() > 1) record(null,START_BLOCK);
	}
}

public void set_redo(int operator)
{
	redo_operator = operator;
	for (int i=0;i<redo_operands_size;i++) redo_operands[i]=0;
}

public void set_redo(int operator, int op1)
{
	redo_operator = operator;
	for (int i=0;i<redo_operands_size;i++) redo_operands[i]=0;
	redo_operands[0]=op1;
}

public void set_redo(int operator, int op1, int op2)
{
	redo_operator = operator;
	for (int i=0;i<redo_operands_size;i++) redo_operands[i]=0;
	redo_operands[0]=op1;
	redo_operands[1]=op2;
}

public void set_redo(int operator, int op1, int op2, int op3)
{
	redo_operator = operator;
	for (int i=0;i<redo_operands_size;i++) redo_operands[i]=0;
	redo_operands[0]=op1;
	redo_operands[1]=op2;
	redo_operands[2]=op3;
}

public void set_redo(int operator, int op1, int op2, int op3,
						int op4, int op5, int op6)
{
	redo_operator = operator;
	for (int i=0;i<redo_operands_size;i++) redo_operands[i]=0;
	redo_operands[0]=op1;
	redo_operands[1]=op2;
	redo_operands[2]=op3;
	redo_operands[3]=op4;
	redo_operands[4]=op5;
	redo_operands[5]=op6;
}

void drawPositions(Graphics g)
{
//	This method tracks user input in red : all given points are drawn
//	and each one of them is connected to the next one with a straight
//	line. This arrangement is particularly useful for polygon input,
//	where one can see the intermediate state of the polygon as it
//	evolves. Circle and line input is traced in a similar manner.
//	The final mouse pointer is also tracked as moved around.

	Point prev, curr;
	Color pen = g.getColor();
	g.setColor(InputColor);

	if ((input_flag == POLYGON) || (input_flag == LINE))
	{
		if (positions.size() > 0)
		{
			prev = (Point)positions.elementAt(0);
			drawPoint(g,prev);
			for (int i=1;i<positions.size();i++)
			{
				curr = (Point)positions.elementAt(i);
				g.drawLine(prev.x,prev.y,curr.x,curr.y);
				prev=curr;
			}
			g.drawLine(prev.x,prev.y,curr_x,curr_y);
		}
	}
	else if (input_flag == CIRCLE)
	{
		if (positions.size() > 0) // It will be necessarily 1.
		{
			prev = (Point)positions.elementAt(0);
			drawPoint(g,prev);

			// Draw the circle from (prev.x,prev.y) to (curr_x,curr_y)
			int diameter = (int)Math.sqrt(((prev.x-curr_x)*(prev.x-curr_x))
					+ ((prev.y-curr_y)*(prev.y-curr_y)));
			Point up = new Point((prev.x+curr_x-diameter)/2,
						(prev.y+curr_y-diameter)/2);
			g.drawOval(up.x,up.y,diameter,diameter);
		}
	}

	g.setColor(pen);
}

void drawPoint(Graphics g, int x, int y)
	{g.drawLine(x,y,x,y);}

void drawPoint(Graphics g, Point po)
	{g.drawLine(po.x,po.y,po.x,po.y);}

}	// End DrawingEditor

class Drawing implements Constants
{
Vector shapes;
Dimension size;

DrawingEditor parent;

static Dimension DefaultSize=new Dimension(450,450);
static Dimension MaximumSize=new Dimension(1000,1000);

//	Light yellow background
Color backgroundColor=Colors.LightYellow;

private void init(Dimension dim, DrawingEditor editor)
{
	size = new Dimension(dim);
	parent = editor;

	shapes = new Vector(10,5);
}

Drawing(DrawingEditor editor)
{
	init(DefaultSize, editor);
}

Drawing(Dimension dim, DrawingEditor editor)
{
	init(dim, editor);
}

public Color getBackgroundColor() {return backgroundColor;}

public boolean includes(int x, int y)
{
	if ((x>0) && (x<size.width) && (y>0) && (y<size.height)) return true;
	return false;
}

private void setSize(Dimension dim)
{
	size = new Dimension(dim);
}

public void add(Shape comp)
{
	shapes.addElement(comp);
}

public void clear()
{
	shapes.removeAllElements();
}

public void remove(Shape comp)
{
	shapes.removeElement(comp);
}

public void draw(Graphics g)
{
	Color c = g.getColor();

	g.setColor(backgroundColor);
	g.fillRect(0,0,size.width,size.height);
	g.setColor(c);

	for (int i=0;i<shapes.size();i++)
		((Shape)(shapes.elementAt(i))).draw(g);
}

//////////////////////////////////////////////////////////////////
//			Example : Happy cartoon
//////////////////////////////////////////////////////////////////

public void initExample()
{
	Shape temp;

	int xpos1[] = {87,337,274,150,147};
	int ypos1[] = {225,228,344,344,343};
	shapes.addElement(new Shape(POLYGON,xpos1,ypos1,5,Color.black));
	int xpos2[] = {128,302,216,214};
	int ypos2[] = {225,227,337,337};
	shapes.addElement(new Shape(POLYGON,xpos2,ypos2,4,Color.white));
	int xpos3[] = {104,119,173,247,295,328,328,215,105,104};
	int ypos3[] = {109,176,231,232,188,109,66,35,66,66};
	shapes.addElement(new Shape(POLYGON,xpos3,ypos3,10,Colors.LightPink));

	temp = new Shape(CIRCLE,new Point(123,76),new Point(141,81));
	temp.setFillColor(Color.yellow);
	shapes.addElement(temp);
	
	temp = new Shape(CIRCLE,new Point(178,80),new Point(150,110));
	temp.setFillColor(Color.yellow);
	shapes.addElement(temp);

	temp = new Shape(LINE,new Point(183,93),new Point(243,93));
	temp.setFillColor(Color.blue);
	shapes.addElement(temp);
	
	temp = new Shape(CIRCLE,new Point(279,80),new Point(251,110));
	temp.setFillColor(Color.yellow);
	shapes.addElement(temp);
	
	temp = new Shape(CIRCLE,new Point(287,76),new Point(305,81));
	temp.setFillColor(Color.yellow);
	shapes.addElement(temp);
	
	temp = new Shape(CIRCLE,new Point(224,122),new Point(202,145));
	temp.setFillColor(Color.yellow);
	shapes.addElement(temp);

	int xpos4[] = {158,214,266,263,215,162,163};
	int ypos4[] = {158,186,157,173,212,173,173};
	shapes.addElement(new Shape(POLYGON,xpos4,ypos4,7,Color.yellow));

	temp = new Shape(CIRCLE,new Point(251,100),new Point(267,103));
	temp.setFillColor(Color.blue);
	shapes.addElement(temp);

	temp = new Shape(CIRCLE,new Point(150,99),new Point(166,102));
	temp.setFillColor(Color.blue);
	shapes.addElement(temp);
	
	int xpos5[] = {168,169,260,261,259};
	int ypos5[] = {233,287,233,286,286};
	shapes.addElement(new Shape(POLYGON,xpos5,ypos5,5,Color.red));

	temp = new Shape(CIRCLE,new Point(212,249),new Point(225,267));
	temp.setFillColor(Color.red);
	shapes.addElement(temp);
}

public String toString()
{
	String str="";

	str += (size.width+" "+size.height+" "+
			backgroundColor.getRed()+" "+
			backgroundColor.getGreen()+" "+
			backgroundColor.getBlue()+"\n");
	for (int i=0;i<shapes.size();i++)
		str+=(((Shape)shapes.elementAt(i)).toString()+"\n");
	return str;
}

public void parse(String str)
{
	String token=null;
	int r=255,g=255,b=255;

	StringTokenizer tokenizer = new StringTokenizer(str);

	token = nextToken(tokenizer);
	if (token == null) return;
	size.width=parseWidth(token);

	token = nextToken(tokenizer);
	if (token == null) return;
	size.height=parseHeight(token);

	token = nextToken(tokenizer);
	if (token == null) return;
	r=parseRGB(token);

	token = nextToken(tokenizer);
	if (token == null) return;
	g=parseRGB(token);

	token = nextToken(tokenizer);
	if (token == null) return;
	b=parseRGB(token);

	backgroundColor = new Color(r,g,b);

	while (tokenizer.hasMoreTokens())
	{
		token = nextToken(tokenizer);
		int typ = Shape.parseShapeType(token);
		if (Shape.validType(typ))
		{
			Shape s = new Shape();
			s.setType(typ);
			s.parseWithoutType(tokenizer);
			if (validate(s)) shapes.addElement(s);
		}
	}
}

private boolean validate(Shape s)
{
	if (s == null) return false;
	return (s.size >= 2);
}

private String nextToken(StringTokenizer tokenizer)
{
	if (tokenizer.hasMoreTokens()) return tokenizer.nextToken();
	return null;
}

private int parseWidth(String str)
{
	int n = Integer.parseInt(str);
	if ((n>0) && (n<=MaximumSize.width)) return n;
	return DefaultSize.width;
}

private int parseHeight(String str)
{
	int n = Integer.parseInt(str);
	if ((n>0) && (n<=MaximumSize.height)) return n;
	return DefaultSize.height;
}

private int parseRGB(String str)
{
	int n = Integer.parseInt(str);
	if ((n>=0) && (n<=255)) return n;
	return 255;
}

}	// End Drawing

class Shape implements Constants, HasColors, UndoEntity
{
protected boolean selected = false;
Point dragPosition = new Point(0,0);

Color fillColor = Color.yellow;
Color borderColor = Color.black;
static int SizeLimit = 100;
int type;
int size = 0;
private int xpoints[] = new int[SizeLimit];
private int ypoints[] = new int[SizeLimit];

Shape()
{
	type = LINE;
}

Shape(int typ, Point start, Point end)
{
	// Line or circle
	switch (typ)
	{
		case LINE:
		default:
			type = LINE;
			xpoints[0]=start.x; xpoints[1]=end.x;
			ypoints[0]=start.y; ypoints[1]=end.y;
			size=2;
			break;
		case CIRCLE:
			type = CIRCLE;
			int cx = (start.x+end.x)/2;
			int cy = (start.y+end.y)/2;
			int r = (int)(Math.sqrt(((start.x-end.x)*(start.x-end.x))
					+ ((start.y-end.y)*(start.y-end.y))) / 2);
			xpoints[0]= cx - r; xpoints[1]= cx + r;
			ypoints[0]= cy - r; ypoints[1]= cy + r;
			size=2;
			break;
	}
}

Shape(int typ, Vector points)
{
	// Polygon
	Point temp;
	size = Math.min(points.size(),SizeLimit);

	type = POLYGON;
	for (int i=0;i<size;i++)
	{
		temp = (Point)points.elementAt(i);
		xpoints[i]=temp.x;
		ypoints[i]=temp.y;
	}
}

Shape(int typ, int xpos[], int ypos[], int s)
{
	// Practically used only for polygons
	type = typ;
	size = Math.min(Math.min(Math.min(xpos.length,ypos.length), SizeLimit), s);
	for (int i=0;i<size;i++)
		xpoints[i]=xpos[i];
	for (int j=0;j<size;j++)
		ypoints[j]=ypos[j];
}

Shape(int typ, int xpos[], int ypos[], int s, Color c)
{
	// Practically used only for polygons or during copying
	type = typ;
	size = Math.min(Math.min(Math.min(xpos.length,ypos.length), SizeLimit), s);
	for (int i=0;i<size;i++)
		xpoints[i]=xpos[i];
	for (int j=0;j<size;j++)
		ypoints[j]=ypos[j];
	fillColor = c;
}

Shape(int typ, int xpos[], int ypos[], int s, Color f, Color b)
{
	// Practically used only for polygons or during copying
	type = typ;
	size = Math.min(Math.min(Math.min(xpos.length,ypos.length), SizeLimit), s);
	for (int i=0;i<size;i++)
		xpoints[i]=xpos[i];
	for (int j=0;j<size;j++)
		ypoints[j]=ypos[j];
	fillColor = f;
	borderColor = b;
}

public void setType(int typ)
{
	type = typ;
}

public void edit(DrawingEditor ed)
{
	ColorEditor.open("Edit fill and border colors",this,ed);
}

public void select() {selected = true;}
public void deselect() {selected = false;}
public boolean isSelected() {return selected;}

public boolean includes1(int x, int y)
{
	return rectangleIncludes(getPosition(), width(), height(), x, y);
}

public boolean rectangleIncludes(Point upperleft, int width, int height, int x, int y)
{
	if ((x >= upperleft.x) && (x <= (upperleft.x+width))
			&& (y >= upperleft.y) && (y <= (upperleft.y+height)))
		return true;
	return false;
}

public boolean includes(int x, int y)
{
	switch (type)
	{
		case LINE: return lineIncludes(x,y);
		case CIRCLE: return circleIncludes(x,y);
		case POLYGON: return polygonIncludes(x,y);
	}
	return false;
}

public void moveBy(Point po) { moveBy(po.x,po.y); }

public Color getFillColor() {return fillColor;}
public void setFillColor(Color c) {fillColor = c;}
public Color getBorderColor() {return borderColor;}
public void setBorderColor(Color c) {borderColor = c;}

public void recordFillColor(DrawingEditor editor)
{
	Color c = getFillColor();
	editor.record(this,Constants.SET_FILL_COLOR,
		c.getRed(),c.getGreen(),c.getBlue());
}

public void recordBorderColor(DrawingEditor editor)
{
	Color c = getBorderColor();
	editor.record(this,Constants.SET_BORDER_COLOR,
		c.getRed(),c.getGreen(),c.getBlue());
}

public Point topleft() { return getPosition(); }

Shape copy()
{
	return new Shape(type, xpoints, ypoints, size, fillColor, borderColor);
}

public String name()
{
	return nameFor(type);
}

static String nameFor(int typ)
{
	switch (typ)
	{
		case LINE: return "Line";
		case CIRCLE: return "Circle";
		case POLYGON: return "Polygon";
		default: return "Simple Shape";
	}
}

public void draw(Graphics g)
{
	drawUnselected(g);
	if (selected) drawSelected(g);
}

protected void drawSelected1(Graphics g)
{
	Color pen = g.getColor();
	g.setColor(getBorderColor());
	drawSelectedRectangle(g,getPosition(),width(),height());
	g.setColor(pen);
}

protected void drawSelectedRectangle(Graphics g, Point upperleft, int width, int height)
{
	if (selected)
	{
		drawSelectedPoint(g, upperleft.x, upperleft.y);
		drawSelectedPoint(g, upperleft.x, upperleft.y + (height/2));
		drawSelectedPoint(g, upperleft.x, upperleft.y + height);

		drawSelectedPoint(g, upperleft.x + (width / 2), upperleft.y);
		drawSelectedPoint(g, upperleft.x + (width / 2),
			upperleft.y + height);
		drawSelectedPoint(g, upperleft.x + width, upperleft.y);
		drawSelectedPoint(g, upperleft.x + width,
			upperleft.y + (height/2));
		drawSelectedPoint(g, upperleft.x + width,
			upperleft.y + height);
	}
}

protected void drawSelectedPoint(Graphics g, int x, int y)
{
	g.fillRect(x-2, y-2, 5, 5);
}

void drawDragged(Graphics g)
{ drawDraggedAt(g,dragPosition.x,dragPosition.y); }

public int width()
{
	int leftmost=10000, rightmost=-1, i;

	switch (type)
	{
		case CIRCLE: 
				leftmost = xpoints[0];
				rightmost = xpoints[1];
				break;
		case LINE:
		case POLYGON: 
		default:
				// Find rightmost and leftmost
				for (i=0;i<size;i++)
				{
					if (xpoints[i]>rightmost) rightmost=xpoints[i];
					if (xpoints[i]<leftmost) leftmost=xpoints[i];
				}
				break;
	}

	return rightmost-leftmost;
}

public int height()
{
	int bottommost=-1, topmost=10000, i;

	switch (type)
	{
		case CIRCLE: 
				topmost = ypoints[0];
				bottommost = ypoints[1];
				break;
		case LINE:
		case POLYGON: 
		default:
				// Find bottommost and topmost
				for (i=0;i<size;i++)
				{
					if (ypoints[i]<topmost) topmost=ypoints[i];
					if (ypoints[i]>bottommost) bottommost=ypoints[i];
				}
				break;
	}

	return bottommost-topmost;
}

public Point getPosition()
{
	int leftmost=10000, topmost=10000, i, j;
	switch (type)
	{
		case CIRCLE: 
				leftmost = xpoints[0];
				topmost = ypoints[0];
				break;
		case LINE:
		case POLYGON: 
		default:
				//	Find leftmost and topmost
				for (i=0;i<size;i++)
					if (xpoints[i]<leftmost) leftmost=xpoints[i];
				for (j=0;j<size;j++)
					if (ypoints[j]<topmost) topmost=ypoints[j];
				break;
	}
	return new Point(leftmost,topmost);
}

public Point bottomright()
{
	int rightmost=-1, bottommost=-1, i, j;
	switch (type)
	{
		case LINE:
		case POLYGON: 
		default:
			//	Find rightmost and bottommost
				for (i=0;i<size;i++)
					if (xpoints[i]>rightmost) rightmost=xpoints[i];
				for (j=0;j<size;j++)
					if (ypoints[j]>bottommost) bottommost=ypoints[j];
				break;
		case CIRCLE: 
				rightmost = xpoints[1];
				bottommost = ypoints[1];
				break;
	}
	return new Point(rightmost,bottommost);
}

public void setPosition(int x, int y)
{
	int leftmost=10000, topmost=10000, i, j;

//	Find leftmost and topmost
	Point po = topleft();

	// Reset topmost and leftmost (reset all points relative to them)
	for (i=0;i<size;i++)
		xpoints[i] += (x-po.x);
	for (j=0;j<size;j++)
		ypoints[j] += (y-po.y);
}

public void setPosition(Point po)
	{setPosition(po.x,po.y);}

void drawUnselected(Graphics g)
{
//	Point up;
	int diameter;
	Color pen = g.getColor();

	switch (type)
	{
		case LINE:
			g.setColor(borderColor);
			g.drawLine(xpoints[0],ypoints[0],
					xpoints[1],ypoints[1]);
			g.setColor(pen);
			break;
		case CIRCLE:
			diameter = xpoints[1] - xpoints[0];
			g.setColor(fillColor);
			g.fillOval(xpoints[0],ypoints[0],diameter,diameter);
			g.setColor(borderColor);
			g.drawOval(xpoints[0],ypoints[0],diameter,diameter);
			g.setColor(pen);
			break;
		case POLYGON:
			g.setColor(fillColor);
			g.fillPolygon(xpoints,ypoints,size);
			g.setColor(borderColor);
			g.drawPolygon(xpoints,ypoints,size);
			g.setColor(pen);
			break;
	}
}

void drawDraggedAt(Graphics g, int x, int y)
{
	Point up, offset, pos = getPosition();
	int diameter;
	Color pen = g.getColor();

	offset = new Point(x - getPosition().x, y - getPosition().y);

	switch (type)
	{
		case LINE:
			g.setColor(borderColor);
			g.drawLine(xpoints[0]+offset.x,ypoints[0]+offset.y,
					xpoints[1]+offset.x,ypoints[1]+offset.y);
			g.setColor(pen);
			break;
		case CIRCLE:
			g.setColor(borderColor);
			g.drawOval(xpoints[0]+offset.x,ypoints[0]+offset.y,
					xpoints[1]-xpoints[0],ypoints[1]-ypoints[0]);
			g.setColor(pen);
			break;
		case POLYGON:
			g.setColor(borderColor);
			int newxpoints[] = new int[SizeLimit];
			int newypoints[] = new int[SizeLimit];
			for (int i=0;i<size;i++)
			{
				newxpoints[i]=xpoints[i]+offset.x;
				newypoints[i]=ypoints[i]+offset.y;
			}
			g.drawPolygon(newxpoints,newypoints,size);
			g.setColor(pen);
			break;
	}
}

protected void drawSelected(Graphics g)
{
	Color pen = g.getColor();
	g.setColor(borderColor);

	switch (type)
	{
		case LINE: lineDrawSelected(g); break;
		case CIRCLE: circleDrawSelected(g); break;
		case POLYGON: polygonDrawSelected(g); break;
	}

	g.setColor(pen);
}

void lineDrawSelected(Graphics g)
{
	drawSelectedPoint(g, xpoints[0], ypoints[0]);
	drawSelectedPoint(g, (xpoints[0] + xpoints[1]) / 2, (ypoints[0] + ypoints[1])/ 2);
	drawSelectedPoint(g, xpoints[1], ypoints[1]);
}

void circleDrawSelected(Graphics g)
{
	//drawSelectedRectangle(g,new Point(xpoints[0],ypoints[0]),xpoints[1]-xpoints[0],ypoints[1]-ypoints[0]);
	Point tl = topleft();
	Point br = bottomright();
	drawSelectedRectangle(g,tl,br.x-tl.x,br.y-tl.y);
}

void polygonDrawSelected(Graphics g)
{
	Point upperleft = new Point(10000,10000);
	Point bottomright = new Point(0,0);
	int width, height;
	for (int i=0;i<size;i++)
	{
		if (xpoints[i] < upperleft.x) upperleft.x = xpoints[i];
		if (xpoints[i] > bottomright.x) bottomright.x = xpoints[i];
		if (ypoints[i] < upperleft.y) upperleft.y = ypoints[i];
		if (ypoints[i] > bottomright.y) bottomright.y = ypoints[i];
	}
	width = bottomright.x - upperleft.x;
	height = bottomright.y - upperleft.y;

	drawSelectedRectangle(g,upperleft,width,height);
}

public void undo(int operator, int op1, int op2, int op3, DrawingEditor editor)
{
	editor.select(this);
	switch (operator)
	{
		case ADD:
			this.setPosition(op1,op2);
			editor.target.add(this);
			break;
		case DELETE:
			editor.target.remove(this);
			break;
		case MOVE:
			setPosition(op1,op2);
			break;
		case SET_FILL_COLOR:
			setFillColor(new Color(op1,op2,op3));
			break;
		case SET_BORDER_COLOR:
			setBorderColor(new Color(op1,op2,op3));
			break;
	}
}

public void redo(int operator, int[] operands, DrawingEditor editor)
{
	switch (operator)
	{
		case MOVE_BY:
			editor.record(this,MOVE,getPosition().x,getPosition().y);
			moveBy(operands[0],operands[1]);
			break;
		case DELETE:
			editor.record(this,ADD,getPosition().x,getPosition().y);
			editor.target.remove(this);
			break;
	}
}

protected String nextToken(StringTokenizer tokenizer)
{
	if (tokenizer.hasMoreTokens()) return tokenizer.nextToken();
	return null;
}

public static int parsePosition(String str)
{
	int n = Integer.parseInt(str);
	if (n>0) return n;
	return 1;
}

public static int parseRGB(String str)
{
	int n = Integer.parseInt(str);
	if ((n>=0) && (n<=255)) return n;
	return 255;
}

public boolean lineIncludes(int x,int y)
{
	int lowX;
	int lowY;
	int upX;
	int upY;

	if (xpoints[0] <= xpoints[1])
	{
		lowX = xpoints[0]-1;
		upX = xpoints[1]+1;
	}
	else
	{
		lowX = xpoints[1]-1;
		upX = xpoints[0]+1;
	}
	if (ypoints[0] <= ypoints[1])
	{
		lowY = ypoints[0]-1;
		upY = ypoints[1]+1;
	}
	else
	{
		lowY = ypoints[1]-1;
		upY = ypoints[0]+1;
	}
	if ((x >= lowX) && (x <= upX) && (y >= lowY) & (y <= upY))
		return true;
	return false;
}

public boolean circleIncludes(int x, int y)
{
	return rectangleIncludes(new Point(xpoints[0],ypoints[0]),
					xpoints[1]-xpoints[0],ypoints[1]-ypoints[0],x,y);
}

public boolean polygonIncludes(int x, int y)
{
	Point upperleft = new Point(10000,10000);
	Point bottomright = new Point(0,0);
	int width, height;
	for (int i=0;i<size;i++)
	{
		if (xpoints[i] < upperleft.x) upperleft.x = xpoints[i];
		if (xpoints[i] > bottomright.x) bottomright.x = xpoints[i];
		if (ypoints[i] < upperleft.y) upperleft.y = ypoints[i];
		if (ypoints[i] > bottomright.y) bottomright.y = ypoints[i];
	}
	width = bottomright.x - upperleft.x;
	height = bottomright.y - upperleft.y;

	return rectangleIncludes(upperleft, width, height, x, y);
}

public void moveBy(int x, int y)
{
	int i;
	for (i=0;i<xpoints.length;i++)
		{xpoints[i]+=x;}
	for (i=0;i<ypoints.length;i++)
		{ypoints[i]+=y;}
}

public void dragBy(int x, int y) {dragPosition.x +=x; dragPosition.y +=y;}
public void startDragging()
		{dragPosition.x = getPosition().x; dragPosition.y = getPosition().y;}
public boolean hasBeenDragged()
{
	Point pos = getPosition();
	return ((dragPosition.x != pos.x) || (dragPosition.y != pos.y));
}

public void confirmDragged(Drawing target)
{
	if (target.includes(dragPosition.x,dragPosition.y) &&
		target.includes(dragPosition.x+width(),
						dragPosition.y+height()))
	setPosition(dragPosition.x,dragPosition.y);
}

public void parse(StringTokenizer tokenizer)
{
	String token;
	int r=255,g=255,b=255;

	// Parse type, first.
	token = nextToken(tokenizer);
	if (token == null) return;
	type=parseType(token);

	// Then parse size, fill color and border color from tokenizer
	parseWithoutType(tokenizer);
}

public void parseWithoutType(StringTokenizer tokenizer)
{
	String token;
	int r=255,g=255,b=255;

	// Parse size, fill color and border color from tokenizer
	token = nextToken(tokenizer);
	if (token == null) return;
	size=parseSize(token);

	token = nextToken(tokenizer);
	if (token == null) return;
	r=parseRGB(token);

	token = nextToken(tokenizer);
	if (token == null) return;
	g=parseRGB(token);

	token = nextToken(tokenizer);
	if (token == null) return;
	b=parseRGB(token);

	fillColor = new Color(r,g,b);

	token = nextToken(tokenizer);
	if (token == null) return;
	r=parseRGB(token);

	token = nextToken(tokenizer);
	if (token == null) return;
	g=parseRGB(token);

	token = nextToken(tokenizer);
	if (token == null) return;
	b=parseRGB(token);

	borderColor = new Color(r,g,b);

	// Parse xpoints and ypoints from tokenizer
	for (int i=0;i<size;i++)
	{
		int x, y;
		token = nextToken(tokenizer);
		if (token == null) {size = 0; return;}
		x = parsePosition(token);

		token = nextToken(tokenizer);
		if (token == null) {size = 0; return;}
		y = parsePosition(token);

		if (i < SizeLimit)
		{
			xpoints[i] = x;
			ypoints[i] = y;
		}
	}
}

protected static boolean validType(int typ)
{
	switch (typ)
	{
		case LINE:
		case CIRCLE:
		case POLYGON:
			return true;
		default:
			return false;
	}
}

protected int parseType(String str)
{
	int n = Integer.parseInt(str);
	switch (n)
	{
		case LINE:
		case CIRCLE:
		case POLYGON:
			return n;
		default:
			return CIRCLE;
	}
}

public static int parseShapeType(String str)
{
	int n = Integer.parseInt(str);
//	System.out.println(""+nameFor(n));
	return n;
}

protected int parseSize(String str)
{
	int n = Integer.parseInt(str);
	if ((n>0) && (n<=SizeLimit)) return n;
	return 2;
}

public String toString()
{
	String str = new String(type + " " + size);
	str += (" "+fillColor.getRed()+" "+
		fillColor.getGreen()+" "+
		fillColor.getBlue());
	str += (" "+borderColor.getRed()+" "+
		borderColor.getGreen()+" "+
		borderColor.getBlue());
	for (int i=0;i<size;i++)
		str += (" "+xpoints[i]+" "+ypoints[i]);
	return str;
}

}	// End Shape

class DrawingView extends JPanel
{
DrawingEditor gui;
private Dimension size;

DrawingView(DrawingEditor ed, Dimension dim)
{
	gui = ed;
	size = new Dimension(dim);

	addKeyListener(gui);
	addMouseListener(gui);
	addMouseMotionListener(gui);
}

public Dimension getPreferredSize()
	{return new Dimension(size.width,size.height);}

public void paint(Graphics g)
{
	if (g == null) return;
	super.paint(g);

	// Draw the editor view
	gui.paint(g);
}

}	// End DrawingView

class DrawingFrame extends JFrame implements ActionListener,
				KeyListener, MouseListener, MouseMotionListener
{
DrawingEditor target;

JMenuItem
	fileNew, fileOpen, fileSave, fileSaveAs, fileExit,
	editSelectAll, editCut, editCopy, editPaste,
	windowRefresh, helpHelp, helpAbout;
private String currentFile = null;
private String illegal_filename_message =
	"Illegal filename. Only *.DRE names are allowed.";

static String postfix = ".dre";

DrawingFrame(DrawingEditor app)
{
	super();
	target = app;

	initBar();
}

private void initBar()
{
	JMenuBar bar = new JMenuBar();

	JMenu fileMenu = new JMenu("File");
	fileNew = new JMenuItem("New",KeyEvent.VK_N);
	fileNew.addActionListener(this);
	fileOpen = new JMenuItem("Open",KeyEvent.VK_O);
	fileOpen.addActionListener(this);
	fileSave = new JMenuItem("Save",KeyEvent.VK_S);
	fileSave.addActionListener(this);
	fileSaveAs = new JMenuItem("Save As...");
	fileSaveAs.addActionListener(this);
	fileExit = new JMenuItem("Exit");
	fileExit.addActionListener(this);
	fileMenu.add(fileNew);
	fileMenu.add(fileOpen);
	fileMenu.addSeparator();
	fileMenu.add(fileSave);
	fileMenu.add(fileSaveAs);
	fileMenu.addSeparator();
	fileMenu.add(fileExit);

	JMenu editMenu = new JMenu("Edit");
	editSelectAll = new JMenuItem("Select All",KeyEvent.VK_A);
	editSelectAll.addActionListener(this);
	editCut = new JMenuItem("Cut",KeyEvent.VK_X);
	editCut.addActionListener(this);
	editCopy = new JMenuItem("Copy",KeyEvent.VK_C);
	editCopy.addActionListener(this);
	editPaste = new JMenuItem("Paste",KeyEvent.VK_V);
	editPaste.addActionListener(this);
	editMenu.add(editSelectAll);
	editMenu.addSeparator();
	editMenu.add(editCut);
	editMenu.add(editCopy);
	editMenu.add(editPaste);

	JMenu windowMenu = new JMenu("Window");
	windowRefresh = new JMenuItem("Refresh -- F5");
	windowRefresh.addActionListener(this);
	windowMenu.add(windowRefresh);

	JMenu helpMenu = new JMenu("Help");
	helpHelp = new JMenuItem("Help -- F1");
	helpHelp.addActionListener(this);
	helpAbout = new JMenuItem("About -- F7");
	helpAbout.addActionListener(this);
	helpMenu.add(helpHelp);
	helpMenu.addSeparator();
	helpMenu.add(helpAbout);

	bar.add(fileMenu);
	bar.add(editMenu);
	bar.add(windowMenu);
	bar.add(helpMenu);
	setJMenuBar(bar);

	addKeyListener(this);
	addMouseListener(this);
	addMouseMotionListener(this);
	addWindowListener(new WindowAdapter()
	{
		public void windowClosing(WindowEvent evt)
			{dispose();System.exit(1);}
	});
	requestFocus();
}

public void keyPressed(KeyEvent evt) {target.keyPressed(evt); requestFocus();}
public void keyTyped(KeyEvent evt) {target.keyTyped(evt); requestFocus();}
public void keyReleased(KeyEvent evt) {target.keyReleased(evt); requestFocus();}

public void mousePressed(MouseEvent evt) {target.mousePressed(evt); requestFocus();}
public void mouseClicked(MouseEvent evt) {target.mouseClicked(evt); requestFocus();}
public void mouseReleased(MouseEvent evt) {target.mouseReleased(evt); requestFocus();}
public void mouseEntered(MouseEvent evt) {target.mouseEntered(evt); requestFocus();}
public void mouseExited(MouseEvent evt) {target.mouseExited(evt); requestFocus();}
public void mouseMoved(MouseEvent evt) {target.mouseMoved(evt); requestFocus();}
public void mouseDragged(MouseEvent evt) {target.mouseDragged(evt); requestFocus();}

public void actionPerformed(ActionEvent evt)
{
  if (evt.getSource() instanceof JMenuItem)
  {
	JMenuItem tgt = (JMenuItem)evt.getSource();

	if (tgt == fileNew)
	{
	//	If a file is open, ask save.
		if (currentFile != null)
			ConfirmSaveDialog.open("Save "+currentFile.toUpperCase()+" first ?",
							this,ConfirmSaveDialog.NEW);
	//	New file
		else doNewFile();
	}
	else if (tgt == fileOpen)
	{
	//	If a file is open, ask save.
		if (currentFile != null)
			ConfirmSaveDialog.open("Save "+currentFile.toUpperCase()+" first ?",
							this,ConfirmSaveDialog.OPEN);
	//	Open file
		else doOpenFile();
	}
	else if (tgt == fileSave)
	{
	//	Save file
		if (currentFile == null)
			MessageDialog.open("Warning : You have not opened any files!"
						+" Use <Save As>.");
		else
			ConfirmSaveDialog.open("Save "+currentFile.toUpperCase()+" ?",
							this,0);
	}
	else if (tgt == fileSaveAs)
	//	Save file under a different name
		saveFileAs();
	else if (tgt == fileExit) {dispose();System.exit(1);}
	else if (tgt == editSelectAll) target.selectAll();
	else if (tgt == editCut) target.cut();
	else if (tgt == editCopy) target.copy();
	else if (tgt == editPaste) target.paste();
	else if (tgt == windowRefresh) target.repaint();
	else if (tgt == helpHelp) target.help();
	else if (tgt == helpAbout) target.about();
  }
  requestFocus();
}

public void doNewFile()
{
	currentFile = null;
	target.clear();
	target.message("New design");
	target.repaint();
}

public boolean accept(File dir, String name)
{
	return name.endsWith(postfix);
}

public void doOpenFile()
{
	String filename;
	FileDialog dlg = new FileDialog(this,"Load Drawing from File",FileDialog.LOAD);
	//dlg.setFilenameFilter(this);
	dlg.setVisible(true);
	filename = dlg.getFile();
	if (filename == null) {target.repaint(); return;}
	filename=checkOpen(filename);
	if (filename != null)
	{
		target.clear();
		target.message("Loading from "+filename.toUpperCase()+". Please wait ...");
		if (read(filename))
		{
			currentFile = filename;
			target.message("Loading from "+filename.toUpperCase()+" successful.");
		}
		else target.message("Loading from "+filename.toUpperCase()+" failed.");
	}
	else target.message(illegal_filename_message);
	target.repaint();
}

public void saveOpenedFile()
{
	target.message("Saving to "+currentFile.toUpperCase()+". Please wait ...");
	if (write(currentFile))
	{
		target.message("Saving to "+currentFile.toUpperCase()+" successful.");
	}
	else target.message("Saving to "+currentFile.toUpperCase()+" failed.");
}

private void saveFileAs()
{
	String filename;
	FileDialog dlg = new FileDialog(this,"Save Drawing to File",FileDialog.SAVE);
	//dlg.setFilenameFilter(this);
	dlg.setVisible(true);
	filename = dlg.getFile();
	if (filename != null)
	{
		filename=checkSave(filename);
		if (filename == null)
		{
			target.message(illegal_filename_message);
			return;
		}
		target.message("Saving to "+filename.toUpperCase()+". Please wait ...");
		if (write(filename))
		{
			currentFile = filename;
			target.message("Saving to "+filename.toUpperCase()+" successful.");
		}
		else target.message("Saving to "+filename.toUpperCase()+" failed.");
	}
}

private boolean read(String file)
{
	String str="";
	Drawing newDrawing = new Drawing(target);

	try
	{
		FileInputStream f = new FileInputStream(file);
		int c;
		while ((c=f.read()) != -1)
		{
			str += String.valueOf((char)c);
		}
		newDrawing.parse(str);
	}
	catch (IOException e)
		{return false;}
	target.target = newDrawing;
	target.repaint();
	return true;
}

private boolean write(String file)
{
	String str = target.target.toString();
	try
	{
		FileOutputStream f = new FileOutputStream(file);
		for (int i=0;i<str.length();i++)
			f.write(str.charAt(i));
	}
	catch (IOException e)
		{return false;}
	return true;
}

private String checkOpen(String f)
{
	if (f.endsWith(postfix))
		return f;
	return null;
}

private String checkSave(String f)
{
	int index;

	if (f.endsWith(postfix))
		return f;
	if ( (index=f.indexOf((int)'.'))== -1) return (f+postfix);
	if (index == 0) return ("Drawing"+postfix);
	return (f.substring(0,index)+postfix);
}

public void tryToOpen(String filename)
{
	String theFile = null;
	int index;

	if (filename.endsWith(postfix)) theFile=filename;
	else
		if ( (index=filename.indexOf((int)'.'))== -1) theFile=filename+postfix;

	if (theFile != null)
	{
		if ((new File(theFile)).exists())
		{
			target.clear();
			target.message("Loading from "+theFile.toUpperCase()+". Please wait ...");
			if (read(theFile))
			{
				currentFile = theFile;
				target.message("Loading from "+theFile.toUpperCase()+" successful.");
			}
			else target.message("Loading from "+theFile.toUpperCase()+" failed.");
			target.repaint();
		}
	}
	else target.message("Illegal filename : "+filename.toUpperCase()+". Load failed.");
}

}	// End DrawingFrame


////////////////////////////////////////////////////
//
//			Various dialogs
//
////////////////////////////////////////////////////

class MessageDialog extends JDialog implements ActionListener
{
  protected static JFrame localFrame;
  protected JButton okButton;
  protected JLabel message;

  MessageDialog(JFrame parent, String mesg)
  {
	super(parent,true);

	okButton = new JButton("OK");
	okButton.addActionListener(this);
	JPanel okPanel = new JPanel();
	okPanel.setLayout(new FlowLayout(FlowLayout.CENTER));
	okPanel.add(okButton);

	message = new JLabel(mesg);
	JPanel messagePanel = new JPanel();
	messagePanel.setLayout(new FlowLayout(FlowLayout.CENTER));
	messagePanel.add(message);
	getContentPane().setLayout(new BorderLayout());
	getContentPane().add("North",messagePanel);
	getContentPane().add("South",okPanel);

	addWindowListener(new WindowAdapter()
	{
		public void windowClosing(WindowEvent e)
		{
			setVisible(false);
			if (localFrame != null)
				localFrame.setVisible(false);
		}
	});
  }

  public void actionPerformed(ActionEvent evt)
  {
	if ((JButton)evt.getSource() == okButton)
	{
		setVisible(false);
		if (localFrame != null)
			localFrame.setVisible(false);
	}
  }

  // Shortcut to create a frame automatically, the frame is a static
  // variable, so that all message dialogs can use the same frame
  public static void open(String mesg)
  {
	// If the frame has not been created yet, create it.
	if (localFrame == null)
		localFrame = new JFrame("");
	// Create the dialog now
	MessageDialog dlg = new MessageDialog(localFrame,mesg);
	// Shrink the frame to just fit the dialog
	localFrame.setSize(dlg.getSize().width,dlg.getSize().height);
	// Show the dialog
	dlg.pack();
	dlg.setVisible(true);
  }

}	// End MessageDialog

class TextAreaDialog extends JDialog implements ActionListener
{
protected static JFrame localFrame;
protected JButton okButton;
protected JTextArea message;

TextAreaDialog(JFrame parent, String title, String text)
{
	super(parent,title,true);

	okButton = new JButton("OK");
	okButton.addActionListener(this);
	JPanel okPanel = new JPanel();
	okPanel.setLayout(new FlowLayout(FlowLayout.CENTER));
	okPanel.add(okButton);

	message = new JTextArea(text,20,70);
	message.setEditable(false);
	JScrollPane messagePanel = new JScrollPane();
	messagePanel.setPreferredSize(new Dimension(550,550));
	messagePanel.setViewportView(message);

	getContentPane().setLayout(new BorderLayout());
	getContentPane().add("North",messagePanel);
	getContentPane().add("South",okPanel);

	addWindowListener(new WindowAdapter()
	{
		public void windowClosing(WindowEvent e)
		{
			setVisible(false);
			if (localFrame != null)
				localFrame.setVisible(false);
		}
	});
}

public void actionPerformed(ActionEvent evt)
{
	if ((JButton)evt.getSource() == okButton)
	{
		setVisible(false);
		if (localFrame != null)
			localFrame.setVisible(false);
	}
}

// Shortcut to create a frame automatically, the frame is a static
// variable, so that all message dialogs can use the same frame

public static void open(String title, String text)
{
	// If the frame has not been created yet, create it.
	if (localFrame == null)
		localFrame = new JFrame("");
	// Create the dialog now
	TextAreaDialog dlg = new TextAreaDialog(localFrame,title,text);

	// Shrink the frame to just fit the dialog
	localFrame.setSize(190,130);

	// Show the dialog
	dlg.pack();
	dlg.setVisible(true);
}

}	// End TextAreaDialog

class AboutDialog extends JDialog implements ActionListener
{
protected static JFrame localFrame;
protected JLabel title1, title2, author, mailto, link;
protected JButton okButton = new JButton("OK");

static Color backColor = Color.cyan;

private AboutDialogUse target;

AboutDialog(JFrame parent, AboutDialogUse app, String title,
	String tit1, String tit2, String auth, String mto, String http)
{
	super(parent,title,true);

	target = app;

	title1 = new JLabel(tit1,JLabel.CENTER);
	title1.setFont(new Font("Dialog",Font.BOLD,18));
	title2 = new JLabel(tit2,JLabel.CENTER);
	title2.setFont(new Font("Dialog",Font.BOLD,16));
	author = new JLabel(auth,JLabel.CENTER);
	author.setFont(new Font("Courier",Font.PLAIN,14));
	mailto = new JLabel(mto,JLabel.CENTER);
	mailto.setFont(new Font("Courier",Font.ITALIC,12));
	link = new JLabel(http,JLabel.CENTER);
	link.setFont(new Font("Courier",Font.ITALIC,12));

//	Component foreground colors
	title1.setForeground(Color.blue);
	title2.setForeground(Color.black);
	author.setForeground(Color.magenta);
	mailto.setForeground(Color.gray);
	link.setForeground(Color.gray);
	okButton.setForeground(Color.black);
	okButton.addActionListener(this);

	initLayout();

	addWindowListener(new WindowAdapter()
	{
		public void windowClosing(WindowEvent e)
		{
			setVisible(false);
			if (localFrame != null)
				localFrame.setVisible(false);
		}
	});
}

private void initLayout()
{
	JPanel componentPanel = new JPanel();
	componentPanel.setLayout(new GridLayout(0,1));
	componentPanel.setBackground(backColor);

	componentPanel.add(title1);
	componentPanel.add(title2);
	componentPanel.add(author);
	componentPanel.add(mailto);
	componentPanel.add(link);

	JPanel buttonPanel = new JPanel();
	buttonPanel.setLayout(new FlowLayout());
	buttonPanel.setBackground(backColor);
	buttonPanel.add(okButton);
	getContentPane().setLayout(new BorderLayout());
	setBackground(backColor);
	setForeground(backColor);
	getContentPane().add("North",componentPanel);
	getContentPane().add("Center",buttonPanel);

	setSize(500,190);
	setResizable(false);
}

public void actionPerformed(ActionEvent evt)
{
	if ((JButton)evt.getSource() == okButton)
	{
		setVisible(false);
		if (localFrame != null)
			localFrame.setVisible(false);
	}
}

// Shortcut to create a frame automatically, the frame is a static
// variable, so that all message dialogs can use the same frame

public static void open(AboutDialogUse app, String title,
		String tit1, String tit2, String auth, String mto, String address)
{
	// If the frame has not been created yet, create it.
	if (localFrame == null)
		localFrame = new JFrame("");
	// Create the dialog now
	AboutDialog dlg = new AboutDialog(localFrame,app,title,
						tit1,tit2,auth,mto,address);

	dlg.setVisible(true);
}

}	// End AboutDialog

class ConfirmSaveDialog extends JDialog implements ActionListener
{
  protected JButton yesButton;
  protected JButton noButton;
  protected JLabel message;
  DrawingFrame target;
  private int nextAction = 0;

//	ConfirmSaveDialog action constants
final static int NEW	= 1001;
final static int OPEN	= 1002;

  ConfirmSaveDialog(DrawingFrame parent, String mesg, int next)
  {
	super(parent,true);

	target = parent;
	nextAction = next;

	yesButton = new JButton("Yes");
	yesButton.addActionListener(this);
	noButton = new JButton("No");
	noButton.addActionListener(this);
	JPanel buttonPanel = new JPanel();
	buttonPanel.setLayout(new FlowLayout(FlowLayout.CENTER));
	buttonPanel.add(yesButton);
	buttonPanel.add(noButton);

	message = new JLabel(mesg);
	JPanel messagePanel = new JPanel();
	messagePanel.setLayout(new FlowLayout(FlowLayout.CENTER));
	messagePanel.add(message);

	getContentPane().setLayout(new BorderLayout());
	getContentPane().add("North",messagePanel);
	getContentPane().add("South",buttonPanel);

	addWindowListener(new WindowAdapter()
	{
		public void windowClosing(WindowEvent e)
		{
			setVisible(false);
		}
	});
  }

  public void actionPerformed(ActionEvent evt)
  {
	if ((JButton)evt.getSource() == yesButton)
	{
		target.saveOpenedFile();
		doNext();
		setVisible(false);
	}
	else if ((JButton)evt.getSource() == noButton)
	{
		doNext();
		setVisible(false);
	}
  }

  private void doNext()
  {
	switch (nextAction)
	{
		case NEW: target.doNewFile(); break;
		case OPEN: target.doOpenFile(); break;
		default: break;
	}
  }

  // Shortcut to create a frame automatically, the frame is a static
  // variable, so that all message dialogs can use the same frame

  public static void open(String mesg, DrawingFrame f, int next)
  {
	// Create the dialog now
	ConfirmSaveDialog dlg = new ConfirmSaveDialog(f,mesg, next);

	// Shrink the frame to just fit the dialog
	dlg.setSize(dlg.getSize().width,dlg.getSize().height);

	// Show the dialog
	dlg.pack();
	dlg.setVisible(true);
  }

}	// End ConfirmSaveDialog

class ColorEditor extends JDialog implements ActionListener
{
protected static JFrame localFrame;
protected JButton okButton,cancelButton;
protected List fillColors;
protected List borderColors;
protected HasColors target;
private DrawingEditor editor;

static Color allColors[] =
 		{Color.black,Color.blue,Colors.Brown,Color.cyan,Color.green,
		Colors.LightGreen,Colors.LightMagenta,Colors.LightPink,
		Colors.LightYellow,Color.magenta,Color.orange,Color.pink,Color.red,
		Color.yellow,Color.white};
static String colorNames[] = {"Black","Blue","Brown","Cyan","Green",
		"Light green","Light magenta","Light pink",
		"Light yellow","Magenta","Orange","Pink","Red",
		"Yellow","White"};

ColorEditor(JFrame parent, String title, HasColors arg, DrawingEditor ed)
{
	super(parent,title,true);

	target = arg;
	editor = ed;

	okButton = new JButton("OK");
	okButton.addActionListener(this);
	cancelButton = new JButton("Cancel");
	cancelButton.addActionListener(this);
	JPanel buttonPanel = new JPanel();
	buttonPanel.setLayout(new FlowLayout(FlowLayout.CENTER));
	buttonPanel.add(okButton);
	buttonPanel.add(cancelButton);

	JPanel labelsPanel = new JPanel();
	labelsPanel.setLayout(new GridLayout(1,0));
	JPanel panel1 = new JPanel();
	panel1.setLayout(new FlowLayout());
	panel1.add(new JLabel("Fill color"));
	JPanel panel2 = new JPanel();
	panel2.setLayout(new FlowLayout());
	panel2.add(new JLabel("Border color"));
	labelsPanel.add(panel1);
	labelsPanel.add(panel2);

	JPanel colorsPanel = new JPanel();
	colorsPanel.setLayout(new GridLayout(1,0));
	colorsPanel.add(fillColors = new List(5,false));
	colorsPanel.add(borderColors = new List(5,false));

	initColors();

	getContentPane().setLayout(new BorderLayout());
	getContentPane().add("North",labelsPanel);
	getContentPane().add("Center",colorsPanel);
	getContentPane().add("South",buttonPanel);

	addWindowListener(new WindowAdapter()
	{
		public void windowClosing(WindowEvent e)
		{
			setVisible(false);
			if (localFrame != null)
				localFrame.setVisible(false);
		}
	});
}

private void initColors()
{
	Color current;
	int i;

	for (i=0;i<colorNames.length;i++)
	{
		fillColors.add(colorNames[i]);
		borderColors.add(colorNames[i]);
	}
	current = target.getFillColor();
	for (i=0;i<allColors.length;i++)
		if (current.equals(allColors[i]))
			{fillColors.select(i); break;}
	current = target.getBorderColor();
	for (i=0;i<allColors.length;i++)
		if (current.equals(allColors[i]))
			{borderColors.select(i); return;}
}

public void actionPerformed(ActionEvent evt)
{
	if (evt.getSource() instanceof JButton)
	{
		JButton tgt = (JButton)evt.getSource();
		if (tgt == okButton)
		{
			update();
			setVisible(false);
			if (localFrame != null)
				localFrame.setVisible(false);
		}
		else if (tgt == cancelButton)
		{
			setVisible(false);
			if (localFrame != null)
				localFrame.setVisible(false);
		}
	}
}

public void update()
{
 	Color f = allColors[fillColors.getSelectedIndex()],
		b = allColors[borderColors.getSelectedIndex()];

	editor.record(null,Constants.END_BLOCK);
	target.recordFillColor(editor);
	target.recordBorderColor(editor);
	editor.record(null,Constants.START_BLOCK);

	target.setFillColor(f);
	target.setBorderColor(b);

	editor.set_redo(Constants.SET_FILL_AND_BORDER_COLOR,
					f.getRed(),f.getGreen(),f.getBlue(),
					b.getRed(),b.getGreen(),b.getBlue());
	editor.repaint();
}

// Shortcut to create a frame automatically, the frame is a static
// variable, so that all message dialogs can use the same frame

public static void open(String title, HasColors arg, DrawingEditor ed)
{
	// If the frame has not been created yet, create it.
	if (localFrame == null)
		localFrame = new JFrame("");
	// Create the dialog now
	ColorEditor dlg = new ColorEditor(localFrame,title,arg,ed);

	// Shrink the frame to just fit the dialog
	localFrame.setSize(190,130);

	// Show the dialog
	dlg.pack();
	dlg.setVisible(true);
}

}	// End ColorEditor

class UndoOperation
{
	public UndoEntity target;
	public int operator;
	public int operands[] = new int[3];

	UndoOperation(UndoEntity elem, int op)
	{
		target = elem;
		operator=op;
	}

	UndoOperation(UndoEntity elem, int op, int op1)
	{
		target = elem;
		operator=op;
		operands[0]=op1;
	}

	UndoOperation(UndoEntity elem, int op, int op1, int op2)
	{
		target = elem;
		operator=op;
		operands[0]=op1;
		operands[1]=op2;
	}

	UndoOperation(UndoEntity elem, int op, int op1, int op2, int op3)
	{
		target = elem;
		operator=op;
		operands[0]=op1;
		operands[1]=op2;
		operands[2]=op3;
	}
}

interface UndoEntity
{
public void undo(int operand, int op1, int op2, int op3, DrawingEditor editor);
}

interface AboutDialogUse {}

interface Colors
{
final Color LightYellow = new Color(255,255,200);
final Color LightPink = new Color(255,220,220);
final Color LightGreen = new Color(200,255,200);
final Color LightMagenta = new Color(255,200,255);
final Color Brown = new Color(185,120,100);
}

interface HasColors
{
public Color getFillColor();
public Color getBorderColor();
public void setFillColor(Color c);
public void setBorderColor(Color c);
public void recordFillColor(DrawingEditor ed);
public void recordBorderColor(DrawingEditor ed);
public void edit(DrawingEditor ed);
}

interface Constants
{
// Constants used for input
static int LINE = 1;
static int CIRCLE = 2;
static int POLYGON = 3;

//	Undo constants
int ADD = 600;
int DELETE = 601;
int MOVE = 602;
int EDIT = 605;
int START_BLOCK = 606;
int END_BLOCK = 607;
int CLEAR_ALL = 608;
int SET_FILL_COLOR = 610;
int SET_BORDER_COLOR = 611;

//	Redo constants (some undo constants are also used)
int MOVE_BY = 650;
int SET_FILL_AND_BORDER_COLOR = 662;

}	// End Constants


