////////////////////////////////////////////////////////////////////////////////////
/*
*	DrawingEditor.java -
*			Sample editor program
*	* Author			Elpida Tzafestas
*	* Original release date		December 1997
*	* Latest release date		December 2005
*
*	Features:
*	1	Input of graphical elements with the buttons and mouse
*		* Input two end points for line
*		* Input two end points of a diameter for circle
*		* Input vertices for polygon, until 30 points are given
*			or control-c/C is pressed or double click is detected.
*	2	Code works both as an applet and as an application
*	3	Selection of a shape with click, multiple shapes with
*		shift-click.
*		Select all with control-a/A (alternatively use alt-a/A)
*		Select with the aid of a rubberband drawn around shapes by dragging
*	4	Delete selected shapes with the <DELETE> or <BACKSPACE> key
*	5	Clipboard : control-x/X, control-c/C and control-v/V, for
*		cut, copy and paste of shapes, respectively.
*		Alternatively use alt-x/X, alt-c/C and alt-v/V
*	6	Mouse dragging of a shape or a specific vertex
*	7	Move selected shapes with arrows
*		(move by 1 pixel with plain arrow, by 10 pixels with alt+arrow).
*	8	Resize selected shapes with shift+arrows
*		(increase or decrease by 1 pixel with shift+arrow and shift+control+arrow,
*		 by 10 pixels with alt+shift+arrow and alt+shift+control+arrow).
*		Resize freely with control-r/R on a selection.
*		Alternatively use alt-r/R
*	9	Rotate selected shapes with control+arrows
*	10	Edit shapes fill and border colors by double-clicking.
*	11	Equalize, align and distribute selected shapes
*	12	I/O of designs when in standalone mode
*	13	Undo operations with control-z/Z. Alternatively use alt-z/Z
*	14	Redo the previous editing operation with control-y/Y. Alternatively
*		use alt-y/Y. Redo-able editing operations for selected shapes are move,
*		resize, rotate, set the fill and border colors, delete, execute (animation),
*		align, distribute and equalize
*/
////////////////////////////////////////////////////////////////////////////////////

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.StringTokenizer;
import java.util.Vector;
import javax.swing.*;

public class DrawingEditor extends JPanel
	implements Constants, AboutDialogUse,
	ActionListener, KeyListener, MouseListener, MouseMotionListener
{
// Parent frame for standalone mode or applet
public DrawingFrame parent;
DrawingApplet applet;

// Shapes input by the user
Drawing target;

// Components
JPanel shapePanel;
JButton circle, line, polygon;

JPanel toolbar;
JButton
	deleteButton, clearButton, exampleButton,
	editButton, resizeButton, rotateButton,
	alignButton, distributeButton, fitButton, equalizeButton;
DrawingView drawingView;

// Selections, clipboard and operations
Vector selections = new Vector(10,5);
Vector clipboard = new Vector(10,5);
Vector operations = new Vector(10,5);

// Redo data
int redo_operator = 0;
int redo_operands_size = 6;
int redo_operands[] = new int[redo_operands_size];

// Positions defined with the mouse
Vector positions = new Vector(10,5);
int curr_x = -1, curr_y = -1;

// Variable used for input
int input_flag = 0;

// Position set with the mouse
int posX = -1, posY = -1;

// Variables used for shapes dragging : flag and previous drag position
boolean shapeMouseDrag = false;
int prev_x = -1, prev_y = -1;

//	Variables used for rubberbanding : flag and start position
boolean rubberbandMouseDrag = false;
Point rubberbandStart = new Point(-1,-1);
Point rubberbandEnd = new Point(-1,-1);

// Colors
static Color BackgroundColor = Color.white;
static Color InputColor = Color.red;

// Sizes
public static int SizeX=500, SizeY=400;
final static Dimension IMAGE_VIEW_DIMENSION = new Dimension(400,400);

// Help message
protected static String helpText =
	"1. Input of shapes\r\n"+
	"  1.1 Use the upper panel buttons (circle, line or polygon).\r\n"+
	"  1.2 Press a button first, then click on two points to input a line\r\n"+
	"    or a circle, or click on any number of points up to 30 to input\r\n"+
	"    a polygon. Polygon input terminates with double-click or <Control>-c/C.\r\n"+
	"  1.3 The newly inserted shape will have the default color.\r\n"+
	"  1.4 By definition, the position of a shape is its upper left\r\n"+
	"    corner (the leftmost or uppermost end for a segment).\r\n"+
	"  1.5 A shape's foreground color may be edited, as well as the view's background\r\n"+
	"    color. (Cf. Clipboard Operations below)\r\n"+
	"\r\n"+
	"2. Selection of shapes\r\n"+
	"  2.1 Click on an shape to select it. For lines click over or close to the line.\r\n"+
	"  2.2 Use <Shift>-click to select multiple shapes\r\n"+
	"  2.3 Use <Control>-a/A to select all shapes. Alternatively use <Alt>-a/A\r\n"+
	"  2.4 Click on the free space to deselect all selected shapes.\r\n"+
	"  2.5 Select all shapes completely included in a rubberband drawn by pressing\r\n"+
	"    the mouse on a start point, then dragging it around and releasing it on an end point.\r\n"+
	"\r\n"+
	"3. Deletion of selected shapes\r\n"+
	"  3.1 Use the <Delete> button on the right button panel or press the <Delete>\r\n"+
	"    or <Backspace> key.\r\n"+
	"  3.2 Use <Control>-x/X to delete the selected shapes and put them on the\r\n"+
	"    clipboard. Alternatively use <Alt>x/X.\r\n"+
	"    (Cf. Clipboard Operations below)\r\n"+
	"\r\n"+
	"4. Translation of selected shapes\r\n"+
	"  4.1 Use plain arrow keys <Left>, <Down>, <Right> and <Up> to move the selected\r\n"+
	"    shapes 1 pixel to the left, bottom, right or top, respectively.\r\n"+
	"  4.2 Use <Alt>-ed arrow keys <Left>, <Down>, <Right> and <Up> to move them\r\n"+
	"    10 pixels to the left, bottom, right or top, respectively.\r\n"+
	"  4.3 Use the mouse to drag them to a new position. The dragging operation\r\n"+
	"    is cancelled if the target position is outside the drawing boundaries.\r\n"+
	"\r\n"+
	"5. Rotation of selected shapes\r\n"+
	"  5.1 Use <Control>-arrow to rotate the selections 90 degrees clockwise (for\r\n"+
	"    left and down arrows) or counter-clockwise (for right or up arrows).\r\n"+
	"  5.2 Use the <Rotate> button on the right button panel to turn the selected\r\n"+
	"    shapes 90 degrees clockwise.\r\n"+
	"\r\n"+
	"6. Resizing of selected shapes\r\n"+
	"  6.1 Use <Shift>-arrow to increase the selected shapes by 1 pixel:\r\n"+
	"    Each arrow moves the extreme point at its direction, for instance the right\r\n"+
	"    arrow moves the rightmost point to the right\r\n"+
	"  6.2 Use <Shift>-<Control>-arrow to decrease the selected shapes by 1 pixel:\r\n"+
	"    Each arrow moves the opposite extreme point at its direction, for instance\r\n"+
	"    the right arrow moves the leftmost point to the right\r\n"+
	"  6.3 Use <Shift>-<Alt>-arrow and <Shift>-<Control>-<Alt>-arrow to increase\r\n"+
	"    increase or decrease the selected shapes by 10 pixels, as before.\r\n"+
	"  6.4 Use the <Resize> button in the right button panel to open a size dialog\r\n"+
	"    on a single selected shape. This dialogs allows the user to introduce a new,\r\n"+
	"    different size for the selected shape, without visualizing all the intermediate\r\n"+
	"    states.\r\n"+
	"  6.5 Use <Control>-r/R to resize the selected shapes freely.\r\n"+
	"      Alternatively use <Alt>-r/R\r\n"+
	"\r\n"+
	"7. Edition\r\n"+
	"  Double click on an shape to edit it or use the <Edit> button on the\r\n"+
	"  right button panel. This button has effect only in the case of a\r\n"+
	"  single object selected. Edition concerns changing its fill and border\r\n"+
	"  colors.\r\n"+
	"\r\n"+
	"8. Clipboard operations\r\n"+
	"  Use <Control>-c/C to copy the selected shapes to the clipboard,\r\n"+
	"  <Control>-x/X to cut the selected shapes, <Control>-v/V to paste them\r\n"+
	"  on a free position defined by clicking. Alternatively use <Alt>-c/C, <Alt>-x/X and <Alt>-v/V\r\n"+
	"\r\n"+
	"9. Equalize selected shapes\r\n"+
	"  Use the <Equalize> button in the right button panel to open an equalize\r\n"+
	"  dialog for the selected shapes. Options are: equalize width, height\r\n"+
	"  or both.\r\n"+
	"\r\n"+
	"10. Align selected shapes\r\n"+
	"  10.1 Use the <Align> button in the right button panel to open an alignment\r\n"+
	"    dialog for the selected shapes.\r\n"+
	"  10.2 There are two parameters to set: alignment position (one of top,\r\n"+
	"    bottom, left and right) and alignment reference (first selection, last\r\n"+
	"    selection, average of all selections, minimum or maximum of\r\n"+
	"    them.\r\n"+
	"\r\n"+
	"11. Distribute selected shapes\r\n"+
	"  11.1 Use the <Distribute> button in the right button panel to open a distribute\r\n"+
	"    dialog for the selected shapes.\r\n"+
	"  11.2 There are two parameters to set: distribute direction (horizontal\r\n"+
	"    or vertical) and distribute type (relative or absolute using a given spacing\r\n"+
	"    in number of pixels.\r\n"+
	"\r\n"+
	"12. Cancellation of operations\r\n"+
	"  Use <Control>-z/Z to cancel previous operations. Alternatively use <Alt>-z/Z\r\n"+
	"\r\n"+
	"13. Redo editing operations\r\n"+
	"  Redo the previous editing operation with control-y/Y. Alternatively use\r\n"+
	"  use alt-y/Y. Redo-able editing operations for selected shapes are move,\r\n"+
	"  resize, rotate, set or invert the fill and border colors, switch from opaque\r\n"+
	"  to transparent and vice versa, delete, execute (animation), send to back\r\n"+
	"  or bring to front, align, distribute and equalize.\r\n"+
	"\r\n"+
	"14. I/O of designs in standalone (non-applet) mode\r\n"+
	"  14.1 Use <Control>-o/O or option <Open> from the <File> menu to open a design file.\r\n"+
	"  14.2 Use <Control>-s/S or option <Save> from the <File> menu to save\r\n"+
	"    a previously opened drawing file.\r\n"+
	"    Use option <Save As> from the <File> menu to save a drawing under a new\r\n"+
	"    name. All drawing filenames are text files and use the extension '.dre'.\r\n"+
	"\r\n"+
	"15. Miscellaneous\r\n"+
	"  15.1 Press <F5> to refresh the display at any time.\r\n"+
	"  15.2 Press <F1> to open a help dialog.\r\n"+
	"  15.3 Press <F7> to open an 'About' Dialog.\r\n"+
	"  15.4 Press <F2>, <F3> and <F4> to respectively step through, resume or suspend a\r\n"+
	"    sample animation of the shapes.\r\n"+
	"  15.5 When in standalone (non-applet) mode, use also the <Refresh> option\r\n"+
	"    from the <Window> menu, the <Help> option and the <About> option\r\n"+
	"    from the <Help> menu, the <Step>, <Start> and <Stop> options from the\r\n"+
	"    <Simulation> menu and the <Send to Back> and <Bring to Front> options from\r\n"+
	"    the <Edit> menu, respectively.\r\n"+
	"  15.6 Use the <Clear> button in the right button panel, to open a new blank\r\n"+
	"    design. When in standalone (non-applet) mode, use also the <New>\r\n"+
	"    option from the <File> menu or press <Control>-n or <Control>-N.\r\n"+
	"  15.7 Use the <Example> button in the right button panel to open a new\r\n"+
	"    exemplar design.\r\n"+
	"\r\n";

public static void main(String args[])
{
	DrawingEditor ed = new DrawingEditor();
	DrawingFrame centralFrame = new DrawingFrame(ed);

	ed.parent = centralFrame;
	ed.init();

	centralFrame.setSize(SizeX,SizeY+100);
	centralFrame.setResizable(false);
	centralFrame.setTitle("Mini Drawing Editor");

	centralFrame.setVisible(true);

//	Parsing argument string for a design file name
	if (args.length > 0)
		centralFrame.tryToOpen(args[0]);
}

public Container getContentPane()
{
	if (parent != null)
		// Standalone mode
		return parent.getContentPane();
	else
		// Applet mode (applet != null)
		return applet.getContentPane();
}

public void init()
{
	initializeGUI();

	// Initialize variables
	target = new Drawing(this);
}

void initializeGUI()
{
//	Shape panel
	shapePanel = new JPanel();
	shapePanel.setLayout(new FlowLayout());
	shapePanel.setBackground(Color.gray);
	circle = new JButton("Circle");
	circle.addActionListener(this);
	line = new JButton("Line");
	line.addActionListener(this);
	polygon = new JButton("Polygon");
	polygon.addActionListener(this);
	shapePanel.add(circle);
	shapePanel.add(line);
	shapePanel.add(polygon);

//	Toolbar
	toolbar = new JPanel();
	toolbar.setLayout(new GridLayout(0,1));
	toolbar.setBackground(Color.gray);
	deleteButton = new JButton("Delete");
	deleteButton.addActionListener(this);
	clearButton = new JButton("Clear");
	clearButton.addActionListener(this);
	exampleButton = new JButton("Example");
	exampleButton.addActionListener(this);
	editButton = new JButton("Edit");
	editButton.addActionListener(this);
	resizeButton = new JButton("Resize");
	resizeButton.addActionListener(this);
	rotateButton = new JButton("Rotate");
	rotateButton.addActionListener(this);
	alignButton = new JButton("Align");
	alignButton.addActionListener(this);
	distributeButton = new JButton("Distribute");
	distributeButton.addActionListener(this);
	fitButton = new JButton("Fit");
	fitButton.addActionListener(this);
	equalizeButton = new JButton("Equalize");
	equalizeButton.addActionListener(this);
	toolbar.add(deleteButton);
	toolbar.add(clearButton);
	toolbar.add(exampleButton);
	toolbar.add(editButton);
	toolbar.add(rotateButton);
	toolbar.add(resizeButton);
	toolbar.add(alignButton);
	toolbar.add(distributeButton);
	toolbar.add(equalizeButton);
	toolbar.add(fitButton);

//	Initialize drawing view
	drawingView = new DrawingView(this,IMAGE_VIEW_DIMENSION);

//	Initialize layout
	getContentPane().setLayout(new BorderLayout());
	getContentPane().add("North",shapePanel);
	getContentPane().add("East",toolbar);
	getContentPane().add("West",drawingView);
}

public void repaint()
{
	super.repaint();
	if (drawingView != null)
		drawingView.repaint();
}

public void actionPerformed(ActionEvent evt)
{
  if (evt.getSource() instanceof JButton)
  {
	JButton tgt = (JButton)evt.getSource();

	if (tgt == line) input_flag = LINE;
	else if (tgt == circle) input_flag = CIRCLE;
	else if (tgt == polygon) input_flag = POLYGON;
	else if (tgt == clearButton)
	{
		clear();
		repaint();
	}
	else if (tgt == editButton)
	{
		if (selections.size() == 1)
			((Shape)selections.elementAt(0)).edit(this);
		else if (selections.size() > 1)
	MessageDialog.open("Warning : Cannot edit multiple selections at once!");
	}
	else if (tgt == deleteButton)
	{
		deleteSelections();
		repaint();
	}
	else if (tgt == exampleButton)
	{
		loadExample();
		repaint();
	}
	else if (tgt == rotateButton)
	{
		rotateSelections();
		repaint();
	}
	else if (tgt == resizeButton)
	{
		reset_input();
		if (selections.size() == 1)
			((Shape)selections.elementAt(0)).setSize(this);
		else if (selections.size() > 1)
	MessageDialog.open("Warning : Cannot resize multiple selections at once!");
	}
	else if (tgt == alignButton)
	{
		reset_input();
		alignSelections();
		repaint();
	}
	else if (tgt == distributeButton)
	{
		reset_input();
		distributeSelections();
		repaint();
	}
	else if (tgt == equalizeButton)
	{
		reset_input();
		equalizeSelections();
		repaint();
	}
	drawingView.requestFocus();
  }
}

public void keyTyped(KeyEvent evt) {}
public void keyReleased(KeyEvent evt) {}
public void keyPressed(KeyEvent evt)
{
	char ch = evt.getKeyChar();
	int key = evt.getKeyCode();

	boolean
		del=false, edit=false,
		shift=false, control=false, alt=false,
		single_selection, has_selections;
	Shape comp;

	shift = evt.isShiftDown();
	control = evt.isControlDown();
	alt = (evt.isMetaDown() || evt.isAltDown());

	boolean noModifiers = (!shift) && (!control) && (!alt);
	boolean onlyShift = (shift) && (!control) && (!alt);
	boolean onlyControl = (!shift) && (control) && (!alt);
	boolean onlyAlt = (!shift) && (!control) && (alt);
	boolean altControl = (!shift) && (control) && (alt);
	boolean shiftControl = (shift) && (control) && (!alt);
	boolean shiftAlt = (shift) && (!control) && (alt);
	boolean shiftAltControl = (shift) && (control) && (alt);

	single_selection = (selections.size() == 1);
	has_selections = (selections.size() > 0);

	if ( (key == KeyEvent.VK_BACK_SPACE) || (key == KeyEvent.VK_DELETE) ) del=true;
	else if (key == KeyEvent.VK_ENTER) edit=true;

	if ((ch=='a') || (ch=='A')) ch='a';
	else if ((ch=='c') || (ch=='C')) ch='c';
	else if ((ch=='r') || (ch=='R')) ch='r';
	else if ((ch=='v') || (ch=='V')) ch='v';
	else if ((ch=='x') || (ch=='X')) ch='x';
	else if ((ch=='y') || (ch=='Y')) ch='y';
	else if ((ch=='z') || (ch=='Z')) ch='z';
	else ch=' ';

	if (del && noModifiers)
	{
		deleteSelections();
		repaint();
		return;
	}

	if (edit && noModifiers)
	{
		if (single_selection)
		{
			((Shape)selections.elementAt(0)).edit(this);
			return;
		}
		else if (selections.size() > 1)
	MessageDialog.open("Warning : Cannot edit multiple selections at once!");
		return;
	}

	// Help and about box
	if ((key == KeyEvent.VK_F1) && noModifiers)
	{
		help();
		return;
	}
	else if ((key == KeyEvent.VK_F7) && noModifiers)
	{
		about();
		return;
	}

	// Refresh
	if ((key == KeyEvent.VK_F5) && noModifiers)
	{
		repaint();
		return;
	}
	// Start, stop and step simulation
	if ((key == KeyEvent.VK_F2) && noModifiers)
	{
		simulationStep();
		repaint();
		return;
	}
	else if ((key == KeyEvent.VK_F3) && noModifiers)
	{
		simulationStart();
		repaint();
		return;
	}
	else if ((key == KeyEvent.VK_F4) && noModifiers)
	{
		simulationStop();
		repaint();
		return;
	}

	// Move, rotate and resize with the arrows
	else if (key == KeyEvent.VK_LEFT)
	{
		if (!has_selections) return;
		if (noModifiers)
		{	// Move by 1
			if (!single_selection) record(null,END_BLOCK);
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				record(comp,MOVE,comp.getPosition().x,comp.getPosition().y);
				comp.moveBy(-1,0);
			}
			if (!single_selection) record(null,START_BLOCK);
			if (selections.size() > 0) set_redo(MOVE_BY,-1,0);
			repaint();
		}
		else if (onlyAlt)
		{	// Move by 10
			if (!single_selection) record(null,END_BLOCK);
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				record(comp,MOVE,comp.getPosition().x,comp.getPosition().y);
				comp.moveBy(-10,0);
			}
			if (!single_selection) record(null,START_BLOCK);
			if (selections.size() > 0) set_redo(MOVE_BY,-10,0);
			repaint();
		}
		else if (onlyShift)
		{	// Increase size by 1
			record(null,END_BLOCK);
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				comp.increaseBy(-1,0,this);
			}
			record(null,START_BLOCK);
			if (selections.size() > 0) set_redo(INCREASE_BY,-1,0);
			repaint();
		}
		else if (shiftControl)
		{	// Decrease size by by 1
			record(null,END_BLOCK);
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				comp.decreaseBy(-1,0,this);
			}
			record(null,START_BLOCK);
			if (selections.size() > 0) set_redo(DECREASE_BY,-1,0);
			repaint();
		}
		else if (shiftAlt)
		{	// Increase size by 10
			record(null,END_BLOCK);
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				comp.increaseBy(-10,0,this);
			}
			record(null,START_BLOCK);
			if (selections.size() > 0) set_redo(INCREASE_BY,-10,0);
			repaint();
		}
		else if (shiftAltControl)
		{	// Decrease size by by 10
		//	if (!single_selection) record(null,END_BLOCK);
			record(null,END_BLOCK);
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				comp.decreaseBy(-10,0,this);
			}
			record(null,START_BLOCK);
			if (selections.size() > 0) set_redo(DECREASE_BY,-10,0);
			repaint();
		}
		else if (onlyControl)
		{	// Rotate
			if (!single_selection) record(null,END_BLOCK);
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				record(comp,ROTATE,1,0);
				comp.rotateBy(-1,0);
			}
			if (!single_selection) record(null,START_BLOCK);
			if (selections.size() > 0) set_redo(ROTATE_BY,-1,0);
			repaint();
		}
	}

	else if (key == KeyEvent.VK_RIGHT)
	{
		if (!has_selections) return;
		if (noModifiers)
		{	// Move by 1
			if (!single_selection) record(null,END_BLOCK);
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				record(comp,MOVE,comp.getPosition().x,comp.getPosition().y);
				comp.moveBy(1,0);
			}
			if (!single_selection) record(null,START_BLOCK);
			if (selections.size() > 0) set_redo(MOVE_BY,1,0);
			repaint();
		}
		else if (onlyAlt)
		{	// Move by 10
			if (!single_selection) record(null,END_BLOCK);
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				record(comp,MOVE,comp.getPosition().x,comp.getPosition().y);
				comp.moveBy(10,0);
			}
			if (!single_selection) record(null,START_BLOCK);
			if (selections.size() > 0) set_redo(MOVE_BY,10,0);
			repaint();
		}
		else if (onlyShift)
		{	// Increase size by 1
			record(null,END_BLOCK);
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				comp.increaseBy(1,0,this);
			}
			record(null,START_BLOCK);
			if (selections.size() > 0) set_redo(INCREASE_BY,1,0);
			repaint();
		}
		else if (shiftControl)
		{	// Decrease size by 1
			record(null,END_BLOCK);
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				comp.decreaseBy(1,0,this);
			}
			record(null,START_BLOCK);
			if (selections.size() > 0) set_redo(DECREASE_BY,1,0);
			repaint();
		}
		else if (shiftAlt)
		{	// Increase size by 10
			record(null,END_BLOCK);
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				comp.increaseBy(10,0,this);
			}
			record(null,START_BLOCK);
			if (selections.size() > 0) set_redo(INCREASE_BY,10,0);
			repaint();
		}
		else if (shiftAltControl)
		{	// Decrease size by 10
			record(null,END_BLOCK);
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				comp.decreaseBy(10,0,this);
			}
			record(null,START_BLOCK);
			if (selections.size() > 0) set_redo(DECREASE_BY,10,0);
			repaint();
		}
		else if (onlyControl)
		{	// Rotate
			if (!single_selection) record(null,END_BLOCK);
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				record(comp,ROTATE,-1,0);
				comp.rotateBy(1,0);
			}
			if (!single_selection) record(null,START_BLOCK);
			if (selections.size() > 0) set_redo(ROTATE_BY,1,0);
			repaint();
		}
	}

	else if (key == KeyEvent.VK_UP)
	{
		if (!has_selections) return;
		if (noModifiers)
		{	// Move by 1
			if (!single_selection) record(null,END_BLOCK);
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				record(comp,MOVE,comp.getPosition().x,comp.getPosition().y);
				comp.moveBy(0,-1);
			}
			if (!single_selection) record(null,START_BLOCK);
			if (selections.size() > 0) set_redo(MOVE_BY,0,-1);
			repaint();
		}
		else if (onlyAlt)
		{	// Move by 10
			if (!single_selection) record(null,END_BLOCK);
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				record(comp,MOVE,comp.getPosition().x,comp.getPosition().y);
				comp.moveBy(0,-10);
			}
			if (!single_selection) record(null,START_BLOCK);
			if (selections.size() > 0) set_redo(MOVE_BY,0,-10);
			repaint();
		}
		else if (onlyShift)
		{	// Increase size by 1
			record(null,END_BLOCK);
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				comp.increaseBy(0,-1,this);
			}
			record(null,START_BLOCK);
			if (selections.size() > 0) set_redo(INCREASE_BY,0,-1);
			repaint();
		}
		else if (shiftControl)
		{	// Decrease size by 1
			record(null,END_BLOCK);
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				comp.decreaseBy(0,-1,this);
			}
			record(null,START_BLOCK);
			if (selections.size() > 0) set_redo(DECREASE_BY,0,-1);
			repaint();
		}
		else if (shiftAlt)
		{	// Increase size by 10
			record(null,END_BLOCK);
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				comp.increaseBy(0,-10,this);
			}
			record(null,START_BLOCK);
			if (selections.size() > 0) set_redo(INCREASE_BY,0,-10);
			repaint();
		}
		else if (shiftAltControl)
		{	// Decrease size by 10
			record(null,END_BLOCK);
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				comp.decreaseBy(0,-10,this);
			}
			record(null,START_BLOCK);
			if (selections.size() > 0) set_redo(DECREASE_BY,0,-10);
			repaint();
		}
		else if (onlyControl)
		{	// Rotate
			if (!single_selection) record(null,END_BLOCK);
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				record(comp,ROTATE,0,1);
				comp.rotateBy(0,-1);
			}
			if (!single_selection) record(null,START_BLOCK);
			if (selections.size() > 0) set_redo(ROTATE_BY,0,-1);
			repaint();
		}
	}

	else if (key == KeyEvent.VK_DOWN)
	{
		if (!has_selections) return;
		if (noModifiers)
		{	// Move by 1
			if (!single_selection) record(null,END_BLOCK);
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				record(comp,MOVE,comp.getPosition().x,comp.getPosition().y);
				comp.moveBy(0,1);
			}
			if (!single_selection) record(null,START_BLOCK);
			if (selections.size() > 0) set_redo(MOVE_BY,0,1);
			repaint();
		}
		else if (onlyAlt)
		{	// Move by 10
			if (!single_selection) record(null,END_BLOCK);
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				record(comp,MOVE,comp.getPosition().x,comp.getPosition().y);
				comp.moveBy(0,10);
			}
			if (!single_selection) record(null,START_BLOCK);
			if (selections.size() > 0) set_redo(MOVE_BY,0,10);
			repaint();
		}
		else if (onlyShift)
		{	// Increase size by 1
			record(null,END_BLOCK);
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				comp.increaseBy(0,1,this);
			}
			record(null,START_BLOCK);
			if (selections.size() > 0) set_redo(INCREASE_BY,0,1);
			repaint();
		}
		else if (shiftControl)
		{	// Decrease size by 1
			record(null,END_BLOCK);
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				comp.decreaseBy(0,1,this);
			}
			record(null,START_BLOCK);
			if (selections.size() > 0) set_redo(DECREASE_BY,0,1);
			repaint();
		}
		else if (shiftAlt)
		{	// Increase size by 10
			record(null,END_BLOCK);
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				comp.increaseBy(0,10,this);
			}
			record(null,START_BLOCK);
			if (selections.size() > 0) set_redo(INCREASE_BY,0,10);
			repaint();
		}
		else if (shiftAltControl)
		{	// Decrease size by 10
			record(null,END_BLOCK);
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				comp.decreaseBy(0,10,this);
			}
			record(null,START_BLOCK);
			if (selections.size() > 0) set_redo(DECREASE_BY,0,10);
			repaint();
		}
		else if (onlyControl)
		{	// Rotate
			if (!single_selection) record(null,END_BLOCK);
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				record(comp,ROTATE,0,-1);
				comp.rotateBy(0,1);
			}
			record(null,START_BLOCK);
			if (selections.size() > 0) set_redo(ROTATE_BY,0,1);
			repaint();
		}
	}

	else if ((ch == 'a') && (onlyControl || onlyAlt)) selectAll();
	else if ((ch == 'c') && (onlyControl || onlyAlt))
	{
		System.out.println("Control-C pressed.");
		if (input_flag == POLYGON)
		{
			if (positions.size() > 0)
			{
				comp = new Shape(POLYGON,positions);
				target.add(comp);
				record(comp,DELETE);
				reset_input();
				clearSelections();
				select(comp);
			}
			repaint();
			return;
		}
		// Copy selections
		copy();
	}
	else if ((ch == 'r') && (onlyControl || onlyAlt))
	{	// Resize the selection
		reset_input();
		if (single_selection)
		{
			((Shape)selections.elementAt(0)).setSize(this);
			return;
		}
		if (selections.size() > 1)
		{
	MessageDialog.open("Warning : Cannot resize multiple selections at once!");
			return;
		}
		repaint();
	}
	else if ((ch == 'v') && (onlyControl || onlyAlt))
	{	// Paste selections
		paste();
	}
	else if ((ch == 'x') && (onlyControl || onlyAlt))
	{	// Cut selections
		cut();
		if (selections.size() > 0) set_redo(DELETE);
	}
	else if ((ch == 'y') && (onlyControl || onlyAlt))
	{
		redo();
		repaint();
	}
	else if (ch == 'z')
	{
		if (onlyControl || onlyAlt)
		{
			undo();
			repaint();
		}
	}
}

public void mouseEntered(MouseEvent evt) {}
public void mouseExited(MouseEvent evt) {}
public void mousePressed(MouseEvent evt)
{
	int x = evt.getX(), y = evt.getY();
	Shape comp;

	boolean
		shift = evt.isShiftDown(),
		control = evt.isControlDown(),
		alt = evt.isMetaDown() || evt.isAltDown();

	drawingView.requestFocus();

	if (control || alt) return;
	
	// Input of a shape
	if (!target.includes(x,y)) {reset_input(); repaint(); return;}
	if (input_flag == LINE)
	{
			positions.addElement((Object)(new Point(x,y)));
			curr_x = x; curr_y = y;
			if (positions.size() == 2)
			{
				comp = new Shape(LINE,(Point)positions.elementAt(0),(Point)positions.elementAt(1));
				target.add(comp);
				record(comp,DELETE);
				reset_input();
				clearSelections();
				select(comp);
			}
			repaint();
			return;
	}
	else if (input_flag == CIRCLE)
	{
			positions.addElement((Object)(new Point(x,y)));
			curr_x = x; curr_y = y;
			if (positions.size() == 2)
			{
				comp = new Shape(CIRCLE,(Point)positions.elementAt(0),
							(Point)positions.elementAt(1));
				target.add(comp);
				record(comp,DELETE);
				reset_input();
				clearSelections();
				select(comp);
			}
			repaint();
			return;
	}
	else if (input_flag == POLYGON)
	{
			positions.addElement((Object)(new Point(x,y)));
			curr_x = x; curr_y = y;
			if ((positions.size() == Shape.SizeLimit) || (evt.getClickCount() == 2))
			{
				comp = new Shape(POLYGON,positions);
				target.add(comp);
				record(comp,DELETE);
				reset_input();
				clearSelections();
				select(comp);
			}
			repaint();
			return;
	}
	else
		// Selection with dragging and edition of shapes
	{
		posX = x;
		posY = y;
		int i = target.shapes.size()-1;
		while (i >= 0)
		{
			comp = (Shape)target.shapes.elementAt(i);
			if (comp.includes(x,y))
			{
				if (!shift) clearSelections();
				select(comp);
	
				// Start dragging selections
				setShapeMouseDrag();
				prev_x = x;
				prev_y = y;
				for (int ind=0;ind<selections.size();ind++)
					((Shape)selections.elementAt(ind)).startDragging();
				posX = x;
				posY = y;
				// If double-clicked, edit the selection
				if (evt.getClickCount() == 2)
				{
					resetShapeMouseDrag();
					if (selections.size() == 1)
					{
						((Shape)selections.elementAt(0)).edit(this);
						repaint();
						return;
					}
					else if (selections.size() > 1)
					{
	MessageDialog.open("Warning : Cannot edit multiple selections at once!");
							return;
					}
				}
				repaint();
				return;
			}
			i--;
		}
		clearSelections();
		setRubberbandMouseDrag();
		rubberbandStart.x = rubberbandEnd.x = x;
		rubberbandStart.y = rubberbandEnd.y = y;
		repaint();
	}
}

public void mouseReleased(MouseEvent evt)
{
	int x = evt.getX(), y = evt.getY();
	Shape comp;

	drawingView.requestFocus();

	if ((!shapeMouseDrag) && (!rubberbandMouseDrag)) return;
	if (shapeMouseDrag)
	{
		resetShapeMouseDrag();
		if (!target.includes(x,y)) return;
		Shape sel = (Shape)selections.elementAt(0);
		if (sel.hasBeenDragged())
		{
			if (selections.size()>1) record(null,END_BLOCK);
			set_redo(-1);
			for (int i=0;i<selections.size();i++)
			{
				comp = (Shape)selections.elementAt(i);
				record(comp,MOVE,comp.getPosition().x,comp.getPosition().y);
				comp.dragBy(x-prev_x,y-prev_y);
				comp.confirmDragged(target);
			}
			if (selections.size()>1) record(null,START_BLOCK);
		}
		prev_x = -1; prev_y = -1;
	}
	else 		// rubberbandMouseDrag
	{
		clearSelections();
		for (int i=0;i<target.shapes.size();i++)
		{
			comp = (Shape)target.shapes.elementAt(i);
			if (inRubberband(comp.topleft())
					&& inRubberband(comp.bottomright()))
				select(comp);
		}
		resetRubberbandMouseDrag();
		rubberbandStart.x = rubberbandStart.y
			= rubberbandEnd.x = rubberbandEnd.y = -1;
	}
	repaint();
}

public void mouseClicked(MouseEvent evt) {}

public void mouseMoved(MouseEvent evt)
{
	int x = evt.getX(), y = evt.getY();
	drawingView.requestFocus();
	if (input_flag != 0)
		if (positions.size() > 0) {curr_x = x; curr_y = y; repaint();}
}

public void mouseDragged(MouseEvent evt)
{
	int x = evt.getX(), y = evt.getY();
	Shape comp;

	drawingView.requestFocus();

	if ((!shapeMouseDrag) && (!rubberbandMouseDrag)) return;
	if (shapeMouseDrag)
	{
		for (int i=0;i<selections.size();i++)
		{
			comp = (Shape)selections.elementAt(i);
			comp.dragBy(x-prev_x,y-prev_y);
		}
		prev_x = x; prev_y = y;
	}
	else		// rubberbanding
	{
		rubberbandEnd.x = x;
		rubberbandEnd.y = y;
	}
	repaint();
}

public void paint(Graphics g)
{
	Color pen = g.getColor();
	super.paint(g);
	g.setColor(BackgroundColor);
	g.fillRect(0,0,SizeX,SizeY);
	g.setColor(pen);
	target.draw(g);
	drawPositions(g);

	if (shapeMouseDrag)
		for (int i=0;i<selections.size();i++)
			((Shape)selections.elementAt(i)).
				drawDragged(g);

	if (rubberbandMouseDrag)
		drawRubberband(g,rubberbandStart,rubberbandEnd);
}

public void drawRubberband(Graphics g, Point start, Point end)
{
	Color pen = g.getColor();
	int	originx = Math.min(start.x, end.x),
		originy = Math.min(start.y, end.y),
		width = Math.abs(start.x - end.x),
		height = Math.abs(start.y - end.y),
		endx = Math.max(start.x, end.x),
		endy = Math.max(start.y, end.y);

	g.setColor(Color.blue);

//	Draw a dashed rectangle
	int i, j, step = 5;
	boolean draw = true;

	for (i=originx;i<=(endx-step);i+=step)
	{
		if (draw) g.drawLine(i, originy, i+step, originy);
		draw = (!draw);
	}
	draw = true;
	for (j=originy;j<=(endy-step);j+=step)
	{
		if (draw) g.drawLine(endx, j, endx, j+step);
		draw = (!draw);
	}
	draw = true;
	for (i=endx;i>=(originx+step);i-=step)
	{
		if (draw) g.drawLine(i, endy, i-step, endy);
		draw = (!draw);
	}
	draw = true;
	for (j=endy;j>=(originy+step);j-=step)
	{
		if (draw) g.drawLine(originx, j, originx, j-step);
		draw = (!draw);
	}

	g.setColor(pen);
}

public boolean inRubberband(Point po)
{
	int	leftmost = Math.min(rubberbandStart.x, rubberbandEnd.x),
		topmost = Math.min(rubberbandStart.y, rubberbandEnd.y),
		rightmost = Math.max(rubberbandStart.x, rubberbandEnd.x),
		bottommost = Math.max(rubberbandStart.y, rubberbandEnd.y);

	return ((po.x >= leftmost) && (po.x <= rightmost)
			&& (po.y >= topmost) && (po.y <= bottommost));
}

public void clearSelections()
{
	for (int i=0;i<selections.size();i++)
	{
		Shape comp = (Shape)selections.elementAt(i);
		comp.deselect();
	}
	selections.removeAllElements();
}

void reset_input()
{
	positions.removeAllElements();
	input_flag = 0;
	curr_x = -1;
	curr_y = -1;
}

void message(String mesg)
{
	System.out.println(mesg);
}

private void setShapeMouseDrag() {shapeMouseDrag = true;}
private void resetShapeMouseDrag() {shapeMouseDrag = false;}
private void setRubberbandMouseDrag() {rubberbandMouseDrag = true;}
private void resetRubberbandMouseDrag() {rubberbandMouseDrag = false;}

public void clear()
{
	clearSelections();
	clearOperations();
	target.clear();
	reset_input();
}

public void loadExample()
{
	target.initExample();
	record(null,CLEAR_ALL);
}

public void rotateSelections()
{
	for (int i=0;i<selections.size();i++)
		((Shape)selections.elementAt(i)).rotate(this);
}

public void alignSelections()
{
	if (selections.size() >= 2) AlignDialog.open(this);
}

public void alignSelections(int direction, int reference, int position)
{
	Point pos;
	Shape comp;

	switch (reference)
	{
		case AlignDialog.ALIGN_FIRST:
			pos = ((Shape)selections.
				elementAt(0)).getAlignPosition(position);
			break;
		case AlignDialog.ALIGN_LAST:
			pos = ((Shape)selections.
				elementAt(selections.size()-1)).getAlignPosition(position);
			break;
		case AlignDialog.ALIGN_AVG:
			pos = compute_avg_selections_pos(position);
			break;
		case AlignDialog.ALIGN_MAX:
			pos = compute_max_selections_pos(position);
			break;
		case AlignDialog.ALIGN_MIN:
			pos = compute_min_selections_pos(position);
			break;
		default:
			pos = ((Shape)selections.
				elementAt(0)).getAlignPosition(position);
			break;
	}
	record(null,END_BLOCK);
	for (int i=0;i<selections.size();i++)
	{
		comp = (Shape)selections.elementAt(i);
		record(comp,MOVE,comp.getPosition().x,comp.getPosition().y);
		comp.align(direction, position, pos);
	}
	record(null,START_BLOCK);
	set_redo(ALIGN,direction,reference,position);
}

public void distributeSelections()
{
	if (selections.size() >= 2) DistributeDialog.open(this);
}

public void distributeSelections(int type, int dimension, int spacing)
{
	Point rel;
	Shape el, comp;
	Vector sels;

	switch (type)
	{
		case DistributeConstants.DISTRIBUTE_ABSOLUTE:
			rel = new Point (spacing,spacing);
			break;
		case DistributeConstants.DISTRIBUTE_RELATIVE:
			rel = compute_uniform_indent_selections();
			break;
		default:
			rel = new Point (spacing,spacing);
			break;
	}
	record(null,END_BLOCK);
	switch (dimension)
	{
		case DistributeConstants.DISTRIBUTE_HORIZONTAL:
			sels = sortLeftmostFirst(selections);
			el = (Shape)sels.elementAt(0);
			for (int i=1;i<sels.size();i++)
			{
				comp = (Shape)sels.elementAt(i);
				record(comp,MOVE,comp.getPosition().x,comp.getPosition().y);
				comp.setPosition((el.relativePosition(rel)).x,comp.getPosition().y);
				el = (Shape)sels.elementAt(i);
			}
			break;
		case DistributeConstants.DISTRIBUTE_VERTICAL:
			sels = sortTopmostFirst(selections);
			el = (Shape)sels.elementAt(0);
			for (int i=1;i<sels.size();i++)
			{
				comp = (Shape)sels.elementAt(i);
				record(comp,MOVE,comp.getPosition().x,comp.getPosition().y);
				comp.setPosition(comp.getPosition().x,(el.relativePosition(rel)).y);
				el = (Shape)sels.elementAt(i);
			}
			break;
		default:
			break;
	}
	record(null,START_BLOCK);
	set_redo(DISTRIBUTE,type,dimension,spacing);
}

public Vector sortLeftmostFirst(Vector v)
{
	int i;
	Vector vec = new Vector(v.size());
	for (i=0;i<v.size();i++) vec.addElement(v.elementAt(i));

	// Use bubble sort
	for (int start=0;start<vec.size();start++)
	for (i=start+1;i<vec.size();i++)
	{
		Shape s_i = (Shape)(vec.elementAt(i));
		Shape s_start = (Shape)(vec.elementAt(start));
		int pos_i = s_i.getPosition().x;
		int pos_start = s_start.getPosition().x;
		if (pos_i<pos_start)
		{
			Shape temp = s_start;
			vec.setElementAt(s_i,start);
			vec.setElementAt(temp,i);
		}
	}
	return vec;
}

public Vector sortTopmostFirst(Vector v)
{
	int i;
	Vector vec = new Vector(v.size());
	for (i=0;i<v.size();i++) vec.addElement(v.elementAt(i));

	// Use bubble sort
	for (int start=0;start<vec.size();start++)
	for (i=start+1;i<vec.size();i++)
	{
		Shape s_i = (Shape)(vec.elementAt(i));
		Shape s_start = (Shape)(vec.elementAt(start));
		int pos_i = s_i.getPosition().y;
		int pos_start = s_start.getPosition().y;
		if (pos_i<pos_start)
		{
			Shape temp = s_start;
			vec.setElementAt(s_i,start);
			vec.setElementAt(temp,i);
		}
	}
	return vec;
}

public void equalizeSelections()
{
	if (selections.size() >= 2) EqualizeDialog.open(this);
}

public void equalizeSelections(int reference, int dimension)
{
  	Dimension theSize;
	Shape comp;

	switch (reference)
	{
		case EqualizeConstants.EQUALIZE_FIRST:
			theSize = ((Shape)selections.
				elementAt(0)).getSize();
			break;
		case EqualizeConstants.EQUALIZE_LAST:
			theSize = ((Shape)selections.
				elementAt(selections.size()-1)).getSize();
			break;
		case EqualizeConstants.EQUALIZE_AVG:
			theSize = compute_avg_selections_size();
			break;
		case EqualizeConstants.EQUALIZE_MAX:
			theSize = compute_max_selections_size();
			break;
		case EqualizeConstants.EQUALIZE_MIN:
			theSize = compute_min_selections_size();
			break;
		default:
			theSize = ((Shape)selections.
				elementAt(0)).getSize();
			break;
	}
	if (selections.size() > 1) record(null,END_BLOCK);
	for (int i=0;i<selections.size();i++)
	{
		if ( ((reference == EqualizeConstants.EQUALIZE_FIRST) && (i == 0)) ||
			((reference == EqualizeConstants.EQUALIZE_LAST) && (i == (selections.size()-1))) )
			;	// Ignore the reference, if there is one
		else
		{
			comp = (Shape)selections.elementAt(i);
			comp.equalize(dimension, theSize, this);
		}
	}
	if (selections.size() > 1) record(null,START_BLOCK);
	set_redo(EQUALIZE,reference,dimension);
}

void selectAll()
{
	clearSelections();
	for (int i=0;i<target.shapes.size();i++)
		select((Shape)target.shapes.elementAt(i));
	repaint();
}

void select(Shape s)
{
	if (s.isSelected()) return;
	s.select();
	selections.addElement(s);
}

void deselect(Shape s)
{
	s.deselect();
	selections.removeElement(s);
}

void copy()
{
	clipboard.removeAllElements();
	for (int i=0;i<selections.size();i++)
	{
		Shape comp = ((Shape)selections.elementAt(i)).copy();
		comp.deselect();
		clipboard.addElement(comp);
	}
}

void cut()
{
	clipboard.removeAllElements();
	if (selections.size() != 1) record(null,END_BLOCK);
	for (int i=0;i<selections.size();i++)
	{
		Shape comp = (Shape)selections.elementAt(i);
		clipboard.addElement(comp);
		target.remove(comp);
		record(comp,ADD,comp.getPosition().x,comp.getPosition().y);
	}
	if (selections.size() != 1) record(null,START_BLOCK);
	clearSelections();
	repaint();
}

void paste()
{
	if ((posX < 0) || (posY < 0)) return;
	if (selections.size() != 1) record(null,END_BLOCK);
	for (int i=0;i<clipboard.size();i++)
	{
		Shape comp = (Shape)clipboard.elementAt(i);
		comp.setPosition(posX,posY);
		comp.deselect();
		comp = comp.copy();
		target.add(comp);
		record(comp,DELETE);
	}
	if (selections.size() != 1) record(null,START_BLOCK);
	repaint();
}

public void clearOperations()
{
	for (int i=0;i<operations.size();i++)
	{
		UndoOperation oper = (UndoOperation)operations.elementAt(i);
		oper.target = null;
	}
	operations.removeAllElements();
	set_redo(-1);
}

public void deleteSelections()
{
	if (selections.size() > 1) record(null,END_BLOCK);
	for (int i=0;i<selections.size();i++)
	{
		Shape comp = (Shape)selections.elementAt(i);
		target.remove(comp);
		record(comp,ADD,comp.getPosition().x,comp.getPosition().y);
	}
	if (selections.size() > 1) record(null,START_BLOCK);
	if (selections.size() > 0) set_redo(DELETE);
	clearSelections();
}

public void simulationStep()
{
	target.simulationStep();
}

public void simulationStart()
{
	target.simulationStart();
}

public void simulationStop()
{
	target.simulationStop();
}

public void help()
{
	TextAreaDialog.open("Help",helpText);
}

public void about()
{
	AboutDialog.open(this,
		"About Drawing Editor",
		"Drawing Editor", "Editor for Geometric Shapes (Lines, Circles and Polygons)",
		"Elpida Tzafestas, 1997", "brensham@softlab.ece.ntua.gr",
		"http://www.softlab.ece.ntua.gr/~brensham");
}

public void undo()
{
	int index;
	UndoOperation current;

	if (operations.size() == 0) return;
	clearSelections();
	index = operations.size()-1;
	current = (UndoOperation)operations.elementAt(index);
	if (current.operator == CLEAR_ALL) clear();
	else if (current.operator == START_BLOCK)
	{
		operations.removeElement(current);
		index--;
		current = (UndoOperation)operations.elementAt(index);
		while (current.operator != END_BLOCK)
		{
			current.target.undo(current.operator,
				current.operands[0],current.operands[1],
				current.operands[2],this);
			operations.removeElement(current);
			index--;
			current = (UndoOperation)operations.elementAt(index);
		}
		// Now, current.operator == END_BLOCK, so remove the UndoOperation.
		operations.removeElement(current);
	}
	else
	{
		current.target.undo(current.operator,
			current.operands[0],current.operands[1],current.operands[2],this);
		operations.removeElement(current);
	}
}

public void record(UndoEntity target, int operator)
{
	operations.addElement(new UndoOperation(target,operator));
}

public void record(UndoEntity target, int operator, int op1)
{
	operations.addElement(new UndoOperation(target,operator,op1));
}

public void record(UndoEntity target, int operator, int op1, int op2)
{
	operations.addElement(new UndoOperation(target,operator,op1,op2));
}

public void record(UndoEntity target, int operator, int op1, int op2, int op3)
{
	operations.addElement(new UndoOperation(target,operator,op1,op2,op3));
}

public void redo()
{
	if (selections.size() == 0) return;
	switch (redo_operator)
	{
		case SET_FILL_AND_BORDER_COLOR:
			record(null,END_BLOCK);
			for (int i=0;i<selections.size();i++)
			{
				Shape comp = (Shape)selections.elementAt(i);
				Color f = comp.getFillColor(), b = comp.getBorderColor();
				comp.recordFillColor(this);
				comp.recordBorderColor(this);
				comp.setFillColor(new Color(redo_operands[0],
								redo_operands[1],redo_operands[2]));
				comp.setBorderColor(new Color(redo_operands[3],
								redo_operands[4],redo_operands[5]));
			}
			record(null,START_BLOCK);
			break;
		case -1 :
			break;
		case ALIGN:
			if (selections.size() >= 2) alignSelections(redo_operands[0],
									redo_operands[1],redo_operands[2]);
			break;
		case DISTRIBUTE:
			if (selections.size() >= 2) distributeSelections(redo_operands[0],
									redo_operands[1],redo_operands[2]);
			break;
		case EQUALIZE:
			if (selections.size() >= 2) equalizeSelections(redo_operands[0],
									redo_operands[1]);
			break;
		default:
			if (selections.size() > 1) record(null,END_BLOCK);
			for (int i=0;i<selections.size();i++)
				((Shape)selections.elementAt(i)).redo(redo_operator,
											redo_operands,this);
			if (selections.size() > 1) record(null,START_BLOCK);
	}
}

public void set_redo(int operator)
{
	redo_operator = operator;
	for (int i=0;i<redo_operands_size;i++) redo_operands[i]=0;
}

public void set_redo(int operator, int op1)
{
	redo_operator = operator;
	for (int i=0;i<redo_operands_size;i++) redo_operands[i]=0;
	redo_operands[0]=op1;
}

public void set_redo(int operator, int op1, int op2)
{
	redo_operator = operator;
	for (int i=0;i<redo_operands_size;i++) redo_operands[i]=0;
	redo_operands[0]=op1;
	redo_operands[1]=op2;
}

public void set_redo(int operator, int op1, int op2, int op3)
{
	redo_operator = operator;
	for (int i=0;i<redo_operands_size;i++) redo_operands[i]=0;
	redo_operands[0]=op1;
	redo_operands[1]=op2;
	redo_operands[2]=op3;
}

public void set_redo(int operator, int op1, int op2, int op3,
						int op4, int op5, int op6)
{
	redo_operator = operator;
	for (int i=0;i<redo_operands_size;i++) redo_operands[i]=0;
	redo_operands[0]=op1;
	redo_operands[1]=op2;
	redo_operands[2]=op3;
	redo_operands[3]=op4;
	redo_operands[4]=op5;
	redo_operands[5]=op6;
}

void drawPositions(Graphics g)
{
//	This method tracks user input in red : all given points are drawn
//	and each one of them is connected to the next one with a straight
//	line. This arrangement is particularly useful for polygon input,
//	where one can see the intermediate state of the polygon as it
//	evolves. Circle and line input is traced in a similar manner.
//	The final mouse pointer is also tracked as moved around.

	Point prev, curr;
	Color pen = g.getColor();
	g.setColor(InputColor);

	if ((input_flag == POLYGON) || (input_flag == LINE))
	{
		if (positions.size() > 0)
		{
			prev = (Point)positions.elementAt(0);
			drawPoint(g,prev);
			for (int i=1;i<positions.size();i++)
			{
				curr = (Point)positions.elementAt(i);
				g.drawLine(prev.x,prev.y,curr.x,curr.y);
				prev=curr;
			}
			g.drawLine(prev.x,prev.y,curr_x,curr_y);
		}
	}
	else if (input_flag == CIRCLE)
	{
		if (positions.size() > 0) // It will be necessarily 1.
		{
			prev = (Point)positions.elementAt(0);
			drawPoint(g,prev);

			// Draw the circle from (prev.x,prev.y) to (curr_x,curr_y)
			int diameter = (int)Math.sqrt(((prev.x-curr_x)*(prev.x-curr_x))
					+ ((prev.y-curr_y)*(prev.y-curr_y)));
			Point up = new Point((prev.x+curr_x-diameter)/2,
						(prev.y+curr_y-diameter)/2);
			g.drawOval(up.x,up.y,diameter,diameter);
		}
	}

	g.setColor(pen);
}

void drawPoint(Graphics g, int x, int y)
	{g.drawLine(x,y,x,y);}

void drawPoint(Graphics g, Point po)
	{g.drawLine(po.x,po.y,po.x,po.y);}

private Point compute_avg_selections_pos(int which)
{
	Point pos = new Point(0,0), temp;

	for (int i=0;i<selections.size();i++)
	{
		temp = ((Shape)selections.elementAt(i)).
				getAlignPosition(which);
		pos.x += temp.x;
		pos.y += temp.y;
	}
	pos.x /= selections.size();
	pos.y /= selections.size();

	return pos;
}

private Point compute_max_selections_pos(int which)
{
	Point pos = new Point(0,0), temp;

	for (int i=0;i<selections.size();i++)
	{
		temp = ((Shape)selections.elementAt(i)).
				getAlignPosition(which);
		pos.x = Math.max(pos.x,temp.x);
		pos.y = Math.max(pos.y,temp.y);
	}
	return pos;
}

private Point compute_min_selections_pos(int which)
{
	Point pos = new Point(10000,10000), temp;

	for (int i=0;i<selections.size();i++)
	{
		temp = ((Shape)selections.elementAt(i)).
				getAlignPosition(which);
		pos.x = Math.min(pos.x,temp.x);
		pos.y = Math.min(pos.y,temp.y);
	}
	return pos;
}

private Point compute_uniform_indent_selections()
{
	Point min = new Point(10000,10000);
	Point max = new Point(0,0);
	Point total = new Point(0,0);
	int n = selections.size()-1;

	for (int i=0;i<selections.size();i++)
	{
		min.x = Math.min(min.x,
			((Shape)selections.elementAt(i)).
			getPosition().x);
		min.y = Math.min(min.y,
			((Shape)selections.elementAt(i)).
			getPosition().y);
		max.x = Math.max(max.x,
			(((Shape)selections.elementAt(i)).
			bottomright()).x);
		max.y = Math.max(max.y,
			(((Shape)selections.elementAt(i)).
			bottomright()).y);
		total.x += ((Shape)selections.elementAt(i)).
			width();
		total.y += ((Shape)selections.elementAt(i)).
			height();
	}
	return new Point( ((max.x-min.x-total.x)/n), ((max.y-min.y-total.y)/n) );
}

private Dimension compute_avg_selections_size()
{
	Dimension dim = new Dimension(0,0);

	for (int i=0;i<selections.size();i++)
	{
		dim.width += ((Shape)selections.elementAt(i)).
			width();
		dim.height += ((Shape)selections.elementAt(i)).
			height();
	}
	dim.width /= selections.size();
	dim.height /= selections.size();

	return dim;
}

private Dimension compute_max_selections_size()
{
	Dimension dim = new Dimension(0,0);

	for (int i=0;i<selections.size();i++)
	{
		dim.width = Math.max(dim.width,
			((Shape)selections.elementAt(i)).
			width());
		dim.height = Math.max(dim.height,
			((Shape)selections.elementAt(i)).
			height());
	}
	return dim;
}

private Dimension compute_min_selections_size()
{
	Dimension dim = new Dimension(10000,10000);

	for (int i=0;i<selections.size();i++)
	{
		dim.width = Math.min(dim.width,
			((Shape)selections.elementAt(i)).
			width());
		dim.height = Math.min(dim.height,
			((Shape)selections.elementAt(i)).
			height());
	}
	return dim;
}

}	// End DrawingEditor

class Drawing implements Constants, Runnable
{
Vector shapes;
Dimension size;

protected Thread thread = null;
protected boolean threadControlVariable = false;

DrawingEditor parent;

static Dimension DefaultSize=new Dimension(450,450);
static Dimension MaximumSize=new Dimension(1000,1000);

//	Light yellow background
Color backgroundColor=Colors.LightYellow;

private void init(Dimension dim, DrawingEditor editor)
{
	size = new Dimension(dim);
	parent = editor;

	shapes = new Vector(10,5);
}

Drawing(DrawingEditor editor)
{
	init(DefaultSize, editor);
}

Drawing(Dimension dim, DrawingEditor editor)
{
	init(dim, editor);
}

public Color getBackgroundColor() {return backgroundColor;}

public boolean includes(int x, int y)
{
	if ((x>0) && (x<size.width) && (y>0) && (y<size.height)) return true;
	return false;
}

private void setSize(Dimension dim)
{
	size = new Dimension(dim);
}

public void add(Shape comp)
{
	shapes.addElement(comp);
}

public void clear()
{
	shapes.removeAllElements();
}

public void remove(Shape comp)
{
	shapes.removeElement(comp);
}

public void draw(Graphics g)
{
	Color c = g.getColor();

	g.setColor(backgroundColor);
	g.fillRect(0,0,size.width,size.height);
	g.setColor(c);

	for (int i=0;i<shapes.size();i++)
		((Shape)(shapes.elementAt(i))).draw(g);
}

public void simulationStep()
{
	if (parent.selections.size()>1) parent.record(null,END_BLOCK);
	for (int i=0;i<shapes.size();i++)
		((Shape)shapes.elementAt(i)).execute(parent);
	if (parent.selections.size()>1) parent.record(null,START_BLOCK);
	if (parent.selections.size()>1) parent.set_redo(EXECUTE);
	parent.repaint();
}

//    ------------------------------------------------
//	void simulationStart() and void simulationStop()
//    ------------------------------------------------

/**********************************************************
//	Old code from Java 1.0.
//	-----------------------
//	Normally, these should work as below, but Windows 95
//	does not support resume and suspend very well.

public void simulationStart()
{
	if (thread == null)
	{
		thread = new Thread(this,"Drawing Shapes Simulation.");
		thread.start();
		thread.suspend();
		if (parent.selections.size()>1) parent.set_redo(EXECUTE);
	}
	thread.resume();
}

public void simulationStop()
{
	thread.suspend();
}

public void run()
{
	while (true) {simulationStep();}
}

**********************************************************/

/**********************************************************
//	Old code from Java 1.1.
//	-----------------------
//	The Thread stop() method is deprecated in Java 1.2,
//	whence the solution below.
----------------------------------------------------------

public void simulationStart()
{
	thread = new Thread(this,"Drawing Shapes Simulation.");
	thread.start();
	if (parent.selections.size()>1) parent.set_redo(EXECUTE);
}

public void simulationStop()
{
	if (thread != null)
	{
		if (thread.isAlive()) thread.stop();
		thread = null;
	}
}

public void run()
{
	while (true) {simulationStep();}
}

**********************************************************/

public void simulationStart()
{
	if (thread == null)
	{
		thread = new Thread(this,"Drawing Shapes Simulation.");
		thread.start();
	}
	threadControlVariable = true;
	if (parent.selections.size()>1) parent.set_redo(EXECUTE);
}

public void simulationStop()
{
	threadControlVariable = false;
}

public void run()
{
	while (true)
	{
		if (threadControlVariable) simulationStep();
	}
}

//////////////////////////////////////////////////////////////////
//			Example : Happy cartoon
//////////////////////////////////////////////////////////////////

public void initExample()
{
	Shape temp;

	int xpos1[] = {87,337,274,150,147};
	int ypos1[] = {225,228,344,344,343};
	shapes.addElement(new Shape(POLYGON,xpos1,ypos1,5,Color.black));
	int xpos2[] = {128,302,216,214};
	int ypos2[] = {225,227,337,337};
	shapes.addElement(new Shape(POLYGON,xpos2,ypos2,4,Color.white));
	int xpos3[] = {104,119,173,247,295,328,328,215,105,104};
	int ypos3[] = {109,176,231,232,188,109,66,35,66,66};
	shapes.addElement(new Shape(POLYGON,xpos3,ypos3,10,Colors.LightPink));

	temp = new Shape(CIRCLE,new Point(123,76),new Point(141,81));
	temp.setFillColor(Color.yellow);
	shapes.addElement(temp);
	
	temp = new Shape(CIRCLE,new Point(178,80),new Point(150,110));
	temp.setFillColor(Color.yellow);
	shapes.addElement(temp);

	temp = new Shape(LINE,new Point(183,93),new Point(243,93));
	temp.setFillColor(Color.blue);
	shapes.addElement(temp);
	
	temp = new Shape(CIRCLE,new Point(279,80),new Point(251,110));
	temp.setFillColor(Color.yellow);
	shapes.addElement(temp);
	
	temp = new Shape(CIRCLE,new Point(287,76),new Point(305,81));
	temp.setFillColor(Color.yellow);
	shapes.addElement(temp);
	
	temp = new Shape(CIRCLE,new Point(224,122),new Point(202,145));
	temp.setFillColor(Color.yellow);
	shapes.addElement(temp);

	int xpos4[] = {158,214,266,263,215,162,163};
	int ypos4[] = {158,186,157,173,212,173,173};
	shapes.addElement(new Shape(POLYGON,xpos4,ypos4,7,Color.yellow));

	temp = new Shape(CIRCLE,new Point(251,100),new Point(267,103));
	temp.setFillColor(Color.blue);
	shapes.addElement(temp);

	temp = new Shape(CIRCLE,new Point(150,99),new Point(166,102));
	temp.setFillColor(Color.blue);
	shapes.addElement(temp);
	
	int xpos5[] = {168,169,260,261,259};
	int ypos5[] = {233,287,233,286,286};
	shapes.addElement(new Shape(POLYGON,xpos5,ypos5,5,Color.red));

	temp = new Shape(CIRCLE,new Point(212,249),new Point(225,267));
	temp.setFillColor(Color.red);
	shapes.addElement(temp);
}

public String toString()
{
	String str="";

	str += (size.width+" "+size.height+" "+
			backgroundColor.getRed()+" "+
			backgroundColor.getGreen()+" "+
			backgroundColor.getBlue()+"\n");
	for (int i=0;i<shapes.size();i++)
		str+=(((Shape)shapes.elementAt(i)).toString()+"\n");
	return str;
}

public void parse(String str)
{
	String token=null;
	int r=255,g=255,b=255;

	StringTokenizer tokenizer = new StringTokenizer(str);

	token = nextToken(tokenizer);
	if (token == null) return;
	size.width=parseWidth(token);

	token = nextToken(tokenizer);
	if (token == null) return;
	size.height=parseHeight(token);

	token = nextToken(tokenizer);
	if (token == null) return;
	r=parseRGB(token);

	token = nextToken(tokenizer);
	if (token == null) return;
	g=parseRGB(token);

	token = nextToken(tokenizer);
	if (token == null) return;
	b=parseRGB(token);

	backgroundColor = new Color(r,g,b);

	while (tokenizer.hasMoreTokens())
	{
		token = nextToken(tokenizer);
		int typ = Shape.parseShapeType(token);
		if (Shape.validType(typ))
		{
			Shape s = new Shape();
			s.setType(typ);
			s.parseWithoutType(tokenizer);
			if (validate(s)) shapes.addElement(s);
		}
	}
}

private boolean validate(Shape s)
{
	if (s == null) return false;
	return (s.size >= 2);
}

private String nextToken(StringTokenizer tokenizer)
{
	if (tokenizer.hasMoreTokens()) return tokenizer.nextToken();
	return null;
}

private int parseWidth(String str)
{
	int n = Integer.parseInt(str);
	if ((n>0) && (n<=MaximumSize.width)) return n;
	return DefaultSize.width;
}

private int parseHeight(String str)
{
	int n = Integer.parseInt(str);
	if ((n>0) && (n<=MaximumSize.height)) return n;
	return DefaultSize.height;
}

private int parseRGB(String str)
{
	int n = Integer.parseInt(str);
	if ((n>=0) && (n<=255)) return n;
	return 255;
}

}	// End Drawing

class Shape implements Constants, HasColors, UndoEntity
{
protected boolean selected = false;
Point dragPosition = new Point(0,0);

Color fillColor = Color.yellow;
Color borderColor = Color.black;
static int SizeLimit = 100;
int type;
int size = 0;
private int xpoints[] = new int[SizeLimit];
private int ypoints[] = new int[SizeLimit];

static double MoveProbability = 0.2;
static double GrowProbability = 0.3;
static double ShrinkProbability = 0.7;

Shape()
{
	type = LINE;
}

Shape(int typ, Point start, Point end)
{
	// Line or circle
	switch (typ)
	{
		case LINE:
		default:
			type = LINE;
			xpoints[0]=start.x; xpoints[1]=end.x;
			ypoints[0]=start.y; ypoints[1]=end.y;
			size=2;
			break;
		case CIRCLE:
			type = CIRCLE;
			int cx = (start.x+end.x)/2;
			int cy = (start.y+end.y)/2;
			int r = (int)(Math.sqrt(((start.x-end.x)*(start.x-end.x))
					+ ((start.y-end.y)*(start.y-end.y))) / 2);
			xpoints[0]= cx - r; xpoints[1]= cx + r;
			ypoints[0]= cy - r; ypoints[1]= cy + r;
			size=2;
			break;
	}
}

Shape(int typ, Vector points)
{
	// Polygon
	Point temp;
	size = Math.min(points.size(),SizeLimit);

	type = POLYGON;
	for (int i=0;i<size;i++)
	{
		temp = (Point)points.elementAt(i);
		xpoints[i]=temp.x;
		ypoints[i]=temp.y;
	}
}

Shape(int typ, int xpos[], int ypos[], int s)
{
	// Practically used only for polygons
	type = typ;
	size = Math.min(Math.min(Math.min(xpos.length,ypos.length), SizeLimit), s);
	for (int i=0;i<size;i++)
		xpoints[i]=xpos[i];
	for (int j=0;j<size;j++)
		ypoints[j]=ypos[j];
}

Shape(int typ, int xpos[], int ypos[], int s, Color c)
{
	// Practically used only for polygons or during copying
	type = typ;
	size = Math.min(Math.min(Math.min(xpos.length,ypos.length), SizeLimit), s);
	for (int i=0;i<size;i++)
		xpoints[i]=xpos[i];
	for (int j=0;j<size;j++)
		ypoints[j]=ypos[j];
	fillColor = c;
}

Shape(int typ, int xpos[], int ypos[], int s, Color f, Color b)
{
	// Practically used only for polygons or during copying
	type = typ;
	size = Math.min(Math.min(Math.min(xpos.length,ypos.length), SizeLimit), s);
	for (int i=0;i<size;i++)
		xpoints[i]=xpos[i];
	for (int j=0;j<size;j++)
		ypoints[j]=ypos[j];
	fillColor = f;
	borderColor = b;
}

public void setType(int typ)
{
	type = typ;
}

public void edit(DrawingEditor ed)
{
	ColorEditor.open("Edit fill and border colors",this,ed);
}

public void select() {selected = true;}
public void deselect() {selected = false;}
public boolean isSelected() {return selected;}

public boolean includes1(int x, int y)
{
	return rectangleIncludes(getPosition(), width(), height(), x, y);
}

public boolean rectangleIncludes(Point upperleft, int width, int height, int x, int y)
{
	if ((x >= upperleft.x) && (x <= (upperleft.x+width))
			&& (y >= upperleft.y) && (y <= (upperleft.y+height)))
		return true;
	return false;
}

public boolean includes(int x, int y)
{
	switch (type)
	{
		case LINE: return lineIncludes(x,y);
		case CIRCLE: return circleIncludes(x,y);
		case POLYGON: return polygonIncludes(x,y);
	}
	return false;
}

public Dimension getSize()
{
	return new Dimension(width(),height());
}

public void setSize(DrawingEditor ed)
{
	ResizeDialog.open(this,ed);
}

public void moveBy(Point po) { moveBy(po.x,po.y); }
public void increaseBy(Point po, DrawingEditor editor) { increaseBy(po.x,po.y,editor); }
public void decreaseBy(Point po, DrawingEditor editor) { decreaseBy(po.x,po.y,editor); }

public void rotate(DrawingEditor editor)
{
	editor.record(this,ROTATE,-1,0);
	editor.set_redo(ROTATE_BY,1,0);
	rotateBy(1,0);
}

protected boolean rotateClockwise(int x, int y)
{
	// Clockwise : (+1,0) -- right -- or (0,+1) --down --
	// Counterclockwise : (-1,0) -- left -- or (0,-1) -- up --
	return (x+y > 0);
}

public Color getFillColor() {return fillColor;}
public void setFillColor(Color c) {fillColor = c;}
public Color getBorderColor() {return borderColor;}
public void setBorderColor(Color c) {borderColor = c;}

public void recordFillColor(DrawingEditor editor)
{
	Color c = getFillColor();
	editor.record(this,Constants.SET_FILL_COLOR,
		c.getRed(),c.getGreen(),c.getBlue());
}

public void recordBorderColor(DrawingEditor editor)
{
	Color c = getBorderColor();
	editor.record(this,Constants.SET_BORDER_COLOR,
		c.getRed(),c.getGreen(),c.getBlue());
}

public Point getAlignPosition(int pos)
{
	switch (pos)
	{
		case AlignConstants.ALIGN_TOP:
			return topleft();
		case AlignConstants.ALIGN_BOTTOM:
			return bottomright();
		case AlignConstants.ALIGN_CENTER:
			return new Point((topleft().x+bottomright().x)/2,
						(topleft().y+bottomright().y)/2);
		case AlignConstants.ALIGN_LEFT:
			return topleft();
		case AlignConstants.ALIGN_RIGHT:
			return topright();
		default:
			return getPosition();
	}
}

public void align(int type, int pos, Point po)
{
	Point oldPosition = getPosition();

	switch (type)
	{
		case AlignConstants.ALIGN_HORIZONTAL:	// Align y coordinates
			switch (pos)
			{
				case AlignConstants.ALIGN_TOP:
					setPosition(oldPosition.x,po.y);
					break;
				case AlignConstants.ALIGN_BOTTOM:
					setPosition(oldPosition.x,po.y-height());
					break;
				case AlignConstants.ALIGN_CENTER:
					setPosition(oldPosition.x,po.y-(height()/2));
					break;			
			}
			break;
		case AlignConstants.ALIGN_VERTICAL:		// Align x coordinates
			switch (pos)
			{
				case AlignConstants.ALIGN_LEFT:
					setPosition(po.x,oldPosition.y);
					break;
				case AlignConstants.ALIGN_RIGHT:
					setPosition(po.x-width(),oldPosition.y);
					break;
				case AlignConstants.ALIGN_CENTER:
					setPosition(po.x-(width()/2),oldPosition.y);
					break;
			}
			break;
	}
}

public Point relativePosition(Point rel)
{
	Point pos = getPosition();
	return new Point(pos.x+width()+rel.x,pos.y+height()+rel.y);
}

public Point topleft() { return getPosition(); }

Shape copy()
{
	return new Shape(type, xpoints, ypoints, size, fillColor, borderColor);
}

public String name()
{
	return nameFor(type);
}

static String nameFor(int typ)
{
	switch (typ)
	{
		case LINE: return "Line";
		case CIRCLE: return "Circle";
		case POLYGON: return "Polygon";
		default: return "Simple Shape";
	}
}

public void draw(Graphics g)
{
	drawUnselected(g);
	if (selected) drawSelected(g);
}

protected void drawSelected1(Graphics g)
{
	Color pen = g.getColor();
	g.setColor(getBorderColor());
	drawSelectedRectangle(g,getPosition(),width(),height());
	g.setColor(pen);
}

protected void drawSelectedRectangle(Graphics g, Point upperleft, int width, int height)
{
	if (selected)
	{
		drawSelectedPoint(g, upperleft.x, upperleft.y);
		drawSelectedPoint(g, upperleft.x, upperleft.y + (height/2));
		drawSelectedPoint(g, upperleft.x, upperleft.y + height);

		drawSelectedPoint(g, upperleft.x + (width / 2), upperleft.y);
		drawSelectedPoint(g, upperleft.x + (width / 2),
			upperleft.y + height);
		drawSelectedPoint(g, upperleft.x + width, upperleft.y);
		drawSelectedPoint(g, upperleft.x + width,
			upperleft.y + (height/2));
		drawSelectedPoint(g, upperleft.x + width,
			upperleft.y + height);
	}
}

protected void drawSelectedPoint(Graphics g, int x, int y)
{
	g.fillRect(x-2, y-2, 5, 5);
}

void drawDragged(Graphics g)
{ drawDraggedAt(g,dragPosition.x,dragPosition.y); }

public void execute(DrawingEditor editor)
{
	long direction = Math.round((Math.random() * 4) - 0.5);

	long action = Math.round((Math.random() * 3) - 0.5);

	if (action == 0)
	{
		if (Math.random() < MoveProbability)
			// Move randomly to one of the four positions around
		{
			Point pos = getPosition();
			editor.record(this,MOVE,pos.x,pos.y);
			if (direction == 0) moveBy(0,-1);
			else if (direction == 1) moveBy(1,0);
			else if (direction == 2) moveBy(0,1);
			else if (direction == 3) moveBy(-1,0);
		}
	}
	else if (action == 1)
	{
		if (Math.random() < GrowProbability)
			// Grow by one
		{
			editor.record(null,END_BLOCK);
			if (direction == 0) increaseBy(0,-1,editor);
			else if (direction == 1) increaseBy(1,0,editor);
			else if (direction == 2) increaseBy(0,1,editor);
			else if (direction == 3) increaseBy(-1,0,editor);
			editor.record(null,START_BLOCK);
		}
	}
	else	// action == 2
	if (Math.random() > ShrinkProbability)
		// Shrink by one
	{
		editor.record(null,END_BLOCK);
		if (direction == 0) decreaseBy(0,-1,editor);
		else if (direction == 1) decreaseBy(1,0,editor);
		else if (direction == 2) decreaseBy(0,1,editor);
		else if (direction == 3) decreaseBy(-1,0,editor);
		editor.record(null,START_BLOCK);
	}
}

public int width()
{
	int leftmost=10000, rightmost=-1, i;

	switch (type)
	{
		case CIRCLE: 
				leftmost = xpoints[0];
				rightmost = xpoints[1];
				break;
		case LINE:
		case POLYGON: 
		default:
				// Find rightmost and leftmost
				for (i=0;i<size;i++)
				{
					if (xpoints[i]>rightmost) rightmost=xpoints[i];
					if (xpoints[i]<leftmost) leftmost=xpoints[i];
				}
				break;
	}

	return rightmost-leftmost;
}

public int height()
{
	int bottommost=-1, topmost=10000, i;

	switch (type)
	{
		case CIRCLE: 
				topmost = ypoints[0];
				bottommost = ypoints[1];
				break;
		case LINE:
		case POLYGON: 
		default:
				// Find bottommost and topmost
				for (i=0;i<size;i++)
				{
					if (ypoints[i]<topmost) topmost=ypoints[i];
					if (ypoints[i]>bottommost) bottommost=ypoints[i];
				}
				break;
	}

	return bottommost-topmost;
}

public void setSize(int wi, int hei)
{
	int leftmost=10000, topmost=10000, i, j;
	int oldwi = width();
	int oldhei = height();

//	Find leftmost and topmost
	for (i=0;i<size;i++)
		if (xpoints[i]<leftmost) leftmost=xpoints[i];
	for (j=0;j<size;j++)
		if (ypoints[j]<topmost) topmost=ypoints[j];

//	Scale all points x's and y's relative to leftmost and topmost
	for (i=0;i<size;i++)
		xpoints[i] = (int)(leftmost + ((xpoints[i]-leftmost)*((float)wi/(float)oldwi)));
	for (j=0;j<size;j++)
		ypoints[j] = (int)(topmost + ((ypoints[j]-topmost)*((float)hei/(float)oldhei)));

	if (type == CIRCLE) validateCircleSize();
}

public Point getPosition()
{
	int leftmost=10000, topmost=10000, i, j;
	switch (type)
	{
		case CIRCLE: 
				leftmost = xpoints[0];
				topmost = ypoints[0];
				break;
		case LINE:
		case POLYGON: 
		default:
				//	Find leftmost and topmost
				for (i=0;i<size;i++)
					if (xpoints[i]<leftmost) leftmost=xpoints[i];
				for (j=0;j<size;j++)
					if (ypoints[j]<topmost) topmost=ypoints[j];
				break;
	}
	return new Point(leftmost,topmost);
}

public Point topright()
{
	int rightmost=-1, topmost=10000, i, j;
	switch (type)
	{
		case LINE:
		case POLYGON: 
		default:
				//	Find rightmost and topmost
				for (i=0;i<size;i++)
					if (xpoints[i]>rightmost) rightmost = xpoints[i];
				for (j=0;j<size;j++)
					if (ypoints[j]<topmost) topmost = ypoints[j];
				break;
		case CIRCLE: 
				rightmost = xpoints[1];
				topmost = ypoints[0];
				break;
	}
	return new Point(rightmost,topmost);
}

public Point bottomright()
{
	int rightmost=-1, bottommost=-1, i, j;
	switch (type)
	{
		case LINE:
		case POLYGON: 
		default:
			//	Find rightmost and bottommost
				for (i=0;i<size;i++)
					if (xpoints[i]>rightmost) rightmost=xpoints[i];
				for (j=0;j<size;j++)
					if (ypoints[j]>bottommost) bottommost=ypoints[j];
				break;
		case CIRCLE: 
				rightmost = xpoints[1];
				bottommost = ypoints[1];
				break;
	}
	return new Point(rightmost,bottommost);
}

public void setPosition(int x, int y)
{
	int leftmost=10000, topmost=10000, i, j;

//	Find leftmost and topmost
	Point po = topleft();

	// Reset topmost and leftmost (reset all points relative to them)
	for (i=0;i<size;i++)
		xpoints[i] += (x-po.x);
	for (j=0;j<size;j++)
		ypoints[j] += (y-po.y);
}

public void setPosition(Point po)
	{setPosition(po.x,po.y);}

void drawUnselected(Graphics g)
{
//	Point up;
	int diameter;
	Color pen = g.getColor();

	switch (type)
	{
		case LINE:
			g.setColor(borderColor);
			g.drawLine(xpoints[0],ypoints[0],
					xpoints[1],ypoints[1]);
			g.setColor(pen);
			break;
		case CIRCLE:
			diameter = xpoints[1] - xpoints[0];
			g.setColor(fillColor);
			g.fillOval(xpoints[0],ypoints[0],diameter,diameter);
			g.setColor(borderColor);
			g.drawOval(xpoints[0],ypoints[0],diameter,diameter);
			g.setColor(pen);
			break;
		case POLYGON:
			g.setColor(fillColor);
			g.fillPolygon(xpoints,ypoints,size);
			g.setColor(borderColor);
			g.drawPolygon(xpoints,ypoints,size);
			g.setColor(pen);
			break;
	}
}

void drawDraggedAt(Graphics g, int x, int y)
{
	Point up, offset, pos = getPosition();
	int diameter;
	Color pen = g.getColor();

	offset = new Point(x - getPosition().x, y - getPosition().y);

	switch (type)
	{
		case LINE:
			g.setColor(borderColor);
			g.drawLine(xpoints[0]+offset.x,ypoints[0]+offset.y,
					xpoints[1]+offset.x,ypoints[1]+offset.y);
			g.setColor(pen);
			break;
		case CIRCLE:
			g.setColor(borderColor);
			g.drawOval(xpoints[0]+offset.x,ypoints[0]+offset.y,
					xpoints[1]-xpoints[0],ypoints[1]-ypoints[0]);
			g.setColor(pen);
			break;
		case POLYGON:
			g.setColor(borderColor);
			int newxpoints[] = new int[SizeLimit];
			int newypoints[] = new int[SizeLimit];
			for (int i=0;i<size;i++)
			{
				newxpoints[i]=xpoints[i]+offset.x;
				newypoints[i]=ypoints[i]+offset.y;
			}
			g.drawPolygon(newxpoints,newypoints,size);
			g.setColor(pen);
			break;
	}
}

protected void drawSelected(Graphics g)
{
	Color pen = g.getColor();
	g.setColor(borderColor);

	switch (type)
	{
		case LINE: lineDrawSelected(g); break;
		case CIRCLE: circleDrawSelected(g); break;
		case POLYGON: polygonDrawSelected(g); break;
	}

	g.setColor(pen);
}

void lineDrawSelected(Graphics g)
{
	drawSelectedPoint(g, xpoints[0], ypoints[0]);
	drawSelectedPoint(g, (xpoints[0] + xpoints[1]) / 2, (ypoints[0] + ypoints[1])/ 2);
	drawSelectedPoint(g, xpoints[1], ypoints[1]);
}

void circleDrawSelected(Graphics g)
{
	//drawSelectedRectangle(g,new Point(xpoints[0],ypoints[0]),xpoints[1]-xpoints[0],ypoints[1]-ypoints[0]);
	Point tl = topleft();
	Point br = bottomright();
	drawSelectedRectangle(g,tl,br.x-tl.x,br.y-tl.y);
}

void polygonDrawSelected(Graphics g)
{
	Point upperleft = new Point(10000,10000);
	Point bottomright = new Point(0,0);
	int width, height;
	for (int i=0;i<size;i++)
	{
		if (xpoints[i] < upperleft.x) upperleft.x = xpoints[i];
		if (xpoints[i] > bottomright.x) bottomright.x = xpoints[i];
		if (ypoints[i] < upperleft.y) upperleft.y = ypoints[i];
		if (ypoints[i] > bottomright.y) bottomright.y = ypoints[i];
	}
	width = bottomright.x - upperleft.x;
	height = bottomright.y - upperleft.y;

	drawSelectedRectangle(g,upperleft,width,height);
}

public void undo(int operator, int op1, int op2, int op3, DrawingEditor editor)
{
	editor.select(this);
	switch (operator)
	{
		case ADD:
			this.setPosition(op1,op2);
			editor.target.add(this);
			break;
		case DELETE:
			editor.target.remove(this);
			break;
		case MOVE:
			setPosition(op1,op2);
			break;
		case RESIZE:
			setSize(op1,op2);
			break;
		case ROTATE:
			rotateBy(op1,op2);
			break;
		case SET_FILL_COLOR:
			setFillColor(new Color(op1,op2,op3));
			break;
		case SET_BORDER_COLOR:
			setBorderColor(new Color(op1,op2,op3));
			break;
		case SET_VERTEX:
			xpoints[op1]=op2;
			ypoints[op1]=op3;
			break;
	}
}

public void redo(int operator, int[] operands, DrawingEditor editor)
{
	switch (operator)
	{
		case MOVE_BY:
			editor.record(this,MOVE,getPosition().x,getPosition().y);
			moveBy(operands[0],operands[1]);
			break;
		case INCREASE_BY:
			editor.record(null,END_BLOCK);
			increaseBy(operands[0],operands[1],editor);
			editor.record(null,START_BLOCK);
			break;
		case DECREASE_BY:
			editor.record(null,END_BLOCK);
			decreaseBy(operands[0],operands[1],editor);
			editor.record(null,START_BLOCK);
			break;
		case ROTATE_BY:
			editor.record(this,ROTATE,-operands[0],-operands[1]);
			rotateBy(operands[0],operands[1]);
			break;
		case RESIZE:
			editor.record(this,RESIZE,width(),height());
			setSize(operands[0],operands[1]);
			break;
		case DELETE:
			editor.record(this,ADD,getPosition().x,getPosition().y);
			editor.target.remove(this);
			break;
		case EXECUTE:
			execute(editor);
			break;
	}
}

protected String nextToken(StringTokenizer tokenizer)
{
	if (tokenizer.hasMoreTokens()) return tokenizer.nextToken();
	return null;
}

public static int parsePosition(String str)
{
	int n = Integer.parseInt(str);
	if (n>0) return n;
	return 1;
}

public static int parseRGB(String str)
{
	int n = Integer.parseInt(str);
	if ((n>=0) && (n<=255)) return n;
	return 255;
}

public boolean lineIncludes(int x,int y)
{
	int lowX;
	int lowY;
	int upX;
	int upY;

	if (xpoints[0] <= xpoints[1])
	{
		lowX = xpoints[0]-1;
		upX = xpoints[1]+1;
	}
	else
	{
		lowX = xpoints[1]-1;
		upX = xpoints[0]+1;
	}
	if (ypoints[0] <= ypoints[1])
	{
		lowY = ypoints[0]-1;
		upY = ypoints[1]+1;
	}
	else
	{
		lowY = ypoints[1]-1;
		upY = ypoints[0]+1;
	}
	if ((x >= lowX) && (x <= upX) && (y >= lowY) & (y <= upY))
		return true;
	return false;
}

public boolean circleIncludes(int x, int y)
{
	return rectangleIncludes(new Point(xpoints[0],ypoints[0]),
					xpoints[1]-xpoints[0],ypoints[1]-ypoints[0],x,y);
}

public boolean polygonIncludes(int x, int y)
{
	Point upperleft = new Point(10000,10000);
	Point bottomright = new Point(0,0);
	int width, height;
	for (int i=0;i<size;i++)
	{
		if (xpoints[i] < upperleft.x) upperleft.x = xpoints[i];
		if (xpoints[i] > bottomright.x) bottomright.x = xpoints[i];
		if (ypoints[i] < upperleft.y) upperleft.y = ypoints[i];
		if (ypoints[i] > bottomright.y) bottomright.y = ypoints[i];
	}
	width = bottomright.x - upperleft.x;
	height = bottomright.y - upperleft.y;

	return rectangleIncludes(upperleft, width, height, x, y);
}

public void validateCircle()
{
	int temp;
	if (xpoints[0] > xpoints[1])
	{
		// Swap xpoints
		temp = xpoints[0];
		xpoints[0] = xpoints[1];
		xpoints[1] = temp;
	}
	if (ypoints[0] > ypoints[1])
	{
		// Swap ypoints
		temp = ypoints[0];
		ypoints[0] = ypoints[1];
		ypoints[1] = temp;
	}
}

public void validateCircleSize()
{
	validateCircle();
	int	cx = xpoints[1] - xpoints[0],
		cy = ypoints[1] - ypoints[0];
	if (cx < cy) xpoints[1] += (cy-cx);
	if (cy < cx) ypoints[1] += (cx-cy);
}

public void rotateBy(int x, int y)
{
	switch (type)
	{
		case LINE:
			rotateLineBy(x,y);
			break;
		case CIRCLE:
			// No in situ rotation necessary !
			break;
		case POLYGON:
			rotatePolygonBy(x,y);
			break;
	}
}

public void rotateLineBy(int x, int y)
{
	int i, j, dx, dy, centerx, centery;

//	Compute centerx and centery
	centerx = (xpoints[0]+xpoints[1])/2;
	centery = (ypoints[0]+ypoints[1])/2;
//
//	Exchange all points x and y relative to centerx and centery
//
	if (rotateClockwise(x,y))
	{
		for (i=0;i<size;i++)
		{
			dx = xpoints[i]-centerx;
			dy = ypoints[i]-centery;
			xpoints[i] = centerx - dy;
			ypoints[i] = centery + dx;
		}
	}
	else
	{
		for (i=0;i<size;i++)
		{
			dx = xpoints[i]-centerx;
			dy = ypoints[i]-centery;
			xpoints[i] = centerx + dy;
			ypoints[i] = centery - dx;
		}
	}
}

public void rotatePolygonByOld(int x, int y)
{
	int leftmost=10000, rightmost=-1, topmost=10000, bottommost=-1,
		i, j, dx, dy, centerx, centery;

//	Find leftmost, rightmost, topmost and bottomost
	for (i=0;i<size;i++)
		if (xpoints[i]<leftmost) leftmost=xpoints[i];
		else if (xpoints[i]>rightmost) rightmost=xpoints[i];
	for (j=0;j<size;j++)
		if (ypoints[j]<topmost) topmost=ypoints[j];
		else if (ypoints[j]>bottommost) bottommost=ypoints[j];
//	Compute centerx and centery
	centerx = (rightmost+leftmost)/2;
	centery = (topmost+bottommost)/2;
//
//	Exchange all points x and y relative to centerx and centery
//
	if (rotateClockwise(x,y))
	{
		for (i=0;i<size;i++)
		{
			dx = xpoints[i]-centerx;
			dy = ypoints[i]-centery;
			xpoints[i] = centerx - dy;
			ypoints[i] = centery + dx;
		}
	}
	else
	{
		for (i=0;i<size;i++)
		{
			dx = xpoints[i]-centerx;
			dy = ypoints[i]-centery;
			xpoints[i] = centerx + dy;
			ypoints[i] = centery - dx;
		}
	}
}

public void rotatePolygonBy(int x, int y)
{
	int i, j, dx, dy, centerx, centery;

//	Compute center
	Point c = center();
	centerx = c.x;
	centery = c.y;
//
//	Exchange all points x and y relative to centerx and centery
//
	if (rotateClockwise(x,y))
	{
		for (i=0;i<size;i++)
		{
			dx = xpoints[i]-centerx;
			dy = ypoints[i]-centery;
			xpoints[i] = centerx - dy;
			ypoints[i] = centery + dx;
		}
	}
	else
	{
		for (i=0;i<size;i++)
		{
			dx = xpoints[i]-centerx;
			dy = ypoints[i]-centery;
			xpoints[i] = centerx + dy;
			ypoints[i] = centery - dx;
		}
	}
}

private Point center()
{
	Point p = new Point(0,0);
	for (int i=0;i<size;i++)
	{
		p.x += xpoints[i];
		p.y += ypoints[i];
	}
	double s = (double)size;
	p.x = (int)Math.round(p.x/s);
	p.y = (int)Math.round(p.y/s);
	return p;
}

public void rotateAround(int cx, int cy, boolean right)
{
	int i;
	if (right)
	{
		for (i=0;i<size;i++)
		{
			int x = xpoints[i], y = ypoints[i];
			xpoints[i] = cx + cy - y;
			ypoints[i] = cy - cx + x;
		}
	}
	else // left
	{
		for (i=0;i<size;i++)
		{
			int x = xpoints[i], y = ypoints[i];
			xpoints[i] = cx - cy + y;
			ypoints[i] = cy + cx - x;
		}
	}
	if (type == CIRCLE) validateCircle();
}

public void moveBy(int x, int y)
{
	int i;
	for (i=0;i<xpoints.length;i++)
		{xpoints[i]+=x;}
	for (i=0;i<ypoints.length;i++)
		{ypoints[i]+=y;}
}

public void increaseBy(int x, int y, DrawingEditor editor)
{
	if (type == CIRCLE)
		increaseCircleBy(x,y,editor);
	else
		increaseLineOrPolygonBy(x,y,editor);
}

public void increaseCircleBy(int x, int y, DrawingEditor editor)
{
	int i, j;

	editor.record(this,SET_VERTEX,0,xpoints[0],ypoints[0]);
	editor.record(this,SET_VERTEX,1,xpoints[1],ypoints[1]);

	if (x>0)
	{
	// Find rightmost x
		xpoints[1] += x;
		ypoints[0] -= x;
	}
	else if (x<0)
	{
	// Find leftmost x
		xpoints[0] += x;
		ypoints[1] -= x;
	}
	if (y>0)
	{
	// Find bottommost y
		ypoints[1] += y;
		xpoints[0] -= y;
	}
	else if (y<0)
	{
	// Find topmost y
		ypoints[0] += y;
		xpoints[1] -= y;
	}
}

public void increaseLineOrPolygonBy(int x, int y, DrawingEditor editor)
{
	int chosen_x, chosen_y, chosen_i = -1, chosen_j = -1, i, j;

	if (x>0)
	{
	// Find rightmost x
		chosen_x = -1;
		for (i=0;i<size;i++)
		{
			if (xpoints[i]>chosen_x)
			{chosen_x = xpoints[i]; chosen_i = i;}
		}
		editor.record(this,SET_VERTEX,chosen_i,xpoints[chosen_i],ypoints[chosen_i]);
		xpoints[chosen_i] += x;
	}
	else if (x<0)
	{
	// Find leftmost x
		chosen_x = 10000;
		for (i=0;i<size;i++)
		{
			if (xpoints[i]<chosen_x)
			{chosen_x = xpoints[i]; chosen_i = i;}
		}
		editor.record(this,SET_VERTEX,chosen_i,xpoints[chosen_i],ypoints[chosen_i]);
		xpoints[chosen_i] += x;
	}
	if (y>0)
	{
	// Find bottommost y
		chosen_y = -1;
		for (j=0;j<size;j++)
		{
			if (ypoints[j]>chosen_y)
			{chosen_y = ypoints[j]; chosen_j = j;}
		}
		editor.record(this,SET_VERTEX,chosen_j,xpoints[chosen_j],ypoints[chosen_j]);
		ypoints[chosen_j] += y;
	}
	else if (y<0)
	{
	// Find topmost y
		chosen_y = 10000;
		for (j=0;j<size;j++)
		{
			if (ypoints[j]<chosen_y)
			{chosen_y = ypoints[j]; chosen_j = j;}
		}
		editor.record(this,SET_VERTEX,chosen_j,xpoints[chosen_j],ypoints[chosen_j]);
		ypoints[chosen_j] += y;
	}
}

public void decreaseBy(int x, int y, DrawingEditor editor)
{
	if (type == CIRCLE)
		decreaseCircleBy(x,y,editor);
	else
		decreaseLineOrPolygonBy(x,y,editor);
}

public void decreaseCircleBy(int x, int y, DrawingEditor editor)
{
	int i, j;

	editor.record(this,SET_VERTEX,0,xpoints[0],ypoints[0]);
	editor.record(this,SET_VERTEX,1,xpoints[1],ypoints[1]);

	if (x>0)
	{
	// Find leftmost x
		xpoints[0] += x;
		ypoints[1] -= x;
	}
	else if (x<0)
	{
	// Find rightmost x
		xpoints[1] += x;
		ypoints[0] -= x;
	}
	if (y>0)
	{
	// Find topmost y
		ypoints[0] += y;
		xpoints[1] -= y;
	}
	else if (y<0)
	{
	// Find bottommost y
		ypoints[1] += y;
		xpoints[0] -= y;
	}
}

public void decreaseLineOrPolygonBy(int x, int y, DrawingEditor editor)
{
	int chosen_x, chosen_y, chosen_i = -1, chosen_j = -1, i, j;

	if (x>0)
	{
	// Find leftmost x
		chosen_x = 10000;
		for (i=0;i<size;i++)
		{
			if (xpoints[i]<chosen_x)
			{chosen_x = xpoints[i]; chosen_i = i;}
		}
		editor.record(this,SET_VERTEX,chosen_i,xpoints[chosen_i],ypoints[chosen_i]);
		xpoints[chosen_i] += x;
	}
	else if (x<0)
	{
	// Find rightmost x
		chosen_x = -1;
		for (i=0;i<size;i++)
		{
			if (xpoints[i]>chosen_x)
			{chosen_x = xpoints[i]; chosen_i = i;}
		}
		editor.record(this,SET_VERTEX,chosen_i,xpoints[chosen_i],ypoints[chosen_i]);
		xpoints[chosen_i] += x;
	}
	if (y>0)
	{
	// Find topmost y
		chosen_y = 10000;
		for (j=0;j<size;j++)
		{
			if (ypoints[j]<chosen_y)
			{chosen_y = ypoints[j]; chosen_j = j;}
		}
		editor.record(this,SET_VERTEX,chosen_j,xpoints[chosen_j],ypoints[chosen_j]);
		ypoints[chosen_j] += y;
	}
	else if (y<0)
	{
	// Find bottommost y
		chosen_y = -1;
		for (j=0;j<size;j++)
		{
			if (ypoints[j]>chosen_y)
			{chosen_y = ypoints[j]; chosen_j = j;}
		}
		editor.record(this,SET_VERTEX,chosen_j,xpoints[chosen_j],ypoints[chosen_j]);
		ypoints[chosen_j] += y;
	}
}

public void dragBy(int x, int y) {dragPosition.x +=x; dragPosition.y +=y;}
public void startDragging()
		{dragPosition.x = getPosition().x; dragPosition.y = getPosition().y;}
public boolean hasBeenDragged()
{
	Point pos = getPosition();
	return ((dragPosition.x != pos.x) || (dragPosition.y != pos.y));
}

public void confirmDragged(Drawing target)
{
	if (target.includes(dragPosition.x,dragPosition.y) &&
		target.includes(dragPosition.x+width(),
						dragPosition.y+height()))
	setPosition(dragPosition.x,dragPosition.y);
}

public void equalize(int dim, Dimension po, DrawingEditor editor)
{
	if (type == CIRCLE)
	{
		int new_dim = 0, old_dim = width();
		if (dim == EqualizeConstants.EQUALIZE_WIDTH) new_dim = po.width;
		else if (dim == EqualizeConstants.EQUALIZE_HEIGHT) new_dim = po.height;
		else if (dim == EqualizeConstants.EQUALIZE_BOTH) new_dim = Math.min(po.width,po.height);
		editor.record(this,SET_VERTEX,0,xpoints[0],ypoints[0]);
		editor.record(this,SET_VERTEX,1,xpoints[1],ypoints[1]);
		xpoints[1] = xpoints[0] + (xpoints[1]-xpoints[0])*new_dim/old_dim;
		ypoints[1] = ypoints[0] + (ypoints[1]-ypoints[0])*new_dim/old_dim;
	}
	else switch (dim)
	{
		case EqualizeConstants.EQUALIZE_WIDTH:
			while (Math.abs(width()-po.width) > 1)
			{
				int leftmost=10000, rightmost=-1, i, rightmost_index=-1;

				// Find rightmost and leftmost
				for (i=0;i<size;i++)
				{
					if (xpoints[i]>rightmost)
						{rightmost=xpoints[i]; rightmost_index=i;}
					if (xpoints[i]<leftmost) leftmost=xpoints[i];
				}
				editor.record(this,SET_VERTEX,rightmost_index,
					xpoints[rightmost_index],ypoints[rightmost_index]);
				xpoints[rightmost_index]=leftmost+po.width;
			}
			break;
		case EqualizeConstants.EQUALIZE_HEIGHT:
			while (Math.abs(height()-po.height) > 1)
			{
				int topmost=10000, bottommost=-1, i, bottommost_index=-1;

				// Find topmost and bottommost
				for (i=0;i<size;i++)
				{
					if (ypoints[i]<topmost) topmost=ypoints[i];
					if (ypoints[i]>bottommost)
						{bottommost=ypoints[i]; bottommost_index=i;}
				}
				editor.record(this,SET_VERTEX,bottommost_index,
					xpoints[bottommost_index],ypoints[bottommost_index]);
				ypoints[bottommost_index]=topmost+po.height;
			}
			break;
		case EqualizeConstants.EQUALIZE_BOTH:
			while (Math.abs(width()-po.width) > 1)
			{
				int leftmost=10000, rightmost=-1, i, rightmost_index=-1;

				// Find rightmost and leftmost
				for (i=0;i<size;i++)
				{
					if (xpoints[i]>rightmost)
						{rightmost=xpoints[i]; rightmost_index=i;}
					if (xpoints[i]<leftmost) leftmost=xpoints[i];
				}
				xpoints[rightmost_index]=leftmost+po.width;
				editor.record(this,SET_VERTEX,rightmost_index,
					xpoints[rightmost_index],ypoints[rightmost_index]);
			}
			while (Math.abs(height()-po.height) > 1)
			{
				int topmost=10000, bottommost=-1, i, bottommost_index=-1;

				// Find topmost and bottommost
				for (i=0;i<size;i++)
				{
					if (ypoints[i]<topmost) topmost=ypoints[i];
					if (ypoints[i]>bottommost)
						{bottommost=ypoints[i]; bottommost_index=i;}
				}
				editor.record(this,SET_VERTEX,bottommost_index,
					xpoints[bottommost_index],ypoints[bottommost_index]);
				ypoints[bottommost_index]=topmost+po.height;
			}
			break;
	}
}

public void parse(StringTokenizer tokenizer)
{
	String token;
	int r=255,g=255,b=255;

	// Parse type, first.
	token = nextToken(tokenizer);
	if (token == null) return;
	type=parseType(token);

	// Then parse size, fill color and border color from tokenizer
	parseWithoutType(tokenizer);
}

public void parseWithoutType(StringTokenizer tokenizer)
{
	String token;
	int r=255,g=255,b=255;

	// Parse size, fill color and border color from tokenizer
	token = nextToken(tokenizer);
	if (token == null) return;
	size=parseSize(token);

	token = nextToken(tokenizer);
	if (token == null) return;
	r=parseRGB(token);

	token = nextToken(tokenizer);
	if (token == null) return;
	g=parseRGB(token);

	token = nextToken(tokenizer);
	if (token == null) return;
	b=parseRGB(token);

	fillColor = new Color(r,g,b);

	token = nextToken(tokenizer);
	if (token == null) return;
	r=parseRGB(token);

	token = nextToken(tokenizer);
	if (token == null) return;
	g=parseRGB(token);

	token = nextToken(tokenizer);
	if (token == null) return;
	b=parseRGB(token);

	borderColor = new Color(r,g,b);

	// Parse xpoints and ypoints from tokenizer
	for (int i=0;i<size;i++)
	{
		int x, y;
		token = nextToken(tokenizer);
		if (token == null) {size = 0; return;}
		x = parsePosition(token);

		token = nextToken(tokenizer);
		if (token == null) {size = 0; return;}
		y = parsePosition(token);

		if (i < SizeLimit)
		{
			xpoints[i] = x;
			ypoints[i] = y;
		}
	}
}

protected static boolean validType(int typ)
{
	switch (typ)
	{
		case LINE:
		case CIRCLE:
		case POLYGON:
			return true;
		default:
			return false;
	}
}

protected int parseType(String str)
{
	int n = Integer.parseInt(str);
	switch (n)
	{
		case LINE:
		case CIRCLE:
		case POLYGON:
			return n;
		default:
			return CIRCLE;
	}
}

public static int parseShapeType(String str)
{
	int n = Integer.parseInt(str);
//	System.out.println(""+nameFor(n));
	return n;
}

protected int parseSize(String str)
{
	int n = Integer.parseInt(str);
	if ((n>0) && (n<=SizeLimit)) return n;
	return 2;
}

public String toString()
{
	String str = new String(type + " " + size);
	str += (" "+fillColor.getRed()+" "+
		fillColor.getGreen()+" "+
		fillColor.getBlue());
	str += (" "+borderColor.getRed()+" "+
		borderColor.getGreen()+" "+
		borderColor.getBlue());
	for (int i=0;i<size;i++)
		str += (" "+xpoints[i]+" "+ypoints[i]);
	return str;
}

}	// End Shape

class DrawingView extends JPanel
{
DrawingEditor gui;
private Dimension size;

DrawingView(DrawingEditor ed, Dimension dim)
{
	gui = ed;
	size = new Dimension(dim);

	addKeyListener(gui);
	addMouseListener(gui);
	addMouseMotionListener(gui);
}

public Dimension getPreferredSize()
	{return new Dimension(size.width,size.height);}

public void paint(Graphics g)
{
	if (g == null) return;
	super.paint(g);

	// Draw the editor view
	gui.paint(g);
}

}	// End DrawingView

class DrawingFrame extends JFrame implements ActionListener,
				KeyListener, MouseListener, MouseMotionListener
{
DrawingEditor target;

JMenuItem
	fileNew, fileOpen, fileSave, fileSaveAs, fileExit,
	editSelectAll, editCut, editCopy, editPaste,
	simulationStep, simulationStart, simulationStop,
	windowRefresh, helpHelp, helpAbout;
private String currentFile = null;
private String illegal_filename_message =
	"Illegal filename. Only *.DRE names are allowed.";

static String postfix = ".dre";

DrawingFrame(DrawingEditor app)
{
	super();
	target = app;

	initBar();
}

private void initBar()
{
	JMenuBar bar = new JMenuBar();

	JMenu fileMenu = new JMenu("File");
	fileNew = new JMenuItem("New",KeyEvent.VK_N);
	fileNew.addActionListener(this);
	fileOpen = new JMenuItem("Open",KeyEvent.VK_O);
	fileOpen.addActionListener(this);
	fileSave = new JMenuItem("Save",KeyEvent.VK_S);
	fileSave.addActionListener(this);
	fileSaveAs = new JMenuItem("Save As...");
	fileSaveAs.addActionListener(this);
	fileExit = new JMenuItem("Exit");
	fileExit.addActionListener(this);
	fileMenu.add(fileNew);
	fileMenu.add(fileOpen);
	fileMenu.addSeparator();
	fileMenu.add(fileSave);
	fileMenu.add(fileSaveAs);
	fileMenu.addSeparator();
	fileMenu.add(fileExit);

	JMenu editMenu = new JMenu("Edit");
	editSelectAll = new JMenuItem("Select All",KeyEvent.VK_A);
	editSelectAll.addActionListener(this);
	editCut = new JMenuItem("Cut",KeyEvent.VK_X);
	editCut.addActionListener(this);
	editCopy = new JMenuItem("Copy",KeyEvent.VK_C);
	editCopy.addActionListener(this);
	editPaste = new JMenuItem("Paste",KeyEvent.VK_V);
	editPaste.addActionListener(this);
	editMenu.add(editSelectAll);
	editMenu.addSeparator();
	editMenu.add(editCut);
	editMenu.add(editCopy);
	editMenu.add(editPaste);

	JMenu simulationMenu = new JMenu("Simulation");
	simulationStep = new JMenuItem("Step -- F2");
	simulationStep.addActionListener(this);
	simulationStart = new JMenuItem("Start -- F3");
	simulationStart.addActionListener(this);
	simulationStop = new JMenuItem("Stop -- F4");
	simulationStop.addActionListener(this);
	simulationMenu.add(simulationStep);
	simulationMenu.add(simulationStart);
	simulationMenu.add(simulationStop);

	JMenu windowMenu = new JMenu("Window");
	windowRefresh = new JMenuItem("Refresh -- F5");
	windowRefresh.addActionListener(this);
	windowMenu.add(windowRefresh);

	JMenu helpMenu = new JMenu("Help");
	helpHelp = new JMenuItem("Help -- F1");
	helpHelp.addActionListener(this);
	helpAbout = new JMenuItem("About -- F7");
	helpAbout.addActionListener(this);
	helpMenu.add(helpHelp);
	helpMenu.addSeparator();
	helpMenu.add(helpAbout);

	bar.add(fileMenu);
	bar.add(editMenu);
	bar.add(simulationMenu);
	bar.add(windowMenu);
	bar.add(helpMenu);
	setJMenuBar(bar);

	addKeyListener(this);
	addMouseListener(this);
	addMouseMotionListener(this);
	addWindowListener(new WindowAdapter()
	{
		public void windowClosing(WindowEvent evt)
			{dispose();System.exit(1);}
	});
	requestFocus();
}

public void keyPressed(KeyEvent evt) {target.keyPressed(evt); requestFocus();}
public void keyTyped(KeyEvent evt) {target.keyTyped(evt); requestFocus();}
public void keyReleased(KeyEvent evt) {target.keyReleased(evt); requestFocus();}

public void mousePressed(MouseEvent evt) {target.mousePressed(evt); requestFocus();}
public void mouseClicked(MouseEvent evt) {target.mouseClicked(evt); requestFocus();}
public void mouseReleased(MouseEvent evt) {target.mouseReleased(evt); requestFocus();}
public void mouseEntered(MouseEvent evt) {target.mouseEntered(evt); requestFocus();}
public void mouseExited(MouseEvent evt) {target.mouseExited(evt); requestFocus();}
public void mouseMoved(MouseEvent evt) {target.mouseMoved(evt); requestFocus();}
public void mouseDragged(MouseEvent evt) {target.mouseDragged(evt); requestFocus();}

public void actionPerformed(ActionEvent evt)
{
  if (evt.getSource() instanceof JMenuItem)
  {
	JMenuItem tgt = (JMenuItem)evt.getSource();

	if (tgt == fileNew)
	{
	//	If a file is open, ask save.
		if (currentFile != null)
			ConfirmSaveDialog.open("Save "+currentFile.toUpperCase()+" first ?",
							this,ConfirmSaveDialog.NEW);
	//	New file
		else doNewFile();
	}
	else if (tgt == fileOpen)
	{
	//	If a file is open, ask save.
		if (currentFile != null)
			ConfirmSaveDialog.open("Save "+currentFile.toUpperCase()+" first ?",
							this,ConfirmSaveDialog.OPEN);
	//	Open file
		else doOpenFile();
	}
	else if (tgt == fileSave)
	{
	//	Save file
		if (currentFile == null)
			MessageDialog.open("Warning : You have not opened any files!"
						+" Use <Save As>.");
		else
			ConfirmSaveDialog.open("Save "+currentFile.toUpperCase()+" ?",
							this,0);
	}
	else if (tgt == fileSaveAs)
	//	Save file under a different name
		saveFileAs();
	else if (tgt == fileExit) {dispose();System.exit(1);}
	else if (tgt == editSelectAll) target.selectAll();
	else if (tgt == editCut) target.cut();
	else if (tgt == editCopy) target.copy();
	else if (tgt == editPaste) target.paste();
	else if (tgt == simulationStep) {target.simulationStep(); target.repaint();}
	else if (tgt == simulationStart) {target.simulationStart(); target.repaint();}
	else if (tgt == simulationStop) {target.simulationStop(); target.repaint();}
	else if (tgt == windowRefresh) target.repaint();
	else if (tgt == helpHelp) target.help();
	else if (tgt == helpAbout) target.about();
  }
  requestFocus();
}

public void doNewFile()
{
	currentFile = null;
	target.clear();
	target.message("New design");
	target.repaint();
}

public boolean accept(File dir, String name)
{
	return name.endsWith(postfix);
}

public void doOpenFile()
{
	String filename;
	FileDialog dlg = new FileDialog(this,"Load Drawing from File",FileDialog.LOAD);
	//dlg.setFilenameFilter(this);
	dlg.setVisible(true);
	filename = dlg.getFile();
	if (filename == null) {target.repaint(); return;}
	filename=checkOpen(filename);
	if (filename != null)
	{
		target.clear();
		target.message("Loading from "+filename.toUpperCase()+". Please wait ...");
		if (read(filename))
		{
			currentFile = filename;
			target.message("Loading from "+filename.toUpperCase()+" successful.");
		}
		else target.message("Loading from "+filename.toUpperCase()+" failed.");
	}
	else target.message(illegal_filename_message);
	target.repaint();
}

public void saveOpenedFile()
{
	target.message("Saving to "+currentFile.toUpperCase()+". Please wait ...");
	if (write(currentFile))
	{
		target.message("Saving to "+currentFile.toUpperCase()+" successful.");
	}
	else target.message("Saving to "+currentFile.toUpperCase()+" failed.");
}

private void saveFileAs()
{
	String filename;
	FileDialog dlg = new FileDialog(this,"Save Drawing to File",FileDialog.SAVE);
	//dlg.setFilenameFilter(this);
	dlg.setVisible(true);
	filename = dlg.getFile();
	if (filename != null)
	{
		filename=checkSave(filename);
		if (filename == null)
		{
			target.message(illegal_filename_message);
			return;
		}
		target.message("Saving to "+filename.toUpperCase()+". Please wait ...");
		if (write(filename))
		{
			currentFile = filename;
			target.message("Saving to "+filename.toUpperCase()+" successful.");
		}
		else target.message("Saving to "+filename.toUpperCase()+" failed.");
	}
}

private boolean read(String file)
{
	String str="";
	Drawing newDrawing = new Drawing(target);

	try
	{
		FileInputStream f = new FileInputStream(file);
		int c;
		while ((c=f.read()) != -1)
		{
			str += String.valueOf((char)c);
		}
		newDrawing.parse(str);
	}
	catch (IOException e)
		{return false;}
	target.target = newDrawing;
	target.repaint();
	return true;
}

private boolean write(String file)
{
	String str = target.target.toString();
	try
	{
		FileOutputStream f = new FileOutputStream(file);
		for (int i=0;i<str.length();i++)
			f.write(str.charAt(i));
	}
	catch (IOException e)
		{return false;}
	return true;
}

private String checkOpen(String f)
{
	if (f.endsWith(postfix))
		return f;
	return null;
}

private String checkSave(String f)
{
	int index;

	if (f.endsWith(postfix))
		return f;
	if ( (index=f.indexOf((int)'.'))== -1) return (f+postfix);
	if (index == 0) return ("Drawing"+postfix);
	return (f.substring(0,index)+postfix);
}

public void tryToOpen(String filename)
{
	String theFile = null;
	int index;

	if (filename.endsWith(postfix)) theFile=filename;
	else
		if ( (index=filename.indexOf((int)'.'))== -1) theFile=filename+postfix;

	if (theFile != null)
	{
		if ((new File(theFile)).exists())
		{
			target.clear();
			target.message("Loading from "+theFile.toUpperCase()+". Please wait ...");
			if (read(theFile))
			{
				currentFile = theFile;
				target.message("Loading from "+theFile.toUpperCase()+" successful.");
			}
			else target.message("Loading from "+theFile.toUpperCase()+" failed.");
			target.repaint();
		}
	}
	else target.message("Illegal filename : "+filename.toUpperCase()+". Load failed.");
}

}	// End DrawingFrame


////////////////////////////////////////////////////
//
//			Various dialogs
//
////////////////////////////////////////////////////

class MessageDialog extends JDialog implements ActionListener
{
  protected static JFrame localFrame;
  protected JButton okButton;
  protected JLabel message;

  MessageDialog(JFrame parent, String mesg)
  {
	super(parent,true);

	okButton = new JButton("OK");
	okButton.addActionListener(this);
	JPanel okPanel = new JPanel();
	okPanel.setLayout(new FlowLayout(FlowLayout.CENTER));
	okPanel.add(okButton);

	message = new JLabel(mesg);
	JPanel messagePanel = new JPanel();
	messagePanel.setLayout(new FlowLayout(FlowLayout.CENTER));
	messagePanel.add(message);
	getContentPane().setLayout(new BorderLayout());
	getContentPane().add("North",messagePanel);
	getContentPane().add("South",okPanel);

	addWindowListener(new WindowAdapter()
	{
		public void windowClosing(WindowEvent e)
		{
			setVisible(false);
			if (localFrame != null)
				localFrame.setVisible(false);
		}
	});
  }

  public void actionPerformed(ActionEvent evt)
  {
	if ((JButton)evt.getSource() == okButton)
	{
		setVisible(false);
		if (localFrame != null)
			localFrame.setVisible(false);
	}
  }

  // Shortcut to create a frame automatically, the frame is a static
  // variable, so that all message dialogs can use the same frame
  public static void open(String mesg)
  {
	// If the frame has not been created yet, create it.
	if (localFrame == null)
		localFrame = new JFrame("");
	// Create the dialog now
	MessageDialog dlg = new MessageDialog(localFrame,mesg);
	// Shrink the frame to just fit the dialog
	localFrame.setSize(dlg.getSize().width,dlg.getSize().height);
	// Show the dialog
	dlg.pack();
	dlg.setVisible(true);
  }

}	// End MessageDialog

class TextAreaDialog extends JDialog implements ActionListener
{
protected static JFrame localFrame;
protected JButton okButton;
protected JTextArea message;

TextAreaDialog(JFrame parent, String title, String text)
{
	super(parent,title,true);

	okButton = new JButton("OK");
	okButton.addActionListener(this);
	JPanel okPanel = new JPanel();
	okPanel.setLayout(new FlowLayout(FlowLayout.CENTER));
	okPanel.add(okButton);

	message = new JTextArea(text,20,70);
	message.setEditable(false);
	JScrollPane messagePanel = new JScrollPane();
	messagePanel.setPreferredSize(new Dimension(550,550));
	messagePanel.setViewportView(message);

	getContentPane().setLayout(new BorderLayout());
	getContentPane().add("North",messagePanel);
	getContentPane().add("South",okPanel);

	addWindowListener(new WindowAdapter()
	{
		public void windowClosing(WindowEvent e)
		{
			setVisible(false);
			if (localFrame != null)
				localFrame.setVisible(false);
		}
	});
}

public void actionPerformed(ActionEvent evt)
{
	if ((JButton)evt.getSource() == okButton)
	{
		setVisible(false);
		if (localFrame != null)
			localFrame.setVisible(false);
	}
}

// Shortcut to create a frame automatically, the frame is a static
// variable, so that all message dialogs can use the same frame

public static void open(String title, String text)
{
	// If the frame has not been created yet, create it.
	if (localFrame == null)
		localFrame = new JFrame("");
	// Create the dialog now
	TextAreaDialog dlg = new TextAreaDialog(localFrame,title,text);

	// Shrink the frame to just fit the dialog
	localFrame.setSize(190,130);

	// Show the dialog
	dlg.pack();
	dlg.setVisible(true);
}

}	// End TextAreaDialog

class AboutDialog extends JDialog implements ActionListener
{
protected static JFrame localFrame;
protected JLabel title1, title2, author, mailto, link;
protected JButton okButton = new JButton("OK");

static Color backColor = Color.cyan;

private AboutDialogUse target;

AboutDialog(JFrame parent, AboutDialogUse app, String title,
	String tit1, String tit2, String auth, String mto, String http)
{
	super(parent,title,true);

	target = app;

	title1 = new JLabel(tit1,JLabel.CENTER);
	title1.setFont(new Font("Dialog",Font.BOLD,18));
	title2 = new JLabel(tit2,JLabel.CENTER);
	title2.setFont(new Font("Dialog",Font.BOLD,16));
	author = new JLabel(auth,JLabel.CENTER);
	author.setFont(new Font("Courier",Font.PLAIN,14));
	mailto = new JLabel(mto,JLabel.CENTER);
	mailto.setFont(new Font("Courier",Font.ITALIC,12));
	link = new JLabel(http,JLabel.CENTER);
	link.setFont(new Font("Courier",Font.ITALIC,12));

//	Component foreground colors
	title1.setForeground(Color.blue);
	title2.setForeground(Color.black);
	author.setForeground(Color.magenta);
	mailto.setForeground(Color.gray);
	link.setForeground(Color.gray);
	okButton.setForeground(Color.black);
	okButton.addActionListener(this);

	initLayout();

	addWindowListener(new WindowAdapter()
	{
		public void windowClosing(WindowEvent e)
		{
			setVisible(false);
			if (localFrame != null)
				localFrame.setVisible(false);
		}
	});
}

private void initLayout()
{
	JPanel componentPanel = new JPanel();
	componentPanel.setLayout(new GridLayout(0,1));
	componentPanel.setBackground(backColor);

	componentPanel.add(title1);
	componentPanel.add(title2);
	componentPanel.add(author);
	componentPanel.add(mailto);
	componentPanel.add(link);

	JPanel buttonPanel = new JPanel();
	buttonPanel.setLayout(new FlowLayout());
	buttonPanel.setBackground(backColor);
	buttonPanel.add(okButton);
	getContentPane().setLayout(new BorderLayout());
	setBackground(backColor);
	setForeground(backColor);
	getContentPane().add("North",componentPanel);
	getContentPane().add("Center",buttonPanel);

	setSize(500,190);
	setResizable(false);
}

public void actionPerformed(ActionEvent evt)
{
	if ((JButton)evt.getSource() == okButton)
	{
		setVisible(false);
		if (localFrame != null)
			localFrame.setVisible(false);
	}
}

// Shortcut to create a frame automatically, the frame is a static
// variable, so that all message dialogs can use the same frame

public static void open(AboutDialogUse app, String title,
		String tit1, String tit2, String auth, String mto, String address)
{
	// If the frame has not been created yet, create it.
	if (localFrame == null)
		localFrame = new JFrame("");
	// Create the dialog now
	AboutDialog dlg = new AboutDialog(localFrame,app,title,
						tit1,tit2,auth,mto,address);

	dlg.setVisible(true);
}

}	// End AboutDialog

class ConfirmSaveDialog extends JDialog implements ActionListener
{
  protected JButton yesButton;
  protected JButton noButton;
  protected JLabel message;
  DrawingFrame target;
  private int nextAction = 0;

//	ConfirmSaveDialog action constants
final static int NEW	= 1001;
final static int OPEN	= 1002;

  ConfirmSaveDialog(DrawingFrame parent, String mesg, int next)
  {
	super(parent,true);

	target = parent;
	nextAction = next;

	yesButton = new JButton("Yes");
	yesButton.addActionListener(this);
	noButton = new JButton("No");
	noButton.addActionListener(this);
	JPanel buttonPanel = new JPanel();
	buttonPanel.setLayout(new FlowLayout(FlowLayout.CENTER));
	buttonPanel.add(yesButton);
	buttonPanel.add(noButton);

	message = new JLabel(mesg);
	JPanel messagePanel = new JPanel();
	messagePanel.setLayout(new FlowLayout(FlowLayout.CENTER));
	messagePanel.add(message);

	getContentPane().setLayout(new BorderLayout());
	getContentPane().add("North",messagePanel);
	getContentPane().add("South",buttonPanel);

	addWindowListener(new WindowAdapter()
	{
		public void windowClosing(WindowEvent e)
		{
			setVisible(false);
		}
	});
  }

  public void actionPerformed(ActionEvent evt)
  {
	if ((JButton)evt.getSource() == yesButton)
	{
		target.saveOpenedFile();
		doNext();
		setVisible(false);
	}
	else if ((JButton)evt.getSource() == noButton)
	{
		doNext();
		setVisible(false);
	}
  }

  private void doNext()
  {
	switch (nextAction)
	{
		case NEW: target.doNewFile(); break;
		case OPEN: target.doOpenFile(); break;
		default: break;
	}
  }

  // Shortcut to create a frame automatically, the frame is a static
  // variable, so that all message dialogs can use the same frame

  public static void open(String mesg, DrawingFrame f, int next)
  {
	// Create the dialog now
	ConfirmSaveDialog dlg = new ConfirmSaveDialog(f,mesg, next);

	// Shrink the frame to just fit the dialog
	dlg.setSize(dlg.getSize().width,dlg.getSize().height);

	// Show the dialog
	dlg.pack();
	dlg.setVisible(true);
  }

}	// End ConfirmSaveDialog

class ColorEditor extends JDialog implements ActionListener
{
protected static JFrame localFrame;
protected JButton okButton,cancelButton;
protected List fillColors;
protected List borderColors;
protected HasColors target;
private DrawingEditor editor;

static Color allColors[] =
 		{Color.black,Color.blue,Colors.Brown,Color.cyan,Color.green,
		Colors.LightGreen,Colors.LightMagenta,Colors.LightPink,
		Colors.LightYellow,Color.magenta,Color.orange,Color.pink,Color.red,
		Color.yellow,Color.white};
static String colorNames[] = {"Black","Blue","Brown","Cyan","Green",
		"Light green","Light magenta","Light pink",
		"Light yellow","Magenta","Orange","Pink","Red",
		"Yellow","White"};

ColorEditor(JFrame parent, String title, HasColors arg, DrawingEditor ed)
{
	super(parent,title,true);

	target = arg;
	editor = ed;

	okButton = new JButton("OK");
	okButton.addActionListener(this);
	cancelButton = new JButton("Cancel");
	cancelButton.addActionListener(this);
	JPanel buttonPanel = new JPanel();
	buttonPanel.setLayout(new FlowLayout(FlowLayout.CENTER));
	buttonPanel.add(okButton);
	buttonPanel.add(cancelButton);

	JPanel labelsPanel = new JPanel();
	labelsPanel.setLayout(new GridLayout(1,0));
	JPanel panel1 = new JPanel();
	panel1.setLayout(new FlowLayout());
	panel1.add(new JLabel("Fill color"));
	JPanel panel2 = new JPanel();
	panel2.setLayout(new FlowLayout());
	panel2.add(new JLabel("Border color"));
	labelsPanel.add(panel1);
	labelsPanel.add(panel2);

	JPanel colorsPanel = new JPanel();
	colorsPanel.setLayout(new GridLayout(1,0));
	colorsPanel.add(fillColors = new List(5,false));
	colorsPanel.add(borderColors = new List(5,false));

	initColors();

	getContentPane().setLayout(new BorderLayout());
	getContentPane().add("North",labelsPanel);
	getContentPane().add("Center",colorsPanel);
	getContentPane().add("South",buttonPanel);

	addWindowListener(new WindowAdapter()
	{
		public void windowClosing(WindowEvent e)
		{
			setVisible(false);
			if (localFrame != null)
				localFrame.setVisible(false);
		}
	});
}

private void initColors()
{
	Color current;
	int i;

	for (i=0;i<colorNames.length;i++)
	{
		fillColors.add(colorNames[i]);
		borderColors.add(colorNames[i]);
	}
	current = target.getFillColor();
	for (i=0;i<allColors.length;i++)
		if (current.equals(allColors[i]))
			{fillColors.select(i); break;}
	current = target.getBorderColor();
	for (i=0;i<allColors.length;i++)
		if (current.equals(allColors[i]))
			{borderColors.select(i); return;}
}

public void actionPerformed(ActionEvent evt)
{
	if (evt.getSource() instanceof JButton)
	{
		JButton tgt = (JButton)evt.getSource();
		if (tgt == okButton)
		{
			update();
			setVisible(false);
			if (localFrame != null)
				localFrame.setVisible(false);
		}
		else if (tgt == cancelButton)
		{
			setVisible(false);
			if (localFrame != null)
				localFrame.setVisible(false);
		}
	}
}

public void update()
{
 	Color f = allColors[fillColors.getSelectedIndex()],
		b = allColors[borderColors.getSelectedIndex()];

	editor.record(null,Constants.END_BLOCK);
	target.recordFillColor(editor);
	target.recordBorderColor(editor);
	editor.record(null,Constants.START_BLOCK);

	target.setFillColor(f);
	target.setBorderColor(b);

	editor.set_redo(Constants.SET_FILL_AND_BORDER_COLOR,
					f.getRed(),f.getGreen(),f.getBlue(),
					b.getRed(),b.getGreen(),b.getBlue());
	editor.repaint();
}

// Shortcut to create a frame automatically, the frame is a static
// variable, so that all message dialogs can use the same frame

public static void open(String title, HasColors arg, DrawingEditor ed)
{
	// If the frame has not been created yet, create it.
	if (localFrame == null)
		localFrame = new JFrame("");
	// Create the dialog now
	ColorEditor dlg = new ColorEditor(localFrame,title,arg,ed);

	// Shrink the frame to just fit the dialog
	localFrame.setSize(190,130);

	// Show the dialog
	dlg.pack();
	dlg.setVisible(true);
}

}	// End ColorEditor

interface AlignConstants
{
//	Alignment constants
final static int ALIGN_HORIZONTAL = 256;
final static int ALIGN_VERTICAL = 257;
final static int ALIGN_TOP = 270;
final static int ALIGN_CENTER = 271;
final static int ALIGN_BOTTOM = 272;
final static int ALIGN_LEFT = 273;
final static int ALIGN_RIGHT = 274;
final static int ALIGN_FIRST = 280;
final static int ALIGN_LAST = 281;
final static int ALIGN_AVG = 282;
final static int ALIGN_MIN = 283;
final static int ALIGN_MAX = 284;
}	// End AlignConstants

class AlignDialog extends JDialog
	implements ActionListener, ItemListener, AlignConstants
{
protected static JFrame localFrame;

protected JPanel horizontalAlignmentPanel;
protected JPanel verticalAlignmentPanel;
protected JPanel mainPanel;

protected JRadioButton topPosition, centerHorzPosition, bottomPosition;
protected JRadioButton leftPosition, centerVertPosition, rightPosition;
protected JRadioButton horzFirstRef, horzLastRef, horzAvgRef, horzUpperRef, horzLowerRef;
protected JRadioButton vertFirstRef, vertLastRef, vertAvgRef, vertLeftRef, vertRightRef;

protected JButton okButton, cancelButton;
protected JRadioButton horizontalDirection, verticalDirection;

public int vertPosition = 0, horzPosition = 0;
public int horzReference = 0, vertReference = 0;

private DrawingEditor target;
private static String HorizontalAlignment = "Horizontally";
private static String VerticalAlignment= "Vertically";

AlignDialog(JFrame parent, DrawingEditor app)
{
	super(parent,true);

	target = app;

	horizontalAlignmentPanel = new JPanel();
	horizontalAlignmentPanel.setLayout(new BorderLayout());

	JPanel horzPositionPanel = new JPanel();
	horzPositionPanel.setLayout(new FlowLayout());
	horzPositionPanel.add(new JLabel("Position"));
	ButtonGroup horzPositionGroup = new ButtonGroup();
	topPosition = new JRadioButton("Top");
	topPosition.setSelected(true);
	horzPositionGroup.add(topPosition);
	centerHorzPosition = new JRadioButton("Center");
	centerHorzPosition.setSelected(false);
	horzPositionGroup.add(centerHorzPosition);
	bottomPosition = new JRadioButton("Bottom");
	bottomPosition.setSelected(false);
	horzPositionGroup.add(bottomPosition);
	horzPositionPanel.add(topPosition);
	horzPositionPanel.add(centerHorzPosition);
	horzPositionPanel.add(bottomPosition);
	topPosition.addItemListener(this);
	centerHorzPosition.addItemListener(this);
	bottomPosition.addItemListener(this);

	JPanel horzReferencePanel = new JPanel();
	horzReferencePanel.setLayout(new FlowLayout());
	horzReferencePanel.add(new JLabel("Reference"));
	ButtonGroup horzReferenceGroup = new ButtonGroup();
	horzFirstRef = new JRadioButton("First");
	topPosition.setSelected(true);
	horzReferenceGroup.add(horzFirstRef);
	horzLastRef = new JRadioButton("Last");
	topPosition.setSelected(false);
	horzReferenceGroup.add(horzLastRef);
	horzAvgRef = new JRadioButton("Average");
	topPosition.setSelected(false);
	horzReferenceGroup.add(horzAvgRef);
	horzUpperRef = new JRadioButton("Uppermost");
	topPosition.setSelected(false);
	horzReferenceGroup.add(horzUpperRef);
	horzLowerRef = new JRadioButton("Lowermost");
	topPosition.setSelected(false);
	horzReferenceGroup.add(horzLowerRef);
	horzReferencePanel.add(horzFirstRef);
	horzReferencePanel.add(horzLastRef);
	horzReferencePanel.add(horzAvgRef);
	horzReferencePanel.add(horzUpperRef);
	horzReferencePanel.add(horzLowerRef);
	horzFirstRef.addItemListener(this);
	horzLastRef.addItemListener(this);
	horzAvgRef.addItemListener(this);
	horzUpperRef.addItemListener(this);
	horzLowerRef.addItemListener(this);

	horizontalAlignmentPanel.add("North",horzPositionPanel);
	horizontalAlignmentPanel.add("South",horzReferencePanel);

	verticalAlignmentPanel = new JPanel();
	verticalAlignmentPanel.setLayout(new BorderLayout());

	JPanel vertPositionPanel = new JPanel();
	vertPositionPanel.setLayout(new FlowLayout());
	vertPositionPanel.add(new JLabel("Position"));
	ButtonGroup vertPositionGroup = new ButtonGroup();
	leftPosition = new JRadioButton("Left");
	leftPosition.setSelected(true);
	vertPositionGroup.add(leftPosition);
	centerVertPosition = new JRadioButton("Center");
	centerVertPosition.setSelected(false);
	vertPositionGroup.add(centerVertPosition);
	rightPosition = new JRadioButton("Right");
	rightPosition.setSelected(false);
	vertPositionGroup.add(rightPosition);
	vertPositionPanel.add(leftPosition);
	vertPositionPanel.add(centerVertPosition);
	vertPositionPanel.add(rightPosition);
	leftPosition.addItemListener(this);
	centerVertPosition.addItemListener(this);
	rightPosition.addItemListener(this);

	JPanel vertReferencePanel = new JPanel();
	vertReferencePanel.setLayout(new FlowLayout());
	vertReferencePanel.add(new JLabel("Reference"));
	ButtonGroup vertReferenceGroup = new ButtonGroup();
	vertFirstRef = new JRadioButton("First");
	vertFirstRef.setSelected(true);
	vertReferenceGroup.add(vertFirstRef);
	vertLastRef = new JRadioButton("Last");
	vertLastRef.setSelected(false);
	vertReferenceGroup.add(vertLastRef);
	vertAvgRef = new JRadioButton("Average");
	vertAvgRef.setSelected(false);
	vertReferenceGroup.add(vertAvgRef);
	vertLeftRef = new JRadioButton("Leftmost");
	vertLeftRef.setSelected(false);
	vertReferenceGroup.add(vertLeftRef);
	vertRightRef = new JRadioButton("Rightmost");
	vertRightRef.setSelected(false);
	vertReferenceGroup.add(vertRightRef);
	vertReferencePanel.add(vertFirstRef);
	vertReferencePanel.add(vertLastRef);
	vertReferencePanel.add(vertAvgRef);
	vertReferencePanel.add(vertLeftRef);
	vertReferencePanel.add(vertRightRef);
	verticalAlignmentPanel.add("North",vertPositionPanel);
	verticalAlignmentPanel.add("South",vertReferencePanel);
	vertFirstRef.addItemListener(this);
	vertLastRef.addItemListener(this);
	vertAvgRef.addItemListener(this);
	vertLeftRef.addItemListener(this);
	vertRightRef.addItemListener(this);

	mainPanel = new JPanel();
	mainPanel.setLayout(new CardLayout());
	mainPanel.add(HorizontalAlignment,horizontalAlignmentPanel);
	mainPanel.add(VerticalAlignment,verticalAlignmentPanel);

	JPanel choicePanel = new JPanel();
	choicePanel.setLayout(new FlowLayout());
	ButtonGroup directionGroup = new ButtonGroup();
	horizontalDirection = new JRadioButton(HorizontalAlignment);
	horizontalDirection.setSelected(true);
	directionGroup.add(horizontalDirection);
	horizontalDirection.addItemListener(this);
	verticalDirection = new JRadioButton(VerticalAlignment);
	verticalDirection.setSelected(false);
	directionGroup.add(verticalDirection);
	verticalDirection.addItemListener(this);
	choicePanel.add(horizontalDirection);
	choicePanel.add(verticalDirection);

	doDefault();

	JPanel buttonPanel = new JPanel();
	buttonPanel.setLayout(new FlowLayout());
	okButton = new JButton("OK");
	okButton.addActionListener(this);
	cancelButton = new JButton("Cancel");
	cancelButton.addActionListener(this);
	buttonPanel.add(okButton);
	buttonPanel.add(cancelButton);

	getContentPane().setLayout(new BorderLayout());
	getContentPane().add("North",choicePanel);
	getContentPane().add("Center",mainPanel);
	getContentPane().add("South",buttonPanel);

	addWindowListener(new WindowAdapter()
	{
		public void windowClosing(WindowEvent e)
		{
			setVisible(false);
			if (localFrame != null)
				localFrame.setVisible(false);
		}
	});
}

public int reference()
{
	if (horizontalAlignmentPanel.isVisible()) return horzReference;
	else return vertReference;
}

public int position()
{
	if (horizontalAlignmentPanel.isVisible()) return horzPosition;
	else return vertPosition;
}

private void doDefault()
{
	horzPosition = ALIGN_TOP;
	topPosition.setSelected(true);

	vertPosition = ALIGN_LEFT;
	leftPosition.setSelected(true);

	horzReference = ALIGN_FIRST;
	horzFirstRef.setSelected(true);

	vertReference = ALIGN_FIRST;
	vertFirstRef.setSelected(true);

	((CardLayout)mainPanel.getLayout()).show(mainPanel,HorizontalAlignment);
	horizontalDirection.setSelected(true);
}

public void itemStateChanged(ItemEvent evt)
{
  if (evt.getSource() instanceof JRadioButton)
  {
	JRadioButton tgt = (JRadioButton)evt.getSource();
	if (tgt == horizontalDirection)
	{
		if (!horizontalAlignmentPanel.isVisible())
		{
			((CardLayout)mainPanel.getLayout()).
				show(mainPanel,HorizontalAlignment);
		}
		return;
	}

	if (tgt == verticalDirection)
	{
		if (!verticalAlignmentPanel.isVisible())
		{
			((CardLayout)mainPanel.getLayout()).
				show(mainPanel,VerticalAlignment);
		}
		return;
	}

	if (tgt == topPosition) {horzPosition = ALIGN_TOP; return ;}
	if (tgt == centerHorzPosition) {horzPosition = ALIGN_CENTER; return ;}
	if (tgt == bottomPosition) {horzPosition = ALIGN_BOTTOM; return ;}

	if (tgt == horzFirstRef) {horzReference = ALIGN_FIRST; return;}
	if (tgt == horzLastRef) {horzReference = ALIGN_LAST; return;}
	if (tgt == horzAvgRef) {horzReference = ALIGN_AVG; return;}

	if (tgt == horzUpperRef) {horzReference = ALIGN_MIN; return;}
	if (tgt == horzLowerRef) {horzReference = ALIGN_MAX; return;}

	if (tgt == leftPosition) {vertPosition = ALIGN_LEFT; return;}
	if (tgt == centerVertPosition) {vertPosition = ALIGN_CENTER; return;}
	if (tgt == rightPosition) {vertPosition = ALIGN_RIGHT; return;}

	if (tgt == vertFirstRef) {vertReference = ALIGN_FIRST; return;}
	if (tgt == vertLastRef) {vertReference = ALIGN_LAST; return;}
	if (tgt == vertAvgRef) {vertReference = ALIGN_AVG; return;}
	if (tgt == vertLeftRef) {vertReference = ALIGN_MIN; return;}
	if (tgt == vertRightRef) {vertReference = ALIGN_MAX; return;}
  }
}

public void actionPerformed(ActionEvent evt)
{
	if (evt.getSource() instanceof JButton)
	{
		JButton tgt = (JButton)evt.getSource();
		if (tgt == okButton)
		{
			setVisible(false);
			if (localFrame != null)
				localFrame.setVisible(false);
			if (horizontalAlignmentPanel.isVisible())
				target.alignSelections(ALIGN_HORIZONTAL,reference(),position());
			else target.alignSelections(ALIGN_VERTICAL,reference(),position());
			target = null;
		}
		else if (tgt == cancelButton)
		{
			setVisible(false);
			if (localFrame != null)
				localFrame.setVisible(false);
			target = null;
		}
	}
}

// Shortcut to create a frame automatically, the frame is a static
// variable, so that all align dialogs can use the same frame

public static void open(DrawingEditor app)
{
	// If the frame has not been created yet, create it.
	if (localFrame == null)
		localFrame = new JFrame("Align selections");
	// Create the dialog now
	AlignDialog dlg = new AlignDialog(localFrame, app);
	// Shrink the frame to just fit the dialog
	localFrame.setSize(dlg.getSize().width,dlg.getSize().height);
	// Show the dialog
	dlg.pack();
	dlg.setVisible(true);
}
}	// End AlignDialog

interface DistributeConstants
{
//	Distribute constants
final static int DISTRIBUTE_HORIZONTAL = 420;
final static int DISTRIBUTE_VERTICAL = 421;
final static int DISTRIBUTE_RELATIVE = 430;
final static int DISTRIBUTE_ABSOLUTE = 431;
}	// End DistributeConstants

class DistributeDialog extends JDialog
	implements ActionListener, ItemListener, DistributeConstants
{
private static int minSpacing = 0;
private static int maxSpacing = 100;

protected static JFrame localFrame;
protected JButton okButton;
protected JButton cancelButton;
protected JRadioButton horizontalDim, verticalDim;
protected JRadioButton relativeSpacing, absoluteSpacing;
protected JButton incrSpacing,decrSpacing;
protected TextField spacingArea;

private DrawingEditor target;

private int distributeDimension = 0, distributeType = 0, spacing = 0;

DistributeDialog(JFrame parent, DrawingEditor app)
{
	super(parent,true);

	target = app;

	JPanel dimensionPanel = new JPanel();
	dimensionPanel.setLayout(new FlowLayout());
	ButtonGroup dimensionGroup = new ButtonGroup();
	horizontalDim = new JRadioButton("Horizontally");
	horizontalDim.setSelected(true);
	dimensionGroup.add(horizontalDim);
	verticalDim = new JRadioButton("Vertically");
	verticalDim.setSelected(false);
	dimensionGroup.add(verticalDim);
	distributeDimension = DISTRIBUTE_HORIZONTAL;
	dimensionPanel.add(horizontalDim);
	dimensionPanel.add(verticalDim);
	horizontalDim.addItemListener(this);
	verticalDim.addItemListener(this);

	JPanel spacingPanel = new JPanel();
	spacingPanel.setLayout(new FlowLayout());
	ButtonGroup spacingGroup = new ButtonGroup();
	relativeSpacing = new JRadioButton("Relative spacing");
	relativeSpacing.setSelected(true);
	spacingGroup.add(relativeSpacing);
	absoluteSpacing = new JRadioButton("Absolute spacing");
	absoluteSpacing.setSelected(false);
	spacingGroup.add(absoluteSpacing);
	relativeSpacing.addItemListener(this);
	absoluteSpacing.addItemListener(this);
	incrSpacing = new JButton("+");
	decrSpacing = new JButton("-");
	incrSpacing.addActionListener(this);
	decrSpacing.addActionListener(this);
	Font bigBoldFont = new Font("System",Font.BOLD,14);
	incrSpacing.setFont(bigBoldFont);
	decrSpacing.setFont(bigBoldFont);
	spacingArea = new TextField(3);
	clearSpacing();
	spacingArea.setEditable(false);
	spacingPanel.add(relativeSpacing);
	spacingPanel.add(absoluteSpacing);
	spacingPanel.add(new JLabel("Spacing ("+minSpacing+" to "+maxSpacing+")"));
	spacingPanel.add(spacingArea);
	spacingPanel.add(incrSpacing);
	spacingPanel.add(decrSpacing);

	doDefault();

	JPanel buttonPanel = new JPanel();
	buttonPanel.setLayout(new FlowLayout());
	okButton = new JButton("OK");
	cancelButton = new JButton("Cancel");
	okButton.addActionListener(this);
	cancelButton.addActionListener(this);
	buttonPanel.add(okButton);
	buttonPanel.add(cancelButton);

	getContentPane().setLayout(new BorderLayout());
	getContentPane().add("North",dimensionPanel);
	getContentPane().add("Center",spacingPanel);
	getContentPane().add("South",buttonPanel);

	addWindowListener(new WindowAdapter()
	{
		public void windowClosing(WindowEvent e)
		{
			setVisible(false);
			if (localFrame != null)
				localFrame.setVisible(false);
		}
	});
}

private void doDefault()
{
	distributeDimension = DISTRIBUTE_HORIZONTAL;
	horizontalDim.setSelected(true);
	distributeType = DISTRIBUTE_RELATIVE;
	relativeSpacing.setSelected(true);
	clearSpacing();
}

public void actionPerformed(ActionEvent evt)
{
	if (evt.getSource() instanceof JButton)
	{
		JButton tgt = (JButton)evt.getSource();
		if (tgt == incrSpacing)
		{
			if (spacing<maxSpacing) {spacing++; spacingArea.setText(""+spacing);}
		}
		else if (tgt == decrSpacing)
		{
			if (spacing>minSpacing) {spacing--; spacingArea.setText(""+spacing);}
		}
		else if (tgt == okButton)
		{
			setVisible(false);
			if (localFrame != null)
				localFrame.setVisible(false);
			target.distributeSelections(distributeType,
								distributeDimension,spacing);
			target = null;
		}
		else if (tgt == cancelButton)
		{
			setVisible(false);
			if (localFrame != null)
				localFrame.setVisible(false);
			target = null;
		}
	}
}

public void itemStateChanged(ItemEvent evt)
{
	if (evt.getSource() instanceof JRadioButton)
	{
		JRadioButton tgt = (JRadioButton)evt.getSource();
		if (tgt == horizontalDim)
			distributeDimension = DISTRIBUTE_HORIZONTAL;
		else if (tgt == verticalDim)
			distributeDimension = DISTRIBUTE_VERTICAL;
		else if (tgt == relativeSpacing)
			{distributeType = DISTRIBUTE_RELATIVE; clearSpacing();}
		else if (tgt == absoluteSpacing)
			{distributeType = DISTRIBUTE_ABSOLUTE; initSpacing();}
	}
}

private void clearSpacing()
{
	spacing = 0;
	spacingArea.setText("");
	incrSpacing.setEnabled(false);
	decrSpacing.setEnabled(false);
}

private void initSpacing()
{
	spacing = 0;
	spacingArea.setText("0");
	incrSpacing.setEnabled(true);
	decrSpacing.setEnabled(true);
}

// Shortcut to create a frame automatically, the frame is a static
// variable, so that all align dialogs can use the same frame

public static void open(DrawingEditor app)
{
	// If the frame has not been created yet, create it.
	if (localFrame == null)
		localFrame = new JFrame("DistributeSelections");
	// Create the dialog now
	DistributeDialog dlg = new DistributeDialog(localFrame,app);
	// Shrink the frame to just fit the dialog
	localFrame.setSize(dlg.getSize().width,dlg.getSize().height);
	// Show the dialog
	dlg.pack();
	dlg.setVisible(true);
 }

}	// End DistributeDialog

interface EqualizeConstants
{
//	Equalize constants
final static int EQUALIZE_WIDTH = 340;
final static int EQUALIZE_HEIGHT = 341;
final static int EQUALIZE_BOTH = 342;
final static int EQUALIZE_FIRST = 350;
final static int EQUALIZE_LAST = 351;
final static int EQUALIZE_AVG = 352;
final static int EQUALIZE_MIN = 353;
final static int EQUALIZE_MAX = 354;
}	// End EqualizeConstants

class EqualizeDialog extends JDialog
	implements Constants, ActionListener, ItemListener, EqualizeConstants
{

protected static JFrame localFrame;
protected JButton okButton;
protected JButton cancelButton;

protected JRadioButton widthDim, heightDim, bothDim;
protected JRadioButton firstRef, lastRef, avgRef, minRef, maxRef;

private DrawingEditor target;
public int equalizeDimension = 0, equalizeReference = 0;

EqualizeDialog(JFrame parent, DrawingEditor app)
{
	super(parent,true);

	target = app;

	JPanel dimensionPanel = new JPanel();
	dimensionPanel.setLayout(new FlowLayout());
	ButtonGroup dimensionGroup = new ButtonGroup();
	widthDim = new JRadioButton("Width");
	widthDim.setSelected(true);
	widthDim.addItemListener(this);
	dimensionGroup.add(widthDim);
	heightDim = new JRadioButton("Height");
	heightDim.setSelected(false);
	heightDim.addItemListener(this);
	dimensionGroup.add(heightDim);
	bothDim = new JRadioButton("Both");
	bothDim.setSelected(false);
	bothDim.addItemListener(this);
	dimensionGroup.add(bothDim);
	dimensionPanel.add(new JLabel("Dimension"));
	dimensionPanel.add(widthDim);
	dimensionPanel.add(heightDim);
	dimensionPanel.add(bothDim);

	JPanel referencePanel = new JPanel();
	referencePanel.setLayout(new FlowLayout());
	ButtonGroup referenceGroup = new ButtonGroup();
	firstRef = new JRadioButton("First");
	firstRef.setSelected(true);
	firstRef.addItemListener(this);
	referenceGroup.add(firstRef);
	lastRef = new JRadioButton("Last");
	lastRef.setSelected(false);
	lastRef.addItemListener(this);
	referenceGroup.add(lastRef);
	avgRef = new JRadioButton("Average");
	avgRef.setSelected(false);
	avgRef.addItemListener(this);
	referenceGroup.add(avgRef);
	minRef = new JRadioButton("Min");
	minRef.setSelected(false);
	minRef.addItemListener(this);
	referenceGroup.add(minRef);
	maxRef = new JRadioButton("Max");
	maxRef.setSelected(false);
	maxRef.addItemListener(this);
	referenceGroup.add(maxRef);
	referencePanel.add(new JLabel("Reference"));
	referencePanel.add(firstRef);
	referencePanel.add(lastRef);
	referencePanel.add(avgRef);
	referencePanel.add(minRef);
	referencePanel.add(maxRef);

	doDefault();

	JPanel buttonPanel = new JPanel();
	buttonPanel.setLayout(new FlowLayout());
	okButton = new JButton("OK");
	okButton.addActionListener(this);
	cancelButton = new JButton("Cancel");
	cancelButton.addActionListener(this);
	buttonPanel.add(okButton);
	buttonPanel.add(cancelButton);

	getContentPane().setLayout(new BorderLayout());
	getContentPane().add("North",dimensionPanel);
	getContentPane().add("Center",referencePanel);
	getContentPane().add("South",buttonPanel);

	addWindowListener(new WindowAdapter()
	{
		public void windowClosing(WindowEvent e)
		{
			setVisible(false);
			if (localFrame != null)
				localFrame.setVisible(false);
		}
	});
}

private void doDefault()
{
	equalizeDimension = EQUALIZE_WIDTH;
	widthDim.setSelected(true);
	equalizeReference = EQUALIZE_FIRST;
	firstRef.setSelected(true);
}

public void itemStateChanged(ItemEvent evt)
{
	if (evt.getSource() instanceof JRadioButton)
	{
		JRadioButton tgt = (JRadioButton)evt.getSource();

		if (tgt == widthDim) equalizeDimension = EQUALIZE_WIDTH;
		else if (tgt == heightDim) equalizeDimension = EQUALIZE_HEIGHT;
		else if (tgt == bothDim) equalizeDimension = EQUALIZE_BOTH;
	
		else if (tgt == firstRef) equalizeReference = EQUALIZE_FIRST;
		else if (tgt == lastRef) equalizeReference = EQUALIZE_LAST;
		else if (tgt == avgRef) equalizeReference = EQUALIZE_AVG;
		else if (tgt == minRef) equalizeReference = EQUALIZE_MIN;
		else if (tgt == maxRef) equalizeReference = EQUALIZE_MAX;
	}
}

public void actionPerformed(ActionEvent evt)
{
	if (evt.getSource() instanceof JButton)
	{
		JButton tgt = (JButton)evt.getSource();
		if (tgt == okButton)
		{
			setVisible(false);
			if (localFrame != null)
				localFrame.setVisible(false);
			target.equalizeSelections(equalizeReference,equalizeDimension);
			target = null;
		}
		else if (tgt == cancelButton)
		{
			setVisible(false);
			if (localFrame != null)
				localFrame.setVisible(false);
			target = null;
		}
	}
}

// Shortcut to create a frame automatically, the frame is a static
// variable, so that all align dialogs can use the same frame

public static void open(DrawingEditor app)
{
	// If the frame has not been created yet, create it.
	if (localFrame == null)
		localFrame = new JFrame("Equalize selections");
	// Create the dialog now
	EqualizeDialog dlg = new EqualizeDialog(localFrame,app);
	// Shrink the frame to just fit the dialog
	localFrame.setSize(dlg.getSize().width,dlg.getSize().height);
	// Show the dialog
	dlg.pack();
	dlg.setVisible(true);
}

}	// End EqualizeDialog

class ResizeDialog extends JDialog
	implements Constants, ActionListener
{
protected static JFrame localFrame;

private static int minWidth = 5;
private static int maxWidth = 100;
private static int minHeight = 5;
private static int maxHeight = 100;

protected JButton okButton;
protected JButton cancelButton;

protected JButton width_incr, width_decr, height_incr, height_decr;
protected TextField /*sizeArea,*/ widthArea, heightArea;

int width = 0, height = 0;
private Shape target;
private DrawingEditor editor;

ResizeDialog(JFrame parent, Shape s, Dimension dim, DrawingEditor ed)
{
	super(parent,true);

	target = s;
	editor = ed;

	JPanel widthPanel = new JPanel();
	widthPanel.setLayout(new FlowLayout());
	widthArea = new TextField(3);
	widthArea.setText(""+dim.width);
	widthArea.setEditable(false);
	width_incr = new JButton("+");
	width_decr = new JButton("-");
	Font bigBoldFont = new Font("System",Font.BOLD,14);
	width_incr.setFont(bigBoldFont);
	width_decr.setFont(bigBoldFont);
	widthPanel.add(new JLabel("Width  ("+minWidth+" to "+maxWidth+")"));
	widthPanel.add(widthArea);
	widthPanel.add(width_incr);
	widthPanel.add(width_decr);
	width_incr.addActionListener(this);
	width_decr.addActionListener(this);
	width = dim.width;

	JPanel heightPanel = new JPanel();
	heightPanel.setLayout(new FlowLayout());
	heightArea = new TextField(3);
	heightArea.setText(""+dim.height);
	heightArea.setEditable(false);
	height_incr = new JButton("+");
	height_decr = new JButton("-");
	height_incr.setFont(bigBoldFont);
	height_decr.setFont(bigBoldFont);
	heightPanel.add(new JLabel("Height ("+minHeight+" to "+maxHeight+")"));
	heightPanel.add(heightArea);
	heightPanel.add(height_incr);
	heightPanel.add(height_decr);
	height_incr.addActionListener(this);
	height_decr.addActionListener(this);
	height = dim.height;

	JPanel buttonPanel = new JPanel();
	buttonPanel.setLayout(new FlowLayout());
	okButton = new JButton("OK");
	okButton.addActionListener(this);
	cancelButton = new JButton("Cancel");
	cancelButton.addActionListener(this);
	buttonPanel.add(okButton);
	buttonPanel.add(cancelButton);

	getContentPane().setLayout(new GridLayout(3,1));
	getContentPane().add(widthPanel);
	getContentPane().add(heightPanel);
	getContentPane().add(buttonPanel);

	addWindowListener(new WindowAdapter()
	{
		public void windowClosing(WindowEvent e)
		{
			setVisible(false);
			if (localFrame != null)
				localFrame.setVisible(false);
		}
	});
}

public void updateTarget()
{
	editor.record((Shape)target,RESIZE,target.width(),target.height());
	editor.set_redo(RESIZE,width,height);
	target.setSize(width,height);
	editor.repaint();
}

public void actionPerformed(ActionEvent evt)
{
	if (evt.getSource() instanceof JButton)
	{
		JButton tgt = (JButton)evt.getSource();
		if (tgt == width_incr)
		{
			if (width<maxWidth) {width++; widthArea.setText(""+width);}
		}
		else if (tgt == width_decr)
		{
			if (width>minWidth) {width--; widthArea.setText(""+width);}
		}
		else if (tgt == height_incr)
		{
			if (height<maxHeight) {height++; heightArea.setText(""+height);}
		}
		else if (tgt == height_decr)
		{
			if (height>minHeight) {height--; heightArea.setText(""+height);}
		}
		else if (tgt == okButton)
		{
			setVisible(false);
			if (localFrame != null)
				localFrame.setVisible(false);
			updateTarget();
			target = null;
		}
		else if (tgt == cancelButton)
		{
			setVisible(false);
			if (localFrame != null)
				localFrame.setVisible(false);
			target = null;
		}
	}
}

// Shortcuts to create a frame automatically, the frame is a static
// variable, so that all resize dialogs can use the same frame

public static void open(Shape elem, DrawingEditor ed)
{
	ResizeDialog dlg;

	// If the frame has not been created yet, create it.
	if (localFrame == null)
		localFrame = new JFrame("New size");
	// Create the dialog now
	dlg = new ResizeDialog(localFrame, elem, elem.getSize(), ed);
	// Shrink the frame to just fit the dialog
	localFrame.setSize(dlg.getSize().width,dlg.getSize().height);

	// Show the dialog

	dlg.pack();
	dlg.setVisible(true);
}

}	// End ResizeDialog

class UndoOperation
{
	public UndoEntity target;
	public int operator;
	public int operands[] = new int[3];

	UndoOperation(UndoEntity elem, int op)
	{
		target = elem;
		operator=op;
	}

	UndoOperation(UndoEntity elem, int op, int op1)
	{
		target = elem;
		operator=op;
		operands[0]=op1;
	}

	UndoOperation(UndoEntity elem, int op, int op1, int op2)
	{
		target = elem;
		operator=op;
		operands[0]=op1;
		operands[1]=op2;
	}

	UndoOperation(UndoEntity elem, int op, int op1, int op2, int op3)
	{
		target = elem;
		operator=op;
		operands[0]=op1;
		operands[1]=op2;
		operands[2]=op3;
	}
}

interface UndoEntity
{
public void undo(int operand, int op1, int op2, int op3, DrawingEditor editor);
}

interface AboutDialogUse {}

interface Colors
{
final Color LightYellow = new Color(255,255,200);
final Color LightPink = new Color(255,220,220);
final Color LightGreen = new Color(200,255,200);
final Color LightMagenta = new Color(255,200,255);
final Color Brown = new Color(185,120,100);
}

interface HasColors
{
public Color getFillColor();
public Color getBorderColor();
public void setFillColor(Color c);
public void setBorderColor(Color c);
public void recordFillColor(DrawingEditor ed);
public void recordBorderColor(DrawingEditor ed);
public void edit(DrawingEditor ed);
}

interface Constants
{
// Constants used for input
static int LINE = 1;
static int CIRCLE = 2;
static int POLYGON = 3;

//	Undo constants
int ADD = 600;
int DELETE = 601;
int MOVE = 602;
int RESIZE = 603;
int ROTATE = 604;
int EDIT = 605;
int START_BLOCK = 606;
int END_BLOCK = 607;
int CLEAR_ALL = 608;
int SET_FILL_COLOR = 610;
int SET_BORDER_COLOR = 611;
int SET_VERTEX = 614;

//	Redo constants (some undo constants are also used)
int MOVE_BY = 650;
int INCREASE_BY = 651;
int DECREASE_BY = 652;
int ROTATE_BY = 653;
int ALIGN = 658;
int DISTRIBUTE = 659;
int EQUALIZE = 660;
int EXECUTE = 661;
int SET_FILL_AND_BORDER_COLOR = 662;

}	// End Constants


